/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.jms.cluster;

import java.io.*;
import java.util.ArrayList;
import java.util.logging.*;

import javax.jms.*;
import javax.annotation.*;

import com.caucho.config.ConfigException;
import com.caucho.jms.JmsRuntimeException;
import com.caucho.jms2.*;
import com.caucho.jms.queue.*;
import com.caucho.jms.message.*;
import com.caucho.server.resin.*;
import com.caucho.server.cluster.*;
import com.caucho.server.hmux.*;
import com.caucho.util.*;
import com.caucho.vfs.*;

/**
 * Implements a cluster send queue.
 */
public class HmuxClient
{
  private static final L10N L = new L10N(ClientQueue.class);
  private static final Logger log
    = Logger.getLogger(ClientQueue.class.getName());

  private String _serverId = "client";
  private Server _selfServer;
  private Cluster _selfCluster;
  
  private Cluster _targetCluster;
  private ClusterServer []_targetServers;

  private MessageFactory _messageFactory = new MessageFactory();

  private ArrayList<String> _hostNames = new ArrayList<String>();

  private String _name;
  private String _subscriptionName;

  public HmuxClient()
  {
    _selfCluster = Cluster.getLocal();

    Resin resin = Resin.getLocal();

    if (resin != null) {
      _selfServer = resin.getServer();
      _serverId = resin.getServerId();
    }

    StringBuilder sb = new StringBuilder();
    sb.append(_serverId);
    sb.append('.');
    long rand = RandomUtil.getRandomLong();
    Base64.encode(sb, rand);
    Base64.encode(sb, Alarm.getExactTime());
    
    _subscriptionName = sb.toString();
  }

  public void setName(String name)
  {
    _name = name;
  }

  public String getName()
  {
    return _name;
  }

  public void setCluster(String clusterName)
  {
    Resin resin = Resin.getLocal();

    _targetCluster = resin.findCluster(clusterName);
  }
  
  public void addHost(String host)
  {
    _hostNames.add(host);

    if (host.indexOf(':') < 0) {
      throw new ConfigException(L.l("Host '{0}' must have an explicit port.",
				    host));
    }
  }

  @PostConstruct
  public void init()
  {
    if (_targetCluster != null && _hostNames.size() > 0)
      throw new ConfigException(L.l("HmuxClient must not have both a cluster and explicit hosts"));
    
    if (_targetCluster == null && _hostNames.size() == 0)
      throw new ConfigException(L.l("HmuxClient requires a valid cluster."));

    if (_targetCluster == null) {
      try {
	_targetCluster = new Cluster();

	for (int i = 0; i < _hostNames.size(); i++) {
	  String hostName = _hostNames.get(i);
	  String host;
	  int port;

	  int p = hostName.indexOf(':');

	  host = hostName.substring(0, p);
	  port = Integer.parseInt(hostName.substring(p + 1));
	
	  ClusterServer server = _targetCluster.createServer();
	  server.setId(hostName);
	  server.setAddress(host);
	  server.setPort(port);
	  server.init();

	  _targetCluster.addServer(server);
	}
      } catch (RuntimeException e) {
	throw e;
      } catch (Exception e) {
	throw ConfigException.create(e);
      }
    }

    _targetServers = _targetCluster.getServerList();

    if (_targetServers == null || _targetServers.length == 0) {
      throw new ConfigException(L.l("ClientQueue[{0}] has no target servers in cluster '{1}'.",
				    getName(), _targetCluster.getId()));

    }
  }

  public void generateMessageID(StringBuilder cb)
  {
    int index = RandomUtil.nextInt(_targetServers.length);

    ClusterServer server = _targetServers[index];

    server.generateBackupCode(cb);

    cb.append('.');
    cb.append(_serverId);
    cb.append('.');
  }

  /**
   * Adds the message to the persistent store.  Called if there are no
   * active listeners.
   */
  protected void enqueue(String queueName, MessageImpl msg, long expires)
    throws JMSException
  {
    try {
      String id = msg.getJMSMessageID();
    
      ClusterServer primary = _targetCluster.getPrimary(id, 3);
      ServerConnector connector = primary.getServerConnector();

      ClusterStream stream = connector.openIfLive();

      if (stream == null)
	throw new JMSException(L.l("{0} can't connect to {1}", this, connector));

      try {
	ReadStream rs = stream.getReadStream();
	WriteStream ws = stream.getWriteStream();

	if (log.isLoggable(Level.FINE))
	  log.fine(stream.getDebugId() + " jms sending " + msg);
      
	writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_JMS_PROTOCOL);
	writeRequestString(ws, HmuxReceiver.HMUX_QUEUE, queueName);
	
	writeInt(ws, HmuxReceiver.HMUX_MESSAGE, msg.getType().ordinal());
	writeRequestString(ws, HmuxReceiver.HMUX_MSG_ID,
			   msg.getJMSMessageID());

	OutputStream propStream;
	propStream = new DataStream(ws, HmuxReceiver.HMUX_PROPERTIES);

	msg.writeProperties(propStream);

	propStream.close();

	OutputStream bodyWriter = new DataStream(ws, HmuxReceiver.HMUX_BODY);

	msg.writeBody(bodyWriter);

	bodyWriter.close();
	
	writeRequestData(ws, HmuxReceiver.HMUX_SEND);
	ws.write(HmuxRequest.HMUX_QUIT);
        
        stream = waitForQuit(stream, rs);
      } finally {
	if (stream != null)
	  stream.close();
      }
    } catch (IOException e) {
      throw new JMSExceptionWrapper(e);
    }
  }

  /**
   * Polls the next message from the store.
   */
  public MessageImpl receive(String queueName,
                             boolean isAutoAck,
                             String subscriber)
    throws JMSException
  {
    try {
      ClusterServer primary = _targetCluster.getServerList()[0];
      ServerConnector connector = primary.getServerConnector();

      ClusterStream stream = connector.openIfLive();

      if (stream == null)
	throw new JMSException(L.l("{0} can't connect to {1}", this, connector));

      boolean isLoggable = log.isLoggable(Level.FINER);

      try {
	ReadStream rs = stream.getReadStream();
	WriteStream ws = stream.getWriteStream();

	writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_JMS_PROTOCOL);
	writeRequestString(ws, HmuxReceiver.HMUX_QUEUE, queueName);

        if (subscriber != null)
          writeRequestString(ws, HmuxReceiver.HMUX_SUBSCRIBER, subscriber);
        
	ws.write(HmuxReceiver.HMUX_RECEIVE);
        ws.write(0);
        ws.write(1);
        ws.write(isAutoAck ? 1 : 0);
	ws.write(HmuxRequest.HMUX_QUIT);

        if (isLoggable)
          log.finer(stream.getDebugId() + " jms query receive " + (isAutoAck ? " (auto-ack)" : ""));

	ws.flush();
	int value;

	HmuxDataInputStream dataStream;
	MessageImpl msg = null;
	// XXX: need to handle acks

	while (true) {
	  int code = rs.read();

	  switch (code) {
	  case HmuxRequest.HMUX_QUIT:
            if (isLoggable)
              log.finer(stream.getDebugId() + " Q quit");
            
	    stream.free();
	    stream = null;
	    return msg;
	    
	  case HmuxRequest.HMUX_EXIT:
            if (isLoggable)
              log.finer(stream.getDebugId() + " X exit");
            
	    return msg;

	  case HmuxReceiver.HMUX_MESSAGE:
	    int len = (rs.read() << 8) + rs.read();

	    value = ((rs.read() << 24)
		     + (rs.read() << 16)
		     + (rs.read() << 8)
		     + (rs.read()));

	    msg = _messageFactory.createMessage(value);

	    if (isLoggable)
	      log.finer(stream.getDebugId() + " jms receive message " + msg);
	
	    break;

	  case HmuxReceiver.HMUX_MSG_ID:
	    len = (rs.read() << 8) + rs.read();
	    CharBuffer cb = new CharBuffer();
	    rs.readAll(cb, len);

	    msg.setJMSMessageID(cb.toString());

	    if (isLoggable)
	      log.finer(stream.getDebugId() + " jms msg id " + cb);
	    break;

	  case HmuxReceiver.HMUX_REDELIVERED:
	    len = (rs.read() << 8) + rs.read();

	    msg.setJMSRedelivered(true);

	    if (isLoggable)
	      log.finer(stream.getDebugId() + " jms redelivered");
	    break;

	  case HmuxReceiver.HMUX_PROPERTIES:
	    len = (rs.read() << 8) + rs.read();

	    dataStream = new HmuxDataInputStream(rs, code, len);

	    msg.readProperties(dataStream);

	    dataStream.close();

	    if (isLoggable)
	      log.finer(stream.getDebugId() + " jms receive properties");
	    break;

	  case HmuxReceiver.HMUX_BODY:
	    len = (rs.read() << 8) + rs.read();

	    dataStream = new HmuxDataInputStream(rs, code, len);
	    msg.readBody(dataStream);
	    dataStream.close();

	    if (isLoggable)
	      log.finer(stream.getDebugId() + " jms body");
	    break;

	  default:
	    if (log.isLoggable(Level.FINE))
	      log.fine(stream.getDebugId() + " jms unknown code " + code + " " + (char) code);
	    
	    return msg;
	  }
	}
      } finally {
	if (stream != null)
	  stream.close();
      }
    } catch (IOException e) {
      e.printStackTrace();
      throw new JMSExceptionWrapper(e);
    } catch (Exception e) {
      e.printStackTrace();
      throw new RuntimeException(e);
    }
  }

  boolean hasMessage()
  {
    try {
      ClusterServer primary = _targetCluster.getServerList()[0];
      ServerConnector connector = primary.getServerConnector();

      ClusterStream stream = connector.openIfLive();

      if (stream == null)
	throw new JmsRuntimeException(L.l("{0} can't connect to {1}", this, connector));

      try {
        ReadStream rs = stream.getReadStream();
        WriteStream ws = stream.getWriteStream();

        writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_JMS_PROTOCOL);
        writeRequestString(ws, HmuxReceiver.HMUX_QUEUE, getName());
	
        writeRequestData(ws, HmuxReceiver.HMUX_POLL);
        ws.write(HmuxRequest.HMUX_QUIT);

        if (log.isLoggable(Level.FINER))
          log.finer(stream.getDebugId() + " jms poll");

        ws.flush();
        boolean hasMessage = false;

        while (true) {
          int code = rs.read();

          switch (code) {
          case HmuxReceiver.HMUX_HAS_MESSAGE:
            if (log.isLoggable(Level.FINER))
              log.finer(stream.getDebugId() + " " + (char) code + ": jms has message");
            rs.skip(2);
            hasMessage = true;
            break;
            
          case HmuxRequest.HMUX_QUIT:
            if (log.isLoggable(Level.FINER))
              log.finer(stream.getDebugId() + " " + (char) code + ": jms quit");
            stream.free();
            stream = null;
            return hasMessage;
	    
          case HmuxRequest.HMUX_EXIT:
            if (log.isLoggable(Level.FINER))
              log.finer(stream.getDebugId() + " " + (char) code + ": jms exit");
          
            return hasMessage;
	    
          case -1:
            if (log.isLoggable(Level.FINER))
              log.finer(stream.getDebugId() + " " + (char) code + ": end of file");
            return hasMessage;

          default:
            if (log.isLoggable(Level.FINE))
              log.finer(stream.getDebugId() + " " + (char) code + ": jms unknown");
            return hasMessage;
          }
        }
      } finally {
        if (stream != null)
          stream.close();
      }
    } catch (IOException e) {
      throw new JmsRuntimeException();
    }
  }

  public void acknowledge(String msgId)
  {
    try {
      ClusterServer primary = _targetCluster.getServerList()[0];
      ServerConnector connector = primary.getServerConnector();

      ClusterStream stream = connector.openIfLive();

      if (stream == null)
	throw new JmsRuntimeException(L.l("{0} can't connect to {1}", this, connector));

      try {
        ReadStream rs = stream.getReadStream();
        WriteStream ws = stream.getWriteStream();

        writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_JMS_PROTOCOL);
        writeRequestString(ws, HmuxReceiver.HMUX_QUEUE, getName());
	
        writeRequestString(ws, HmuxReceiver.HMUX_ACK, msgId);
        ws.write(HmuxRequest.HMUX_QUIT);

        ws.flush();

        stream = waitForQuit(stream, rs);
      } finally {
        if (stream != null)
          stream.close();
      }
    } catch (IOException e) {
      throw new JmsRuntimeException();
    }
  }

  public void rollback(String msgId)
  {
    try {
      ClusterServer primary = _targetCluster.getServerList()[0];
      ServerConnector connector = primary.getServerConnector();

      ClusterStream stream = connector.openIfLive();

      if (stream == null)
	throw new JmsRuntimeException(L.l("{0} can't connect to {1}", this, connector));
      
      try {
        ReadStream rs = stream.getReadStream();
        WriteStream ws = stream.getWriteStream();

        writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_JMS_PROTOCOL);
        writeRequestString(ws, HmuxReceiver.HMUX_QUEUE, getName());
	
        writeRequestString(ws, HmuxReceiver.HMUX_ROLLBACK, msgId);
        ws.write(HmuxRequest.HMUX_QUIT);

        ws.flush();
        int value;

        stream = waitForQuit(stream, rs);
      } finally {
        if (stream != null)
          stream.close();
      }
    } catch (IOException e) {
      throw new JmsRuntimeException();
    }
  }

  private ClusterStream waitForQuit(ClusterStream stream,
                                    ReadStream rs)
    throws IOException
  {
    while (true) {
      int code = rs.read();

      switch (code) {
      case HmuxRequest.HMUX_QUIT:
        if (log.isLoggable(Level.FINER))
          log.finer(stream.getDebugId() + " " + (char) code + ": jms quit");
          
        stream.free();
        return null;
	    
      case HmuxRequest.HMUX_EXIT:
        if (log.isLoggable(Level.FINER))
          log.finer(stream.getDebugId() + " " + (char) code + ": jms exit");
          
        return stream;
	    
      case -1:
        if (log.isLoggable(Level.FINER))
          log.finer(stream.getDebugId() + " " + (char) code + ": end of file");
          
        return stream;

      default:
        if (log.isLoggable(Level.FINE))
          log.finer(stream.getDebugId() + " " + (char) code + ": jms unknown");
	    
        return stream;
      }
    }
  }

  private void writeRequestString(WriteStream s, int code, String string)
    throws IOException
  {
    if (string == null) {
      s.write(code);
      s.write(0);
      s.write(0);
    }
    else {
      int length = string.length();
      s.write(code);
      s.write(length >> 8);
      s.write(length);

      s.print(string, 0, length);
    }
  }

  private void writeRequestString(WriteStream s, int code,
				  char []buffer, int offset, int length)
    throws IOException
  {
    s.write(code);
    s.write(length >> 8);
    s.write(length);

    s.print(buffer, offset, length);
  }

  private void writeInt(WriteStream s, int code, int value)
    throws IOException
  {
    s.write(code);
    s.write(0);
    s.write(4);
    s.write(value >> 24);
    s.write(value >> 16);
    s.write(value >> 8);
    s.write(value);
  }

  private void writeRequestData(WriteStream s, int code)
    throws IOException
  {
    s.write(code);
    s.write(0);
    s.write(0);
  }

  private void writeRequestData(WriteStream os,
                                int code, byte []data, int length)
    throws IOException
  {
    os.write(code);
    os.write(length >> 8);
    os.write(length);

    os.write(data, 0, length);
  }

  protected String dbgId()
  {
    return "JmsClientQueue[] ";
  }

  static class DataStream extends OutputStream {
    private WriteStream _out;
    private int _code;
    private boolean _hasData;

    DataStream(WriteStream out, int code)
    {
      _out = out;
      _code = code;
    }

    public void write(int ch)
      throws IOException
    {
      _hasData = true;

      _out.write(_code);
      _out.write(0);
      _out.write(1);
      _out.write(ch);
    }

    public void write(byte []buffer, int offset, int length)
      throws IOException
    {
      _hasData = true;

      _out.write(_code);
      _out.write(length >> 8);
      _out.write(length);
      _out.write(buffer, offset, length);
    }

    public void close()
      throws IOException
    {
      if (_hasData) {
	_out.write(HmuxReceiver.HMUX_DATA_END);
	_out.write(0);
	_out.write(0);
      }

      _out = null;
    }
  }
}
