/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.jms.cluster;

import java.io.*;
import java.lang.ref.*;
import java.util.*;
import java.util.logging.*;

import javax.jms.JMSException;

import com.caucho.config.ConfigException;
import com.caucho.loader.*;
import com.caucho.jms.message.*;
import com.caucho.jms.queue.AbstractQueue;
import com.caucho.server.hmux.*;
import com.caucho.util.*;
import com.caucho.vfs.*;

/**
 * Receives hmux requests
 */
public class HmuxReceiver implements HmuxExtension
{
  private static final L10N L = new L10N(HmuxReceiver.class);
  
  private static final Logger log
    = Logger.getLogger(HmuxReceiver.class.getName());

  private static final EnvironmentLocal<HmuxReceiver> _localReceiver
    = new EnvironmentLocal<HmuxReceiver>();
  
  public static final int HMUX_ACK = 'a';
  public static final int HMUX_QUEUE = 'b';
  
  public static final int HMUX_MESSAGE = 'c';
  public static final int HMUX_MSG_ID = 'd';
  
  public static final int HMUX_RECEIVE = 'e';
  public static final int HMUX_SEND = 'f';
  public static final int HMUX_SUBSCRIBER = 'g';
  
  public static final int HMUX_POLL = 'h';
  public static final int HMUX_HAS_MESSAGE = 'i';
  public static final int HMUX_ROLLBACK = 'j';
  public static final int HMUX_REDELIVERED = 'k';
  
  public static final int HMUX_PROPERTIES = 'l';
  public static final int HMUX_BODY = 'm';
  public static final int HMUX_DATA_END = 'z';

  private HmuxProtocol _hmuxProtocol;

  private HashMap<String,WeakReference<ServerDestination>> _queueMap
    = new HashMap<String,WeakReference<ServerDestination>>();

  private HmuxReceiver()
  {
    _hmuxProtocol = HmuxProtocol.getLocal();

    if (_hmuxProtocol == null)
      throw new ConfigException(L.l("JmsCluster is missing a configured cluster-port."));

    _hmuxProtocol.putExtension(HmuxRequest.HMUX_JMS_PROTOCOL, this);
  }

  public static HmuxReceiver create()
  {
    synchronized (_localReceiver) {
      HmuxReceiver receiver = _localReceiver.get();
      
      if (receiver == null) {
	receiver = new HmuxReceiver();

	_localReceiver.set(receiver, receiver.getClassLoader());
      }

      return receiver;
    }
  }

  ClassLoader getClassLoader()
  {
    return _hmuxProtocol.getClassLoader();
  }

  public ServerDestination getQueue(String name)
  {
    synchronized (_queueMap) {
      WeakReference<ServerDestination> ref = _queueMap.get(name);

      if (ref != null)
	return ref.get();
      else
	return null;
    }
  }

  public void setQueue(String name, ServerDestination queue)
  {
    synchronized (_queueMap) {
      WeakReference<ServerDestination> ref = _queueMap.get(name);

      if (ref != null && ref.get() != null) {
	// throw new IllegalStateException(L.l("{0} already exists", queue));
        log.warning(L.l("{0} already exists", queue));
        return;
      }
      
      log.fine(L.l("Jms adding cluster queue {0}", queue));

      _queueMap.put(name, new WeakReference<ServerDestination>(queue));
    }
  }

  public void removeQueue(String name)
  {
    log.fine(L.l("Jms removing cluster queue {0}", name));

    synchronized (_queueMap) {
      _queueMap.remove(name);
    }
  }

  /**
   * Handles the request from the server.
   */
  public int handleRequest(HmuxRequest req,
			   ReadStream is,
			   WriteStream os)
    throws IOException
  {
    CharBuffer cb = new CharBuffer();
    boolean isLoggable = log.isLoggable(Level.FINER);
    int code;
    int len;
    ServerDestination queue = null;
    MessageImpl msg = null;
    HmuxDataInputStream dataStream;
    String subscriber = null;
    long timeout = 0;
    int value;

    try {
      while (true) {
	code = is.read();

	switch (code) {
	case -1:
	  if (isLoggable)
	    log.finer(req.dbgId() + "jms end of file");
	  return HmuxRequest.HMUX_EXIT;

	case HMUX_QUEUE:
	  len = (is.read() << 8) + is.read();
	  cb.clear();
	  is.readAll(cb, len);

	  queue = getQueue(cb.toString());

	  if (queue == null) {
	    if (log.isLoggable(Level.FINE))
	      log.fine(req.dbgId() + (char) code + ": jms queue[" + cb + "] does not exist");
	    return HmuxRequest.HMUX_EXIT;
	  }
	
	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms queue[" + cb + "] " + queue);
	
	  break;

	case HMUX_SUBSCRIBER:
	  len = (is.read() << 8) + is.read();
	  cb.clear();
	  is.readAll(cb, len);

	  subscriber = cb.toString();

          if (isLoggable)
            log.fine(req.dbgId() + (char) code + ": jms subscriber " + subscriber);
	  break;

	case HMUX_MESSAGE:
	  len = (is.read() << 8) + is.read();

	  value = ((is.read() << 24)
		   + (is.read() << 16)
		   + (is.read() << 8)
		   + (is.read()));

	  msg = queue.createMessage(value);

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms message " + msg);
	
	  break;

	case HMUX_MSG_ID:
	  len = (is.read() << 8) + is.read();
	  cb.clear();
	  is.readAll(cb, len);

	  msg.setJMSMessageID(cb.toString());

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms msg-id " + cb);
	
	  break;

	case HMUX_PROPERTIES:
	  len = (is.read() << 8) + is.read();

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms header(" + len + ")");

	  dataStream = new HmuxDataInputStream(is, code, len);

	  msg.readProperties(dataStream);
	  break;

	case HMUX_BODY:
	  len = (is.read() << 8) + is.read();

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms body(" + len + ")");

	  dataStream = new HmuxDataInputStream(is, code, len);
	  msg.readBody(dataStream);
	  dataStream.close();
	  break;

	case HMUX_SEND:
	  len = (is.read() << 8) + is.read();

	  queue.sendCluster(msg, timeout);

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms send " + msg + " to " + queue);

	  break;

	case HMUX_RECEIVE:
	  len = (is.read() << 8) + is.read();
          boolean isAutoAck = is.read() != 0;
	  msg = queue.receiveCluster(subscriber, isAutoAck);

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms receive " + msg + " from " + queue + (isAutoAck ? " (auto-ack)" : ""));

	  if (msg != null) {
	    sendMessage(os, msg);
	  }
	  break;

	case HMUX_ACK:
	  len = (is.read() << 8) + is.read();
	  cb.clear();
	  is.readAll(cb, len);

          queue.acknowledge(cb.toString());

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms acknowledge " + cb + " for " + queue);
	  break;

	case HMUX_ROLLBACK:
	  len = (is.read() << 8) + is.read();
	  cb.clear();
	  is.readAll(cb, len);

          queue.rollback(cb.toString());

	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms rollback " + cb + " for " + queue);
	  break;

	case HMUX_POLL:
	  len = (is.read() << 8) + is.read();

	  if (queue.hasMessage(subscriber)) {
            if (isLoggable)
              log.finer(req.dbgId() + (char) code + ": jms poll has message");
            
            os.write(HMUX_HAS_MESSAGE);
            os.write(0);
            os.write(0);
          }
          else {
            if (isLoggable)
              log.finer(req.dbgId() + (char) code + ": jms poll");
          }
	  break;

	case HmuxRequest.HMUX_QUIT:
	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms end of request");
	
	  return HmuxRequest.HMUX_QUIT;

	case HmuxRequest.HMUX_EXIT:
	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms end of socket");

	  return HmuxRequest.HMUX_EXIT;

	case HmuxRequest.HMUX_YIELD:
	  if (isLoggable)
	    log.finer(req.dbgId() + (char) code + ": jms yield");

	  return HmuxRequest.HMUX_YIELD;
        
	default:
	  len = (is.read() << 8) + is.read();

	  if (log.isLoggable(Level.FINE))
	    log.fine(req.dbgId() + (char) code + " " + len + " (jms)");
	
	  is.skip(len);

	  return HmuxRequest.HMUX_EXIT;
	}
      }
    } catch (JMSException e) {
      log.log(Level.WARNING, e.toString(), e);
      
      return HmuxRequest.HMUX_EXIT;
    }
  }

  private void sendMessage(WriteStream out, MessageImpl msg)
    throws IOException
  {
    writeInt(out, HmuxReceiver.HMUX_MESSAGE, msg.getType().ordinal());
    writeRequestString(out, HmuxReceiver.HMUX_MSG_ID,
		       msg.getJMSMessageID());

    OutputStream propStream;
    propStream = new HmuxDataOutputStream(out, HmuxReceiver.HMUX_PROPERTIES);

    msg.writeProperties(propStream);

    propStream.close();

    OutputStream bodyWriter
      = new HmuxDataOutputStream(out, HmuxReceiver.HMUX_BODY);

    msg.writeBody(bodyWriter);

    bodyWriter.close();

    if (msg.getJMSRedelivered())
      writeRequestData(out, HmuxReceiver.HMUX_REDELIVERED);
  }

  private void writeInt(WriteStream s, int code, int value)
    throws IOException
  {
    s.write(code);
    s.write(0);
    s.write(4);
    s.write(value >> 24);
    s.write(value >> 16);
    s.write(value >> 8);
    s.write(value);
  }

  private void writeRequestData(WriteStream s, int code)
    throws IOException
  {
    s.write(code);
    s.write(0);
    s.write(0);
  }

  private void writeRequestString(WriteStream s, int code, String string)
    throws IOException
  {
    if (string == null) {
      s.write(code);
      s.write(0);
      s.write(0);
    }
    else {
      int length = string.length();
      s.write(code);
      s.write(length >> 8);
      s.write(length);

      s.print(string, 0, length);
    }
  }

  public String toString()
  {
    return "HmuxJmsReceiver[]";
  }

  static class DataStream extends InputStream {
    private ReadStream _is;
    private int _code;
    private int _length;

    DataStream(ReadStream is, int code, int length)
    {
      _is = is;
      _code = code;
      _length = length;
    }

    public int read()
      throws IOException
    {
      if (_is == null)
	return -1;
      else if (_length == 0) {
	int code = _is.read();
	int len = (_is.read() << 8) + _is.read();
	
	if (code != _code) {
	  System.out.println("CODE: " + (char) code);
	  _is.skip(len);
	  _is = null;

	  return -1;
	}

	_length = len;
      }

      _length--;

      return _is.read();
    }

    public int read(byte []buffer, int offset, int length)
      throws IOException
    {
      if (_is == null)
	return -1;
      else if (_length == 0) {
	int code = _is.read();
	int len = (_is.read() << 8) + _is.read();
	
	if (code != _code) {
	  _is.skip(len);
	  _is = null;

	  return -1;
	}

	_length = len;
      }

      if (_length < length)
	length = _length;

      length = _is.readAll(buffer, offset, length);

      _length -= length;

      return length;
    }

    public void close()
      throws IOException
    {
      while (read() >= 0) {
      }
    }
  }
}

