/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.jms.cluster;

import java.io.*;
import java.util.logging.*;

import javax.annotation.*;
import javax.jms.*;

import com.caucho.config.ConfigException;
import com.caucho.loader.Environment;
import com.caucho.jms.message.*;
import com.caucho.jms.queue.*;
import com.caucho.jms.connection.*;
import com.caucho.jms.file.*;
import com.caucho.server.cluster.*;
import com.caucho.util.L10N;
import com.caucho.vfs.*;

/**
 * Implements a cluster receive queue.
 */
public class ServerQueue extends AbstractQueue
  implements ServerDestination
{
  private static final L10N L = new L10N(ServerQueue.class);
  private static final Logger log
    = Logger.getLogger(ServerQueue.class.getName());

  private Cluster _cluster;
  private MessageFactory _factory;

  private HmuxReceiver _hmuxReceiver;

  private FileQueue _fileQueue;
  private volatile boolean _hasMessage;

  private String _fullName;

  public ServerQueue()
  {
    _cluster = Cluster.getLocal();
    _factory = new MessageFactory();

    _fileQueue = new FileQueue();
  }

  public ServerQueue(String name)
  {
    this();
    
    setName(name);

    init();
  }

  public void setPath(Path path)
  {
    _fileQueue.setPath(path);
  }

  //
  // JMX configuration items
  //

  /**
   * Returns the JMS configuration url.
   */
  public String getUrl()
  {
    return "cluster:name=" + getName() + ";path=" + _fileQueue.getPath().getURL();
  }

  //
  // JMX stats
  //

  public int getQueueSize()
  {
    return _fileQueue.getQueueSize();
  }

  public void init()
  {
    if (getName() == null || "".equals(getName()))
      throw new ConfigException(L.l("ServerQueue requires a non-null name."));

    _fileQueue.setName(getName());

    _fileQueue.init();

    String envName = Environment.getEnvironmentName();

    int p = envName.indexOf(':');
    if (p > 0)
      envName = envName.substring(p + 1);

    if (envName.endsWith("/") || getName().startsWith("/"))
      _fullName = envName + getName();
    else
      _fullName = envName + "/" + getName();
      
    _hmuxReceiver = HmuxReceiver.create();

    if (_hmuxReceiver.getQueue(getName()) == null)
      _hmuxReceiver.setQueue(getName(), this);
    
    _hmuxReceiver.setQueue(_fullName, this);
  }

  public MessageImpl createMessage(int msg)
  {
    return _factory.createMessage(msg);
  }

  @Override
  public void send(JmsSession session, MessageImpl msg, long expires)
    throws JMSException
  {
    _hasMessage = true;
    _fileQueue.send(session, msg, expires);

    notifyMessageAvailable();
  }

  @Override
  public void rollback(String msgId)
  {
    _fileQueue.rollback(msgId);
  }

  @Override
  public void acknowledge(String msgId)
  {
    _fileQueue.acknowledge(msgId);
  }

  /**
   * Polls the next message from the store.  If no message is available,
   * wait for the timeout.
   */
  @Override
  public MessageImpl receive(boolean isAutoAck)
  {
    return _fileQueue.receive(isAutoAck);
  }

  public boolean hasMessage()
  {
    boolean hasMessage;
    
    synchronized (this) {
      hasMessage = _hasMessage;
      _hasMessage = false;
    }

    return hasMessage;
  }

  public void sendCluster(MessageImpl msg, long timeout)
    throws JMSException
  {
    _fileQueue.send(null, msg, timeout);

    _hasMessage = true;

    notifyMessageAvailable();
  }

  /**
   * Polls the next message from the store.  If no message is available,
   * wait for the timeout.
   */
  public MessageImpl receiveCluster(String subscription, boolean isAutoAck)
  {
    return _fileQueue.receive(isAutoAck);
  }

  public boolean hasMessage(String subscription)
  {
    return hasMessage();
  }

  @PreDestroy
  public void close()
  {
    _hmuxReceiver.removeQueue(getName());
  }
}

