/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.jms.cluster;

import java.io.*;
import java.util.HashMap;
import java.util.logging.*;

import javax.annotation.*;
import javax.jms.*;

import com.caucho.config.ConfigException;
import com.caucho.jms.message.*;
import com.caucho.jms.queue.*;
import com.caucho.jms.connection.*;
import com.caucho.jms.file.*;
import com.caucho.server.cluster.*;
import com.caucho.util.Alarm;
import com.caucho.util.L10N;
import com.caucho.vfs.*;

/**
 * Implements a cluster receive topic.
 */
public class ServerTopic extends AbstractTopic
  implements ServerDestination
{
  private static final L10N L = new L10N(ServerTopic.class);
  private static final Logger log
    = Logger.getLogger(ServerTopic.class.getName());

  private Cluster _cluster;
  private MessageFactory _factory;

  private HmuxReceiver _hmuxReceiver;

  private FileTopic _fileTopic;
  private volatile boolean _hasMessage;

  private long _subscriberTimeout = 15 * 60000L;

  private HashMap<String,Subscriber> _subscriberMap
    = new HashMap<String,Subscriber>();

  public ServerTopic()
  {
    _cluster = Cluster.getLocal();
    _factory = new MessageFactory();

    _fileTopic = new FileTopic();
  }

  public ServerTopic(String name)
  {
    this();
    
    setName(name);

    init();
  }

  public void setPath(Path path)
  {
    _fileTopic.setPath(path);
  }

  //
  // JMX configuration attributes
  //

  /**
   * Returns the JMS configuration url.
   */
  public String getUrl()
  {
    return "cluster:name=" + getName();
  }

  public MessageImpl createMessage(int msg)
  {
    return _factory.createMessage(msg);
  }

  @PostConstruct
  public void init()
  {
    if (getName() == null || "".equals(getName()))
      throw new ConfigException(L.l("ServerTopic requires a non-null name."));

    _fileTopic.setName(getName());

    _fileTopic.init();

    _hmuxReceiver = HmuxReceiver.create();

    _hmuxReceiver.setQueue(getName(), this);
  }

  @Override
  public void send(JmsSession session, MessageImpl msg, long expires)
    throws JMSException
  {
    _hasMessage = true;
    _fileTopic.send(session, msg, expires);

    notifyMessageAvailable();
  }

  @Override
  public void rollback(String msgId)
  {
    _fileTopic.rollback(msgId);
  }

  @Override
  public void acknowledge(String msgId)
  {
    _fileTopic.acknowledge(msgId);
  }

  /**
   * Polls the next message from the store.  If no message is available,
   * wait for the timeout.
   */
  @Override
  public MessageImpl receive(boolean isAutoAck)
    throws JMSException
  {
    return _fileTopic.receive(isAutoAck);
  }

  public void sendCluster(MessageImpl msg, long timeout)
    throws JMSException
  {
    _fileTopic.send(null, msg, timeout);

    _hasMessage = true;

    notifyMessageAvailable();
  }

  public MessageImpl receiveCluster(String subscriber, boolean isAutoAck)
    throws JMSException
  {
    Subscriber sub = getSubscriber(subscriber);

    return sub.getQueue().receive(true);
  }

  public boolean hasMessage(String subscriber)
  {
    Subscriber sub = getSubscriber(subscriber);

    return sub.getQueue().hasMessage();
  }

  private Subscriber getSubscriber(String subscriber)
  {
    synchronized (_subscriberMap) {
      Subscriber sub = _subscriberMap.get(subscriber);

      if (sub == null) {
        sub = new Subscriber(subscriber,
                             createSubscriber(null, subscriber, false));
        _subscriberMap.put(subscriber, sub);
      }
      else
        sub.access();

      return sub;
    }
  }

  public void notifyMessageAvailable()
  {
  }

  public AbstractQueue createSubscriber(JmsSession session,
                                        String name,
                                        boolean noLocal)
  {
    return _fileTopic.createSubscriber(session, name, noLocal);
  }

  public void closeSubscriber(AbstractQueue queue)
  {
    _fileTopic.closeSubscriber(queue);
  }

  @PreDestroy
  public void close()
  {
    _hmuxReceiver.removeQueue(getName());
  }

  class Subscriber {
    private String _name;
    private long _expires;
    private AbstractQueue _queue;

    Subscriber(String name, AbstractQueue queue)
    {
      _name = name;
      _expires = Alarm.getCurrentTime() + _subscriberTimeout;
      _queue = queue;
    }

    boolean isExpired()
    {
      return _expires < Alarm.getCurrentTime();
    }

    void access()
    {
      _expires = Alarm.getCurrentTime() + _subscriberTimeout;
    }

    AbstractQueue getQueue()
    {
      return _queue;
    }
  }
}

