/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.profile;

import java.util.logging.*;

import com.caucho.config.*;
import com.caucho.util.*;

/**
 * Resin's bootstrap class.
 */
public class Profile implements Runnable {
  private static final Logger log = Logger.getLogger(Profile.class.getName());
  private static final L10N L = new L10N(Profile.class);
  
  private static Profile _profile = new Profile();
  
  private static long _jniProfile;
  private static long _period = 1000L;
  private static long _ticks;

  private static boolean _isRunning;

  private static int _maxDepth = 32;

  public Profile()
  {
    synchronized (Profile.class) {
      if (_jniProfile == 0) {
	_jniProfile = nativeCreateProfile(16 * 1024);
      }
    }
  }

  public int getDepth()
  {
    return _maxDepth;
  }

  public void setDepth(int depth)
  {
    _maxDepth = depth;
  }
  
  public void setPeriod(long period)
  {
    synchronized (Profile.class) {
      if (period < 10)
	throw new ConfigException(L.l("profile period '{0}ms' is too small.  The period must be greater than 10ms.",
				      period));

      _profile._period = period;
    }
  }
  
  public long getPeriod()
  {
    synchronized (Profile.class) {
      return _profile._period;
    }
  }

  public boolean isActive()
  {
    return _isRunning;
  }

  public void start()
  {
    synchronized (Profile.class) {
      if (_isRunning)
	return;

      _isRunning = true;

      try {
	nativeClear(_jniProfile);
	
	Thread profileThread = new Thread(this, "resin-profile");
	profileThread.start();
      } catch (Exception e) {
	log.log(Level.WARNING, e.toString(), e);
      }
    }
  }

  public void stop()
  {
    synchronized (Profile.class) {
      _isRunning = false;

      Profile.class.notifyAll();
    }
  }

  public ProfileEntry []getResults()
  {
    synchronized (Profile.class) {
      ProfileEntry []results = nativeDisplay(_jniProfile);

      return results;
    }
  }

  public void doNative()
  {
    synchronized (Profile.class) {
      _ticks++;
      nativeProfile(_jniProfile, _maxDepth);
    }
  }
  
  public void run()
  {
    synchronized (Profile.class) {
      while (_isRunning) {
	long now = System.currentTimeMillis();
      
	long delta = now % _period;

	try {
	  Profile.class.wait(_period - delta);
	} catch (Exception e) {
	}

	_ticks++;
	nativeProfile(_jniProfile, _maxDepth);
      }
    }
  }
  
  private static native long nativeCreateProfile(int size);
  
  private static native boolean nativeProfile(long profile, int maxDepth);
  
  private static native void nativeClear(long profile);
  
  private static native ProfileEntry []nativeDisplay(long profile);
  
  static {
    try {
      System.loadLibrary("resin");
    } catch (Throwable e) {
      log.log(Level.FINE, e.toString(), e);
    }
  }
}
