/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.Location;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.util.L10N;

import java.io.IOException;
import java.util.logging.Logger;

/**
 * Represents compilable PHP expression.
 */
abstract public class GenerateExpr {
  private static final L10N L = new L10N(GenerateExpr.class);
  private static final Logger log
    = Logger.getLogger(GenerateExpr.class.getName());

  private Location _location = Location.UNKNOWN;
  
  public static final int COMPILE_ARG_MAX = 5;

  public GenerateExpr(Location location)
  {
    _location = location;
  }
  
  //
  // Java code generation
  //

  /**
   * Properties.
   */

  public Location getLocation()
  {
    return _location;
  }
  
  /**
   * Returns true for a long-valued expression.
   */
  public boolean isLong()
  {
    return false;
  }

  /**
   * Returns true for a double-valued expression.
   */
  public boolean isDouble()
  {
    return false;
  }

  /**
   * Returns true for a number.
   */
  public boolean isNumber()
  {
    return isLong() || isDouble();
  }
  
  /*
   * Returns true for a boolean.
   */
  public boolean isBoolean()
  {
    return false;
  }

  /**
   * Returns true for a String-valued expression.
   */
  public boolean isString()
  {
    return false;
  }

  /**
   * Analyze the expression
   */
  abstract public void analyze(AnalyzeInfo info);
  
  /**
   * Analyze the expression
   */
  public void analyzeAssign(AnalyzeInfo info)
  {
    analyze(info);
  }

  /**
   * Analyze as modified.
   */
  public void analyzeSetModified(AnalyzeInfo info)
  {
    analyze(info);
  }

  /**
   * Analyze as reference
   */
  public void analyzeSetReference(AnalyzeInfo info)
  {
    analyze(info);
  }

  /**
   * Analyze as unset
   */
  public void analyzeUnset(AnalyzeInfo info)
  {
    analyze(info);
  }

  /**
   * Returns the variables state.
   *
   * @param var the variables to test
   * @param owner the owning expression
   */
  /*
  public VarState getVarState(VarExpr var, VarExpr owner)
  {
    return VarState.UNKNOWN;
  }
  */

  /**
   * Returns the variables state.
   *
   * @param leftState the variables to test
   * @param rightState the owning expression
   */
  protected VarState combineBinaryVarState(VarState leftState,
					   VarState rightState)
  {
    if (leftState == VarState.UNSET || leftState == VarState.UNDEFINED)
      return leftState;

    if (rightState == VarState.UNSET || rightState == VarState.UNDEFINED)
      return rightState;

    if (leftState == VarState.VALID || rightState == VarState.VALID)
      return VarState.VALID;
    else
      return VarState.UNKNOWN;
  }

  /**
   * Returns true if the variable is ever assigned.
   *
   * @param var the variable to test
   */
  public boolean isVarAssigned(VarExpr var)
  {
    return false;
  }

  /**
   * Generates code to evaluate the expression.
   *
   * @param out the writer to the Java source code.
   */
  public void generate(PhpWriter out)
    throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName());
  }

  /**
   * Generates code to evaluate the expression, expecting a modification.
   *
   * @param out the writer to the Java source code.
   */
  public void generateDirty(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code for a function arg.
   *
   * @param out the writer to the Java source code.
   */
  public void generateArg(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code for a function arg.
   *
   * @param out the writer to the Java source code.
   */
  public void generateValue(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code for a reference.
   *
   * @param out the writer to the Java source code.
   */
  public void generateRef(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code to recreate the expression.
   *
   * @param out the writer to the Java source code.
   */
  public void generateCopy(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code to recreate the expression, creating an array
   * for an unset value.
   *
   * @param out the writer to the Java source code.
   */
  public void generateArray(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code to recreate the expression, creating an object
   * for an unset value.
   *
   * @param out the writer to the Java source code.
   */
  public void generateObject(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code to evaluate the expression
   *
   * @param out the writer to the Java source code.
   */
  public void generateAssign(PhpWriter out, Expr value,
			     boolean isTop)
    throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName() + " is not a valid left-hand side.");
  }

  /**
   * Generates code to evaluate the expression
   *
   * @param out the writer to the Java source code.
   */
  public void generateAssignRef(PhpWriter out, Expr value, boolean isTop)
    throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName());
  }

  /**
   * Generates code to evaluate the expression
   *
   * @param out the writer to the Java source code.
   */
  public void generateAssignRef(PhpWriter out, String value)
    throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName());
  }
  
  /**
   * Generates code to unset the expression
   *
   * @param out the writer to the Java source code.
   */
  public void generateUnset(PhpWriter out)
    throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName());
  }

  /**
   * Generates code to recreate the expression.
   *
   * @param out the writer to the Java source code.
   */
  public void generateTop(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code for a return value
   *
   * @param out the writer to the Java source code.
   */
  public void generateReturn(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code to recreate the expression.
   *
   * @param out the writer to the Java source code.
   */
  public void generateStatement(PhpWriter out)
    throws IOException
  {
    generateTop(out);
    out.println(";");
  }

  /**
   * Generates code to evaluate a boolean directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateBoolean(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toBoolean()");
  }

  /**
   * Generates code to evaluate a string directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateString(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toString()");
  }

  /**
   * Generates code to evaluate a string directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateStringValue(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toStringValue()");
  }

  /**
   * Generates code to append to a string builder.
   *
   * @param out the writer to the Java source code.
   */
  public void generateAppend(PhpWriter out)
    throws IOException
  {
    generate(out);
  }

  /**
   * Generates code to evaluate a string directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateChar(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toChar()");
  }

  /**
   * Generates code to evaluate the expression directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateInt(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toInt()");
  }
  
  /**
   * Generates code to evaluate the expression directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateLong(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toLong()");
  }

  /**
   * Generates code to evaluate the expression directly
   *
   * @param out the writer to the Java source code.
   */
  public void generateDouble(PhpWriter out)
    throws IOException
  {
    generate(out);
    out.print(".toDouble()");
  }

  /**
   * Generates code to get the out.
   */
  public void generateGetOut(PhpWriter out)
    throws IOException
  {
    // quercus/1l07
    // out.print("_quercus_out");

    // out.print("env.getOut()");
    out.print("env");
  }

  /**
   * Generates code to print the expression directly
   *
   * @param out the writer to the Java source code.
   */
  public void generatePrint(PhpWriter out)
    throws IOException
  {
    if (isLong()) {
      out.print("env.print(");
      generateLong(out);
      out.print(")");
    }
    else if (isDouble()) {
      out.print("env.print(");
      generateDouble(out);
      out.print(")");
    }
    else if (isString()) {
      out.print("env.print(");
      generateString(out);
      out.print(")");
    }
    else {
      generate(out);
      out.print(".print(env)");
    }
  }

  /**
   * Generates code to recreate the expression.
   *
   * @param out the writer to the Java source code.
   */
  public void generateExpr(PhpWriter out)
    throws IOException
  {
    throw new UnsupportedOperationException(getClass().getName());
  }
}

