/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;

import java.io.IOException;
import java.lang.reflect.*;

/**
 * Represents a PHP append ('.') expression.
 */
public class ProAppendExpr extends AppendExpr
  implements CompilingExpr
{
  protected ProAppendExpr(Expr value, ProAppendExpr next)
  {
    super(value, next);
  }

  public GenerateExpr getGenerator()
  {
    return GENERATOR;
  }

  private GenerateExpr GENERATOR = new GenerateExpr(getLocation()) {
      public GenerateExpr getValue()
      {
	return ((CompilingExpr) ProAppendExpr.this.getValue()).getGenerator();
      }
      
      /**
       * Returns true for a string.
       */
      public boolean isString()
      {
	return true;
      }

      /**
       * Analyze the expression
       */
      @Override
      public void analyze(AnalyzeInfo info)
      {
	getValue().analyze(info);

	if (getNext() != null)
	  ((CompilingExpr) getNext()).getGenerator().analyze(info);
      }

      /**
       * Generates code to evaluate the expression as a string.
       *
       * @param out the writer to the Java source code.
       */
      @Override
      public void generate(PhpWriter out)
	throws IOException
      {
	Expr valueExpr = ProAppendExpr.this.getValue();
	AppendExpr ptr = getNext();

	if (valueExpr instanceof StringLiteralExpr) {
	  StringLiteralExpr literal = (StringLiteralExpr) valueExpr;
	  String string = literal.evalConstant().toString();
	  
	  if (! out.getPhp().isUnicodeSemantics())
	    out.print("new StringBuilderValue(");
	  else if (valueExpr instanceof BinaryLiteralExpr)
	    out.print("new BinaryBuilderValue(");
	  else
	    out.print("new UnicodeBuilderValue(");

	  if (! "".equals(string)) {
	    String charArray = out.addCharArray(string);

	    out.print(charArray);
	  }

	  if (ptr != null && ! hasCustomAppend(ptr.getValue())) {
	    CompilingExpr value = (CompilingExpr) ptr.getValue();

	    if (! "".equals(string))
	      out.print(", ");
	    
	    value.getGenerator().generate(out);

	    ptr = ptr.getNext();
	  }

	  out.print(")");
	}
	else {
	  getValue().generate(out);
	  out.print(".toStringBuilder(env)");
	}

	for (; ptr != null; ptr = ptr.getNext()) {
	  CompilingExpr value = (CompilingExpr) ptr.getValue();
	  
	  out.print(".appendUnicode(");
	  value.getGenerator().generateAppend(out);
	  out.print(")");
	}
      }

      /**
       * Generates code to evaluate the expression as a string.
       *
       * @param out the writer to the Java source code.
       */
      @Override
      public void generateStringValue(PhpWriter out)
	throws IOException
      {
	generate(out);
      }

      /**
       * Generates code to evaluate the expression as a string.
       *
       * @param out the writer to the Java source code.
       */
      private void generateSubAppend(PhpWriter out)
	throws IOException
      {
	for (ProAppendExpr ptr = ProAppendExpr.this;
	     ptr != null;
	     ptr = (ProAppendExpr) ptr.getNext()) {
	  out.print(".appendUnicode(");
	  ((CompilingExpr) ptr.getValue()).getGenerator().generateAppend(out);
	  out.print(")");
	}
      }

      /**
       * Generates code to evaluate the expression as a string.
       *
       * @param out the writer to the Java source code.
       */
      public void generateString(PhpWriter out)
	throws IOException
      {
	getValue().generate(out);
	out.print(".toStringBuilder(env)");

	for (ProAppendExpr ptr = (ProAppendExpr) ProAppendExpr.this.getNext();
	     ptr != null;
	     ptr = (ProAppendExpr) ptr.getNext()) {
	  CompilingExpr value = (CompilingExpr) ptr.getValue();
	  
	  out.print(".append(");
	  value.getGenerator().generateAppend(out);
	  out.print(")");
	}

	out.print(".toString()");
      }

      /**
       * Generates code to print the expression to the output
       *
       * @param out the writer to the Java source code.
       */
      public void generatePrint(PhpWriter out)
	throws IOException
      {
	for (AppendExpr ptr = ProAppendExpr.this;
	     ptr != null;
	     ptr = (ProAppendExpr) ptr.getNext()) {
	  ((CompilingExpr) ptr.getValue()).getGenerator().generatePrint(out);
	}
      }
    };

  /**
   * Returns true if the expr class customizes the generateAppend
   * method.
   */
  static boolean hasCustomAppend(Expr expr)
  {
    try {
      GenerateExpr gen = ((CompilingExpr) expr).getGenerator();
      Class cl = gen.getClass();
      
      Method method = cl.getMethod("generateAppend",
				   new Class[] { PhpWriter.class });

      if (method != null
	  && method.getDeclaringClass().equals(GenerateExpr.class))
	return false;
      else
	return true;
    } catch (Exception e) {
      return true;
    }
  }
}
