/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.Location;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;

import java.io.IOException;

/**
 * Represents a PHP array reference expression.
 */
public class ProArrayGetExpr extends ArrayGetExpr
  implements CompilingExpr
{
  public ProArrayGetExpr(Location location, Expr expr, Expr index)
  {
    super(location, expr, index);
  }

  public GenerateExpr getGenerator()
  {
    return GENERATOR;
  }

  private GenerateExpr GENERATOR = new GenerateExpr(getLocation()) {
      public GenerateExpr getExpr()
      {
	return ((CompilingExpr) _expr).getGenerator();
      }
      
      public GenerateExpr getIndex()
      {
	return ((CompilingExpr) _index).getGenerator();
      }
	
      /**
       * Analyze the statement
       */
      public void analyze(AnalyzeInfo info)
      {
	getExpr().analyze(info);
	getIndex().analyze(info);
      }

      /**
       * Analyze the statement
       */
      public void analyzeAssign(AnalyzeInfo info)
      {
	analyze(info);

	getExpr().analyzeSetModified(info);
	// php/3a68
	getExpr().analyzeSetReference(info);
      }

      /**
       * Analyze the statement as modified
       */
      public void analyzeSetModified(AnalyzeInfo info)
      {
	getExpr().analyzeSetModified(info);
      }

      /**
       * Analyze the statement as a reference
       */
      public void analyzeSetReference(AnalyzeInfo info)
      {
	// php/3a69
	getExpr().analyzeSetReference(info);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generate(PhpWriter out)
	throws IOException
      {
	getExpr().generate(out);
	out.print(".get(");
	getIndex().generate(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression, marking as dirty, i.e.
       * copy-on-write for an unset.
       *
       * @param out the writer to the Java source code.
       */
      public void generateDirty(PhpWriter out)
	throws IOException
      {
	getExpr().generate(out);
	out.print(".getDirty(");
	getIndex().generate(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateArray(PhpWriter out)
	throws IOException
      {
	getExpr().generateArray(out);
	out.print(".getArray(");
	getIndex().generate(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateObject(PhpWriter out)
	throws IOException
      {
	getExpr().generateArray(out);
	out.print(".getObject(env, ");
	getIndex().generate(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateRef(PhpWriter out)
	throws IOException
      {
	getExpr().generateArray(out); // php/3d1c
	out.print(".getRef(");
	getIndex().generate(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateArg(PhpWriter out)
	throws IOException
      {
	getExpr().generateArg(out);
	out.print(".getArg(");
	getIndex().generate(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression with a copied result.
       *
       * @param out the writer to the Java source code.
       */
      public void generateCopy(PhpWriter out)
	throws IOException
      {
	generate(out);
	out.print(".copy()"); // php/3a5m
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateAssign(PhpWriter out, Expr value, boolean isTop)
	throws IOException
      {
	GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();
	
	getExpr().generateArray(out);
	out.print(".put(");
	getIndex().generate(out);
	out.print(", ");
	valueGen.generateCopy(out);  // php/3a5k
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateAssignRef(PhpWriter out, Expr value, boolean isTop)
	throws IOException
      {
	GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();
	
	getExpr().generateArray(out);
	out.print(".put(");
	getIndex().generate(out);
	out.print(", ");
	valueGen.generateRef(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateUnset(PhpWriter out)
	throws IOException
      {
        getExpr().generateDirty(out);
	out.print(".remove(");
	getIndex().generate(out);
	out.print(")");
      }
    };
  
}

