/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.Location;
import com.caucho.quercus.env.Value;
import com.caucho.quercus.env.StringValue;
import com.caucho.quercus.program.*;
import com.caucho.quercus.statement.*;
import com.caucho.util.L10N;
import com.caucho.vfs.Path;

import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * Factory for creating PHP expressions and statements
 */
public class ProExprFactory extends ExprFactory {
  private static final L10N L = new L10N(ProExprFactory.class);
  private static final Logger log
    = Logger.getLogger(ProExprFactory.class.getName());

  public ProExprFactory()
  {
  }

  /**
   * Creates a null literal expression.
   */
  public Expr createNull()
  {
    return ProNullLiteralExpr.NULL;
  }

  /**
   * Creates a string literal expression.
   */
  @Override
  public Expr createString(String lexeme)
  {
    return new ProStringLiteralExpr(lexeme);
  }

  /**
   * Creates a binary literal expression.
   */
  @Override
  public Expr createBinary(byte []bytes)
  {
    return new ProBinaryLiteralExpr(bytes);
  }

  /**
   * Creates a binary literal expression.
   */
  @Override
  public Expr createUnicode(String lexeme)
  {
    return new ProUnicodeLiteralExpr(lexeme);
  }

  /**
   * Creates a binary literal expression.
   */
  @Override
  public Expr createBinary(String string, String encoding)
  {
    return new ProBinaryLiteralExpr(string, encoding);
  }

  /**
   * Creates a long literal expression.
   */
  @Override
  public Expr createLong(long value)
  {
    return new ProLongLiteralExpr(value);
  }

  /**
   * Creates a literal expression.
   */
  @Override
  public Expr createLiteral(Value value)
  {
    return new ProLiteralExpr(value);
  }

  /**
   * Creates a var expression.
   */
  @Override
  public VarExpr createVar(VarInfo var)
  {
    return new ProVarExpr(var);
  }

  /**
   * Creates a var expression.
   */
  @Override
  public VarVarExpr createVarVar(Expr var)
  {
    return new ProVarVarExpr(var);
  }

  /**
   * Creates a const expression.
   */
  @Override
  public ConstExpr createConst(String name)
  {
    return new ProConstExpr(name);
  }

  /**
   * Creates a class const expression.
   */
  @Override
  public ClassConstExpr createClassConst(String className, String name)
  {
    return new ProClassConstExpr(className, name);
  }

  /**
   * Creates a this expression.
   */
  @Override
  public ThisExpr createThis(Location location, InterpretedClassDef cl)
  {
    return new ProThisExpr(location, cl);
  }

  /**
   * Creates an array get 'a[0]' expression.
   */
  @Override
  public ArrayGetExpr createArrayGet(Location location,
                                     Expr base,
                                     Expr index)
  {
    return new ProArrayGetExpr(location, base, index);
  }

  /**
   * Creates an array tail 'a[]' expression.
   */
  @Override
  public ArrayTailExpr createArrayTail(Location location, Expr base)
  {
    return new ProArrayTailExpr(location, base);
  }

  /**
   * Creates an object get '$a->b' expression.
   */
  @Override
  public Expr createFieldGet(Location location, Expr base, StringValue name)
  {
    return new ProFieldGetExpr(location, base, name);
  }

  /**
   * Creates an object get '$a->$b' expression.
   */
  @Override
  public Expr createFieldVarGet(Location location, Expr base, Expr name)
  {
    return new ProFieldVarGetExpr(location, base, name);
  }

  /**
   * Creates an object get 'a::b' expression.
   */
  @Override
  public Expr createStaticFieldGet(Location location,
                                   String className,
                                   String name)
  {
    return new ProStaticFieldGetExpr(location, className, name);
  }
  
  /**
   * Creates an object get 'a::${b}' expression.
   */
  public Expr createStaticFieldVarGet(Location location,
                                      String className,
                                      Expr name)
  {
    return new ProStaticFieldVarGetExpr(location, className, name);
  }

  /**
   * Creates a ref '&$a' expression.
   */
  @Override
  public RefExpr createRef(Expr base)
  {
    return new ProRefExpr(base);
  }

  /**
   * Creates an unset '$a' expression.
   */
  @Override
  public Expr createUnsetVar(AbstractVarExpr var)
  {
    return new ProUnsetVarExpr(var);
  }

  /**
   * Creates a char at 'a{0}' expression.
   */
  @Override
  public CharAtExpr createCharAt(Expr base, Expr index)
  {
    return new ProCharAtExpr(base, index);
  }

  /**
   * Creates a post increment 'a++' expression.
   */
  @Override
  public PostIncrementExpr createPostIncrement(Expr expr, int incr)
  {
    return new ProPostIncrementExpr(expr, incr);
  }

  /**
   * Creates a pre increment '++a' expression.
   */
  @Override
  public PreIncrementExpr createPreIncrement(Expr expr, int incr)
  {
    return new ProPreIncrementExpr(expr, incr);
  }

  /**
   * Creates a unary minus '-a' expression.
   */
  @Override
  public Expr createMinus(Expr expr)
  {
    return new ProMinusExpr(expr);
  }

  /**
   * Creates a unary plus '+a' expression.
   */
  @Override
  public Expr createPlus(Expr expr)
  {
    return new ProPlusExpr(expr);
  }

  /**
   * Creates a clone 'clone a' expression.
   */
  @Override
  public Expr createClone(Expr expr)
  {
    return new ProCloneExpr(expr);
  }

  /**
   * Creates a clone 'clone a' expression.
   */
  public Expr createCopy(Expr expr)
  {
    return new ProCopyExpr(expr);
  }

  /**
   * Creates a logical not '!a' expression.
   */
  @Override
  public Expr createNot(Expr expr)
  {
    return new ProNotExpr(expr);
  }

  /**
   * Creates a unary inversion '~a' expression.
   */
  @Override
  public Expr createBitNot(Expr expr)
  {
    return new ProBitNotExpr(expr);
  }

  /**
   * Creates an error suppression '@a' expression.
   */
  @Override
  public Expr createSuppress(Expr expr)
  {
    return new ProSuppressErrorExpr(expr);
  }

  /**
   * Creates a boolean cast
   */
  @Override
  public Expr createToBoolean(Expr expr)
  {
    return new ProToBooleanExpr(expr);
  }

  /**
   * Creates a long cast
   */
  @Override
  public Expr createToLong(Expr expr)
  {
    return new ProToLongExpr(expr);
  }

  /**
   * Creates a double cast
   */
  @Override
  public Expr createToDouble(Expr expr)
  {
    return new ProToDoubleExpr(expr);
  }

  /**
   * Creates a string cast
   */
  @Override
  public Expr createToString(Expr expr)
  {
    return new ProToStringExpr(expr);
  }

  /**
   * Creates a binary string cast
   */
  @Override
  public Expr createToBinary(Expr expr)
  {
    return new ProToBinaryExpr(expr);
  }

  /**
   * Creates a unicode string cast
   */
  @Override
  public Expr createToUnicode(Expr expr)
  {
    return new ProToUnicodeExpr(expr);
  }

  /**
   * Creates an object cast
   */
  @Override
  public Expr createToObject(Expr expr)
  {
    return new ProToObjectExpr(expr);
  }

  /**
   * Creates an array cast
   */
  @Override
  public Expr createToArray(Expr expr)
  {
    return new ProToArrayExpr(expr);
  }

  /**
   * Creates an die 'die("msg")' expression.
   */
  @Override
  public Expr createDie(Expr expr)
  {
    return new ProDieExpr(expr);
  }

  /**
   * Creates an exit 'exit("msg")' expression.
   */
  @Override
  public Expr createExit(Expr expr)
  {
    return new ProExitExpr(expr);
  }

  /**
   * Creates a required
   */
  @Override
  public Expr createRequired()
  {
    return new ProRequiredExpr();
  }

  /**
   * Creates a default
   */
  public Expr createDefault()
  {
    return new ProDefaultExpr();
  }

  /**
   * Creates an addition expression.
   */
  @Override
  public Expr createAdd(Expr left, Expr right)
  {
    return new ProAddExpr(left, right);
  }

  /**
   * Creates a subtraction expression.
   */
  @Override
  public Expr createSub(Expr left, Expr right)
  {
    return new ProSubExpr(left, right);
  }

  /**
   * Creates a multiplication expression.
   */
  @Override
  public Expr createMul(Expr left, Expr right)
  {
    return new ProMulExpr(left, right);
  }

  /**
   * Creates a division expression.
   */
  @Override
  public Expr createDiv(Expr left, Expr right)
  {
    return new ProDivExpr(left, right);
  }

  /**
   * Creates a modulo expression.
   */
  @Override
  public Expr createMod(Expr left, Expr right)
  {
    return new ProModExpr(left, right);
  }

  /**
   * Creates a left-shift expression.
   */
  @Override
  public Expr createLeftShift(Expr left, Expr right)
  {
    return new ProLeftShiftExpr(left, right);
  }

  /**
   * Creates a right-shift expression.
   */
  @Override
  public Expr createRightShift(Expr left, Expr right)
  {
    return new ProRightShiftExpr(left, right);
  }

  /**
   * Creates a bit-and expression.
   */
  @Override
  public Expr createBitAnd(Expr left, Expr right)
  {
    return new ProBitAndExpr(left, right);
  }

  /**
   * Creates a bit-xor expression.
   */
  @Override
  public Expr createBitXor(Expr left, Expr right)
  {
    return new ProBitXorExpr(left, right);
  }

  /**
   * Creates a bit-or expression.
   */
  @Override
  public Expr createBitOr(Expr left, Expr right)
  {
    return new ProBitOrExpr(left, right);
  }

  /**
   * Creates a lt expression.
   */
  @Override
  public Expr createLt(Expr left, Expr right)
  {
    return new ProLtExpr(left, right);
  }

  /**
   * Creates a leq expression.
   */
  @Override
  public Expr createLeq(Expr left, Expr right)
  {
    return new ProLeqExpr(left, right);
  }

  /**
   * Creates a gt expression.
   */
  @Override
  public Expr createGt(Expr left, Expr right)
  {
    return new ProGtExpr(left, right);
  }

  /**
   * Creates a geq expression.
   */
  @Override
  public Expr createGeq(Expr left, Expr right)
  {
    return new ProGeqExpr(left, right);
  }

  /**
   * Creates an eq expression.
   */
  @Override
  public Expr createEq(Expr left, Expr right)
  {
    return new ProEqExpr(left, right);
  }

  /**
   * Creates a neq expression.
   */
  @Override
  public Expr createNeq(Expr left, Expr right)
  {
    return new ProNeqExpr(left, right);
  }

  /**
   * Creates an equals expression.
   */
  @Override
  public Expr createEquals(Expr left, Expr right)
  {
    return new ProEqualsExpr(left, right);
  }
  
  /**
   * Creates an append expression.
   */
  @Override
  protected AppendExpr createAppendImpl(Expr left, AppendExpr right)
  {
    return new ProAppendExpr(left, (ProAppendExpr) right);
  }

  /**
   * Creates an assignment expression.
   */
  @Override
  public Expr createAssign(AbstractVarExpr var, Expr right)
  {
    return new ProAssignExpr(var, right);
  }

  /**
   * Creates an assignment ref expression.
   */
  @Override
  public Expr createAssignRef(AbstractVarExpr var, Expr right)
  {
    return new ProAssignRefExpr(var, right);
  }

  /**
   * Creates an and expression.
   */
  @Override
  public Expr createAnd(Expr left, Expr right)
  {
    return new ProAndExpr(left, right);
  }

  /**
   * Creates an or expression.
   */
  @Override
  public Expr createOr(Expr left, Expr right)
  {
    return new ProOrExpr(left, right);
  }

  /**
   * Creates an xor expression.
   */
  @Override
  public Expr createXor(Expr left, Expr right)
  {
    return new ProXorExpr(left, right);
  }

  /**
   * Creates a comma expression.
   */
  @Override
  public Expr createComma(Expr left, Expr right)
  {
    return new ProCommaExpr(left, right);
  }

  /**
   * Creates an instanceof expression.
   */
  public Expr createInstanceOf(Expr expr, String name)
  {
    return new ProInstanceOfExpr(expr, name);
  }

  /**
   * Creates an instanceof expression.
   */
  public Expr createInstanceOfVar(Expr expr, Expr name)
  {
    return new ProInstanceOfVarExpr(expr, name);
  }

  /**
   * Creates an each expression.
   */
  @Override
  public Expr createEach(Expr expr)
  {
    return new ProEachExpr(expr);
  }

  /**
   * Creates a list expression.
   */
  @Override
  public ListHeadExpr createListHead(ArrayList<Expr> keys)
  {
    return new ProListHeadExpr(keys);
  }

  /**
   * Creates a list expression.
   */
  @Override
  public Expr createList(ListHeadExpr head, Expr value)
  {
    return new ProListExpr(head, value);
  }

  /**
   * Creates a list expression.
   */
  @Override
  public Expr createListEach(Expr []varList, EachExpr value)
  {
    return new ProListEachExpr(varList, value);
  }

  /**
   * Creates an conditional expression.
   */
  @Override
  public Expr createConditional(Expr test, Expr left, Expr right)
  {
    return new ProConditionalExpr(test, left, right);
  }

  /**
   * Creates a array() expression.
   */
  @Override
  public Expr createArrayFun(ArrayList<Expr> keys, ArrayList<Expr> values)
  {
    return new ProArrayFunExpr(keys, values);
  }

  /**
   * Creates a new function call.
   */
  @Override
    public Expr createFunction(Location loc,
			       String name,
			       ArrayList<Expr> args)
  {
    if ("isset".equals(name) && args.size() == 1)
      return new ProIssetExpr(args.get(0));
    else
      return new ProFunctionExpr(loc, name, args);
  }

  /**
   * Creates a new var function call.
   */
  @Override
  public VarFunctionExpr createVarFunction(Location loc,
					   Expr name,
					   ArrayList<Expr> args)
  {
    return new ProVarFunctionExpr(loc, name, args);
  }

  /**
   * Creates a new method call.
   */
  @Override
  public Expr createVarMethodCall(Location loc,
				  Expr objExpr,
				  Expr name,
				  ArrayList<Expr> args)
  {
    return new ProVarMethodCallExpr(loc, objExpr, name, args);
  }

  /**
   * Creates a new call.
   */
  @Override
  public NewExpr createNew(Location loc,
			   String name,
			   ArrayList<Expr> args)
  {
    return new ProNewExpr(loc, name, args);
  }

  /**
   * Creates a new function call.
   */
  @Override
  public Expr createClassMethod(Location loc,
				String className,
				String name,
				ArrayList<Expr> args)
  {
    return new ProClassMethodExpr(loc, className, name, args);
  }

  /**
   * Creates a new function call.
   */
  @Override
  public Expr createStaticMethod(Location loc,
				 String className,
				 String name,
				 ArrayList<Expr> args)
  {
    return new ProStaticMethodExpr(loc, className, name, args);
  }

  /**
   * Creates a new function call.
   */
  @Override
  public Expr createStaticVarMethod(Location loc,
				    String className,
				    Expr name,
				 ArrayList<Expr> args)
  {
    return new ProStaticVarMethodExpr(loc, className, name, args);
  }

  /**
   * Creates a new method call.
   */
  @Override
  public Expr createMethodCall(Location loc,
				       Expr objExpr,
				       String name,
				       ArrayList<Expr> args)
  {
    return new ProMethodCallExpr(loc, objExpr, name, args);
  }

  /**
   * Creates a new function call.
   */
  @Override
  public VarNewExpr createVarNew(Location loc,
				 Expr name,
				 ArrayList<Expr> args)
  {
    return new ProVarNewExpr(loc, name, args);
  }

  /**
   * Creates an include expr
   */
  public Expr createInclude(Location loc,
			    Path source,
			    Expr expr)
  {
    return new ProIncludeExpr(loc, source, expr, false);
  }

  /**
   * Creates an include expr
   */
  public Expr createRequire(Location loc,
			    Path source,
			    Expr expr)
  {
    return new ProIncludeExpr(loc, source, expr, true);
  }

  /**
   * Creates an include expr
   */
  public Expr createIncludeOnce(Location loc,
				Path source,
				Expr expr)
  {
    return new ProIncludeOnceExpr(loc, source, expr, false);
  }

  /**
   * Creates an include expr
   */
  public Expr createRequireOnce(Location loc,
			    Path source,
			    Expr expr)
  {
    return new ProIncludeOnceExpr(loc, source, expr, true);
  }

  /**
   * Creates a Quercus class import.
   */
  public Expr createImport(Location loc,
                           String name,
                           boolean isWildcard)
  {
    return new ProImportExpr(loc, name, isWildcard);
  }

  /**
   * Creates an echo statement
   */
  @Override
  public Statement createEcho(Location loc, Expr expr)
  {
    return new ProEchoStatement(loc, expr);
  }

  /**
   * Creates an expr statement
   */
  @Override
  public Statement createExpr(Location loc, Expr expr)
  {
    return new ProExprStatement(loc, expr);
  }

  /**
   * Creates an expr statement
   */
  @Override
  public BlockStatement createBlockImpl(Location loc, Statement []statements)
  {
    return new ProBlockStatement(loc, statements);
  }

  /**
   * Creates a text statement
   */
  @Override
  public Statement createText(Location loc, String text)
  {
    return new ProTextStatement(loc, text);
  }

  /**
   * Creates a null statement
   */
  @Override
  public Statement createNullStatement()
  {
    return ProNullStatement.PRO_NULL;
  }

  /**
   * Creates an if statement
   */
  @Override
  public Statement createIf(Location loc,
			    Expr test,
			    Statement trueBlock,
			    Statement falseBlock)
  {
    return new ProIfStatement(loc, test, trueBlock, falseBlock);
  }

  /**
   * Creates a switch statement
   */
  @Override
  public Statement createSwitch(Location loc,
				Expr value,
				ArrayList<Expr[]> caseList,
				ArrayList<BlockStatement> blockList,
				Statement defaultBlock)
  {
    return new ProSwitchStatement(loc, value, caseList,
				  blockList, defaultBlock);
  }

  /**
   * Creates a for statement
   */
  @Override
  public Statement createFor(Location loc,
			     Expr init,
			     Expr test,
			     Expr incr,
			     Statement block)
  {
    return new ProForStatement(loc, init, test, incr, block);
  }

  /**
   * Creates a foreach statement
   */
  @Override
  public Statement createForeach(Location loc,
				 Expr objExpr,
				 AbstractVarExpr key,
				 AbstractVarExpr value,
				 boolean isRef,
				 Statement block)
  {
    return new ProForeachStatement(loc, objExpr, key, value, isRef, block);
  }

  /**
   * Creates a while statement
   */
  @Override
  public Statement createWhile(Location loc,
			       Expr test,
			       Statement block)
  {
    return new ProWhileStatement(loc, test, block);
  }

  /**
   * Creates a do-while statement
   */
  @Override
  public Statement createDo(Location loc,
			    Expr test,
			    Statement block)
  {
    return new ProDoStatement(loc, test, block);
  }

  /**
   * Creates a break statement
   */
  @Override
  public BreakStatement createBreak(Location location, Expr target)
  {
    return new ProBreakStatement(location, target);
  }

  /**
   * Creates a continue statement
   */
  @Override
  public ContinueStatement createContinue(Location location, Expr target)
  {
    return new ProContinueStatement(location, target);
  }

  /**
   * Creates a global statement
   */
  @Override
  public Statement createGlobal(Location loc,
				VarExpr var)
  {
    return new ProGlobalStatement(loc, var);
  }

  /**
   * Creates a global var statement
   */
  @Override
  public Statement createVarGlobal(Location loc,
                                   VarVarExpr var)
  {
    return new ProVarGlobalStatement(loc, var);
  }

  /**
   * Creates a static statement
   */
  @Override
  public Statement createStatic(Location loc,
				VarExpr var,
				Expr value)
  {
    return new ProStaticStatement(loc, var, value);
  }

  /**
   * Creates a throw statement
   */
  @Override
  public Statement createThrow(Location loc,
			       Expr value)
  {
    return new ProThrowStatement(loc, value);
  }

  /**
   * Creates a try statement
   */
  @Override
  public TryStatement createTry(Location loc,
				Statement block)
  {
    return new ProTryStatement(loc, block);
  }

  /**
   * Creates a return statement
   */
  @Override
  public Statement createReturn(Location loc,
				Expr value)
  {
    return new ProReturnStatement(loc, value);
  }

  /**
   * Creates a return ref statement
   */
  @Override
  public Statement createReturnRef(Location loc,
				   Expr value)
  {
    return new ProReturnRefStatement(loc, value);
  }

  /**
   * Creates a new function definition def.
   */
  public Statement createFunctionDef(Location loc,
				     Function fun)
  {
    return new ProFunctionDefStatement(loc, fun);
  }

  /**
   * Creates a new function def statement
   */
  @Override
  public Statement createClassDef(Location loc,
				  InterpretedClassDef cl)
  {
    return new ProClassDefStatement(loc, cl);
  }

  /**
   * Creates a new function call.
   */
  @Override
  public Function createFunction(Location loc,
                                 String name,
                                 FunctionInfo info,
                                 Arg []argList,
                                 Statement []statementList)
  {
    return new ProFunction(this, loc, name, info, argList, statementList);
  }

  /**
   * Creates a new object method definition.
   */
  @Override
  public Function createObjectMethod(Location loc,
					 InterpretedClassDef cl,
					 String name,
					 FunctionInfo info,
					 Arg []argList,
					 Statement []statementList)
  {
    return new ProObjectMethod(this, loc, cl, name, info,
                               argList, statementList);
  }

  /**
   * Creates a new object method definition.
   */
  @Override
  public Function createMethodDeclaration(Location loc,
                                          InterpretedClassDef cl,
                                          String name,
                                          FunctionInfo info,
                                          Arg []argList)
  {
    return new ProMethodDeclaration(this, loc, cl, name, info, argList);
  }

  @Override
  public InterpretedClassDef createClassDef(Location location, String name,
                                            String parentName,
                                            String[] ifaceList)
  {
    return new ProClassDef(location, name, parentName, ifaceList);
  }
}

