/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.Location;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.quercus.env.StringValue;

import java.io.IOException;

/**
 * Represents a PHP field reference.
 */
public class ProFieldGetExpr extends FieldGetExpr
  implements CompilingExpr
{
  public ProFieldGetExpr(Location location, Expr objExpr, StringValue name)
  {
    super(objExpr, name);
  }

  public GenerateExpr getGenerator()
  {
    return GENERATOR;
  }

  private GenerateExpr GENERATOR = new GenerateExpr(getLocation()) {
      /**
       * Analyze the statement
       */
      public void analyze(AnalyzeInfo info)
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.analyze(info);
      }

      /**
       * Analyze the statement as an assignement
       */
      public void analyzeAssign(AnalyzeInfo info)
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.analyze(info);

	// php/3a6e
	objGen.analyzeSetReference(info);
	objGen.analyzeSetModified(info);
      }

      /**
       * Analyze the statement as modified
       */
      public void analyzeSetModified(AnalyzeInfo info)
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.analyzeSetModified(info);
      }

      /**
       * Analyze the statement as a reference
       */
      public void analyzeSetReference(AnalyzeInfo info)
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	// php/3a6f
	objGen.analyzeSetReference(info);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generate(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();

	objGen.generate(out);
	out.print(".getField(env, ");
	out.print(_name);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateArg(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.generateArg(out);
	out.print(".getFieldArg(env, ");
	out.print(_name);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateDirty(PhpWriter out)
        throws IOException
      {
        GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();

        objGen.generateObject(out);
        out.print(".getField(env, ");
        out.print(_name);
        out.print(")"); // php/3228 create the field if necessary
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateRef(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.generateObject(out);
	out.print(".getFieldRef(env, ");
	out.print(_name);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression, where the result is copied.
       *
       * @param out the writer to the Java source code.
       */
      public void generateCopy(PhpWriter out)
	throws IOException
      {
	generate(out);
	out.print(".copy()");
      }

      /**
       * Generates code to assign the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateAssign(PhpWriter out, Expr value, boolean isTop)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();
	
	objGen.generateObject(out);
	out.print(".putField(env, ");
	out.print(_name);
	out.print(", ");
	valueGen.generateCopy(out);
	out.print(")");
      }

      /**
       * Generates code to assign the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateAssignRef(PhpWriter out, Expr value, boolean isTop)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();
	
	objGen.generateObject(out);
	out.print(".putField(env, ");
	out.print(_name);
	out.print(", ");
	valueGen.generateRef(out);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateObject(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.generateObject(out);
	out.print(".getFieldObject(env, ");
	out.print(_name);
	out.print(")");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateArray(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.generateObject(out);
	out.print(".getFieldArray(env, ");
	out.print(_name);
	out.print(")");
      }

      /**
       * Generates code to assign the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateUnset(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	objGen.generate(out);
	out.print(".unsetField(");
	out.print(_name);
	out.print(")");
      }

      /**
       * Generates code to recreate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateExpr(PhpWriter out)
	throws IOException
      {
	GenerateExpr objGen = ((CompilingExpr) _objExpr).getGenerator();
	
	out.print("new FieldGetExpr(");

	objGen.generateExpr(out);

	out.print(", \"");

	out.printJavaString(_name.toString());

	out.print("\")");
      }
    };
}

