/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * This file is part of Resin(R) Open Source
 *
 * Each copy or derived work must preserve the copyright notice and this
 * notice unmodified.
 *
 * Resin Open Source is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Resin Open Source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, or any warranty
 * of NON-INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Resin Open Source; if not, write to the
 *
 *   Free Software Foundation, Inc.
 *   59 Temple Place, Suite 330
 *   Boston, MA 02111-1307  USA
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.Location;
import com.caucho.quercus.Quercus;
import com.caucho.quercus.env.Value;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.quercus.program.AbstractFunction;
import com.caucho.quercus.program.CompilingFunction;
import com.caucho.quercus.program.FunctionGenerator;
import com.caucho.quercus.program.QuercusProgram;
import com.caucho.util.L10N;

import java.io.IOException;
import java.util.ArrayList;

/**
 * Represents a PHP function expression.
 */
public class ProFunctionExpr extends FunctionExpr
  implements CompilingExpr
{
  private static final L10N L = new L10N(ProFunctionExpr.class);

  public ProFunctionExpr(Location location, String name, ArrayList<Expr> args)
  {
    super(location, name, args);
  }

  public ProFunctionExpr(Location location, String name, Expr []args)
  {
    super(location, name, args);

  }

  /**
   * Returns the reference of the value.
   * @param location
   */
  public Expr createRef(Location location)
  {
    return new RefExpr(location, this);
  }

  public GenerateExpr getGenerator()
  {
    return GENERATOR;
  }

  private GenerateExpr GENERATOR = new GenerateExpr(getLocation()) {
    
      //
      // Java code generation
      //
  
      /**
       * Analyzes the function.
       */
      public void analyze(AnalyzeInfo info)
      {
	Quercus quercus = info.getFunction().getPhp();
	AbstractFunction fun = quercus.findFunction(_name);

	FunctionGenerator funGen = null;

	if (fun != null && fun instanceof CompilingFunction)
	  funGen = ((CompilingFunction) fun).getGenerator();

	if (funGen != null && funGen.isCallUsesVariableArgs())
	  info.getFunction().setVariableArgs(true);
    
	if (funGen != null && funGen.isCallUsesSymbolTable()) {
	  // php/1729
	  info.getFunction().setUsesSymbolTable(true);
	  info.clear();
	}
    
	for (int i = 0; i < _args.length; i++) {
	  CompilingExpr arg = (CompilingExpr) _args[i];
	  
	  arg.getGenerator().analyze(info);
	}

	// check for read-only and refs

	if (funGen != null) {
	  funGen.analyzeArguments(_args, info);
	}
	else {
	  for (int i = 0; i < _args.length; i++) {
	    CompilingExpr arg = (CompilingExpr) _args[i];
	  
	    arg.getGenerator().analyzeSetModified(info);
	    arg.getGenerator().analyzeSetReference(info);
	  }
	}
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generate(PhpWriter out)
	throws IOException
      {
	generateImpl(out, Value.class, false, false);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateCopy(PhpWriter out)
	throws IOException
      {
	generateImpl(out, Value.class, false, true);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateRef(PhpWriter out)
	throws IOException
      {
	generateImpl(out, Value.class, true, false);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateTop(PhpWriter out)
	throws IOException
      {
	generateImpl(out, void.class, true, false);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateBoolean(PhpWriter out)
	throws IOException
      {
	generateImpl(out, boolean.class, false, false);
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateValue(PhpWriter out)
	throws IOException
      {
	generate(out);  // php/3c4k
	out.print(".toValue()");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateImpl(PhpWriter out, Class retType,
			       boolean isRef, boolean isCopy)
	throws IOException
      {
	// Quercus quercus = out.getPhp();

	// StaticFunction fun = quercus.findFunction(_name);

	QuercusProgram program = out.getProgram();

	AbstractFunction fun = program.findFunction(_name);
    
	if (fun == null || ! fun.isGlobal())
	  fun = program.getPhp().findFunction(_name);

	FunctionGenerator funGen = null;

	if (fun != null && fun instanceof CompilingFunction)
	  funGen = ((CompilingFunction) fun).getGenerator();

	if (funGen != null
	    && funGen.canGenerateCall(_args)
	    && ! funGen.isVoidReturn()) {
	  if (void.class.equals(retType))
	    funGen.generateTop(out, this, _args);
	  else if (boolean.class.equals(retType))
	    funGen.generateBoolean(out, this, _args);
	  else if (isRef && fun.isReturnsReference()) // php/3442
	    funGen.generateRef(out, this, _args);
	  else if (isCopy || isRef) // php/3442
	    funGen.generateCopy(out, this, _args);
	  else
	    funGen.generate(out, this, _args);
	}
	else {
	  if (void.class.equals(retType))
	    isRef = true;
	  else if (boolean.class.equals(retType))
	    isRef = true;
	  else if (! isCopy)
	    isRef = true;
      
	  // super.generate(out);

	  // XXX: need to check where it's from

	  String id = out.addFunctionId(_name);

	  out.print("env._fun[" + id + "]");

	  if (isRef)
	    out.print(".callRef(env");
	  else
	    out.print(".call(env");

	  if (_args.length <= COMPILE_ARG_MAX) {
	    for (int i = 0; i < _args.length; i++) {
	      CompilingExpr arg = (CompilingExpr) _args[i];
	      
	      out.print(", ");

	      arg.getGenerator().generateArg(out);
	    }
	  }
	  else {
	    out.print(", new Value[] {");

	    for (int i = 0; i < _args.length; i++) {
	      CompilingExpr arg = (CompilingExpr) _args[i];
	      
	      if (i != 0)
		out.print(", ");

	      arg.getGenerator().generateArg(out);
	    }

	    out.print("}");
	  }
      
	  out.print(")");
    
	  if (boolean.class.equals(retType))
	    out.print(".toBoolean()");
	  /*
	    else if (isCopy)
	    out.print(".copyReturn()");
	  */
	}
      }

      /**
       * Generates code to recreate the expression.  Used for default values.
       *
       * @param out the writer to the Java source code.
       */
      public void generateExpr(PhpWriter out)
	throws IOException
      {
	out.print("new FunctionExpr(\"");
	out.printJavaString(_name);
	out.print("\", new Expr[] {");
    
	for (int i = 0; i < _args.length; i++) {
	  CompilingExpr arg = (CompilingExpr) _args[i];
	  
	  arg.getGenerator().generateExpr(out);
	  out.print(", ");
	}

	out.print("})");
      }
    };
}

