/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.expr;

import com.caucho.quercus.Location;
import com.caucho.quercus.env.StringValue;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.quercus.program.InterpretedClassDef;

import java.io.IOException;

/**
 * Represents a PHP field reference.
 */
public class ProThisFieldExpr extends ThisFieldExpr
  implements CompilingExpr
{
  public ProThisFieldExpr(Location location,
                          InterpretedClassDef quercusClass,
                          StringValue name)
  {
    super(location, quercusClass, name);
  }

  public GenerateExpr getGenerator()
  {
    return GENERATOR;
  }

  private GenerateExpr GENERATOR = new GenerateExpr(getLocation()) {
      private boolean _hasThis;
    
      /**
       * Analyze the statement
       */
      public void analyze(AnalyzeInfo info)
      {
        _hasThis = info.getFunction().hasThis();
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generate(PhpWriter out)
        throws IOException
      {
        if (false && _quercusClass.isDeclaredField(_name)) {
          out.print("q_this._fields[_f_" + _name + "].toValue()");
        }
        else if (_hasThis) {
          out.print("q_this.getThisField(env, ");
          out.print(_name);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateRef(PhpWriter out)
        throws IOException
      {
        if (_hasThis) {
          out.print("q_this.getThisFieldRef(env, ");
          out.print(_name);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }

      }

      /**
       * Generates code to evaluate the expression, as a copy.
       *
       * @param out the writer to the Java source code.
       */
      public void generateCopy(PhpWriter out)
        throws IOException
      {
        generate(out);
        out.print(".copy()");
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateArg(PhpWriter out)
        throws IOException
      {
        if (_hasThis) {
          out.print("q_this.getThisFieldArg(env, ");
          out.print(_name);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }

      /**
       * Generates code to evaluate the expression, creating an array for the
       * field value if unset.
       *
       * @param out the writer to the Java source code.
       */
      public void generateArray(PhpWriter out)
        throws IOException
      {
        if (_hasThis) {
          out.print("q_this.getThisFieldArray(env, ");
          out.print(_name);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }

      /**
       * Generates code to evaluate the expression, creating an object for the
       * field value if unset.
       *
       * @param out the writer to the Java source code.
       */
      public void generateObject(PhpWriter out)
        throws IOException
      {
        if (_hasThis) {
          out.print("q_this.getThisFieldObject(env, ");
          out.print(_name);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateAssign(PhpWriter out, Expr value, boolean isTop)
        throws IOException
      {
        GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();
        
        if (false && _quercusClass.isDeclaredField(_name)) {
          if (! isTop)
            out.print("(");
          
          out.print("q_this._fields[_f_" + _name + "] = ");
          out.print("q_this._fields[_f_" + _name + "].set(");

          valueGen.generateCopy(out);

          out.print(")");
          
          if (! isTop)
            out.print(")");
        }
        else if (_hasThis) {
          out.print("q_this.putThisField(env, ");
          out.print(_name);
          out.print(", ");
          valueGen.generateCopy(out);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateAssignRef(PhpWriter out, Expr value, boolean isTop)
        throws IOException
      {
        GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();
        
        if (false && _quercusClass.isDeclaredField(_name)) {
          // php/39f5
          if (! isTop)
            out.print("(");
          
          out.print("q_this._fields[_f_" + _name + "] = ");

          valueGen.generateRef(out);
          
          if (! isTop)
            out.print(")");
        }
        else if (_hasThis) {
          out.print("q_this.putThisField(env, ");
          out.print(_name);
          out.print(", ");
          valueGen.generateRef(out);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }

      /**
       * Generates code to evaluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateUnset(PhpWriter out)
        throws IOException
      {
        if (_hasThis) {
          out.print("q_this.unsetThisField(");
          out.print(_name);
          out.print(")");
        }
        else {
          out.print("env.error(\"Cannot use '$this' when not in object context.\")");
        }
      }
    };
}

