/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.function;

import com.caucho.quercus.env.*;
import com.caucho.quercus.module.ModuleContext;
import com.caucho.quercus.lib.file.BinaryInput;
import com.caucho.quercus.program.JavaClassDef;
import com.caucho.util.L10N;
import com.caucho.vfs.Path;

import java.io.InputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URL;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * Code for marshaling (PHP to Java) and unmarshaling (Java to PHP) arguments.
 */
public class ProMarshalFactory extends MarshalFactory {
  private static final L10N L = new L10N(ProMarshalFactory.class);

  public ProMarshalFactory(ModuleContext moduleContext)
  {
    super(moduleContext);
  }

  public Marshal create(Class argType,
			boolean isNotNull,
			boolean isNullAsFalse)
  {
    final Marshal marshal;

    // optimized cases, new types should be added to JavaMarshall

    if (String.class.equals(argType)) {
      marshal = ProStringMarshal.MARSHAL;
    }
    else if (boolean.class.equals(argType)) {
      marshal = ProBooleanMarshal.MARSHAL;
    }
    else if (Boolean.class.equals(argType)) {
      marshal = ProBooleanObjectMarshal.MARSHAL;
    }
    else if (byte.class.equals(argType)) {
      marshal = ProByteMarshal.MARSHAL;
    }
    else if (Byte.class.equals(argType)) {
      marshal = ProByteObjectMarshal.MARSHAL;
    }
    else if (short.class.equals(argType)) {
      marshal = ProShortMarshal.MARSHAL;
    }
    else if (Short.class.equals(argType)) {
      marshal = ProShortObjectMarshal.MARSHAL;
    }
    else if (int.class.equals(argType)) {
      marshal = ProIntegerMarshal.MARSHAL;
    }
    else if (Integer.class.equals(argType)) {
      marshal = ProIntegerObjectMarshal.MARSHAL;
    }
    else if (long.class.equals(argType)) {
      marshal = ProLongMarshal.MARSHAL;
    }
    else if (Long.class.equals(argType)) {
      marshal = ProLongObjectMarshal.MARSHAL;
    }
    else if (LongValue.class.equals(argType)) {
      marshal = ProLongValueMarshal.MARSHAL;
    }
    else if (float.class.equals(argType)) {
      marshal = ProFloatMarshal.MARSHAL;
    }
    else if (Float.class.equals(argType)) {
      marshal = ProFloatObjectMarshal.MARSHAL;
    }
    else if (double.class.equals(argType)) {
      marshal = ProDoubleMarshal.MARSHAL;
    }
    else if (Double.class.equals(argType)) {
      marshal = ProDoubleObjectMarshal.MARSHAL;
    }
    else if (DoubleValue.class.equals(argType)) {
      marshal = ProDoubleValueMarshal.MARSHAL;
    }
    else if (BigDecimal.class.equals(argType)) {
      marshal = ProBigDecimalMarshal.MARSHAL;
    }
    else if (BigInteger.class.equals(argType)) {
      marshal = ProBigIntegerMarshal.MARSHAL;
    }
    else if (char.class.equals(argType)) {
      marshal = ProCharacterMarshal.MARSHAL;
    }
    else if (Character.class.equals(argType)) {
      marshal = ProCharacterObjectMarshal.MARSHAL;
    }
    else if (Path.class.equals(argType)) {
      marshal = ProPathMarshal.MARSHAL;
    }
    else if (Callback.class.equals(argType)) {
      marshal = ProCallbackMarshal.MARSHAL;
    }
    else if (StringValue.class.equals(argType)) {
      marshal = ProStringValueMarshal.MARSHAL;
    }
    else if (UnicodeValue.class.equals(argType)) {
      marshal = ProUnicodeValueMarshal.MARSHAL;
    }
    else if (BinaryValue.class.equals(argType)) {
      marshal = ProBinaryValueMarshal.MARSHAL;
    }
    else if (BytesValue.class.equals(argType)) {
      marshal = ProBinaryValueMarshal.MARSHAL;
    }
    else if (InputStream.class.equals(argType)) {
      marshal = ProInputStreamMarshal.MARSHAL;
    }
    else if (BinaryInput.class.equals(argType)) {
      marshal = ProBinaryInputMarshal.MARSHAL;
    }
    else if (ArrayValue.class.equals(argType)) {
      marshal = ProArrayValueMarshal.MARSHAL;
    }
    else if (Value.class.equals(argType)) {
      marshal = ProValueMarshal.MARSHAL;
    }
    else if (Value.class.isAssignableFrom(argType)) {
      marshal = new ProExtValueMarshal(argType);
    }
    else if (void.class.equals(argType)) {
      marshal = ProVoidMarshal.MARSHAL;
    }
    else if (Calendar.class.equals(argType)){
      marshal = ProCalendarMarshal.MARSHAL;
    }
    else if (Date.class.equals(argType)) {
      marshal = ProDateMarshal.MARSHAL;
    }
    else if (URL.class.equals(argType)) {
      marshal = ProURLMarshal.MARSHAL;
    }
    else if (byte[].class.equals(argType)) {
      marshal = ProJavaByteArrayMarshal.MARSHAL;
    }
    else if (Byte[].class.equals(argType)) {
      marshal = ProJavaByteObjectArrayMarshal.MARSHAL;
    }
    else if (char[].class.equals(argType)) {
      marshal = ProJavaCharacterArrayMarshal.MARSHAL;
    }
    else if (Character[].class.equals(argType)) {
      marshal = ProJavaCharacterObjectArrayMarshal.MARSHAL;
    }
    else if (argType.isArray()) {
      marshal = ProJavaArrayMarshal.MARSHAL;
    }
    else if (Map.class.isAssignableFrom(argType)) {
      JavaClassDef javaDef
        = _moduleContext.getJavaClassDefinition(argType, argType.getName());

      marshal = new ProJavaMapMarshal(javaDef, isNotNull, isNullAsFalse);
    }
    else if (List.class.isAssignableFrom(argType)) {
      JavaClassDef javaDef
        = _moduleContext.getJavaClassDefinition(argType, argType.getName());

      marshal = new ProJavaListMarshal(javaDef, isNotNull, isNullAsFalse);
    }
    else if (Collection.class.isAssignableFrom(argType)) {
      JavaClassDef javaDef
        = _moduleContext.getJavaClassDefinition(argType, argType.getName());

      marshal = new ProJavaCollectionMarshal(javaDef, isNotNull, isNullAsFalse);
    }
    else {
      JavaClassDef javaDef
        = _moduleContext.getJavaClassDefinition(argType, argType.getName());

      marshal = new ProJavaMarshal(javaDef, isNotNull, isNullAsFalse);
    }

    if (!isNullAsFalse)
      return marshal;
    else {
      if (Value.class.equals(argType)
	  || Boolean.class.equals(argType)
	  || Byte.class.equals(argType)
	  || Short.class.equals(argType)
	  || Integer.class.equals(argType)
	  || Long.class.equals(argType)
	  || Float.class.equals(argType)
	  || Double.class.equals(argType)
	  || Character.class.equals(argType)) {

        String shortName = argType.getSimpleName();
        throw new UnsupportedOperationException("@ReturnNullAsFalse cannot be used with return type `"+shortName+"'");
      }

      return new ProNullAsFalseMarshal(marshal);
    }
  }

  public Marshal createReference()
  {
    return ProReferenceMarshal.MARSHAL;
  }
}

