/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.gen;

import com.caucho.java.JavaWriter;
import com.caucho.java.gen.JavaWriterWrapper;
import com.caucho.quercus.Quercus;
import com.caucho.quercus.env.Value;
import com.caucho.quercus.expr.Expr;
import com.caucho.quercus.module.QuercusModule;
import com.caucho.quercus.program.InterpretedClassDef;
import com.caucho.quercus.program.QuercusProgram;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
  
/**
 * Writer which gathers additional info.
 */
public class PhpWriter extends JavaWriterWrapper {
  public static final String SWITCH_LABEL_PREFIX = "switch_";
  public static final String FOREACH_LABEL_PREFIX = "foreach_";
  
  protected QuercusProgram _program;
  
  protected HashMap<Value,String> _valueMap
    = new HashMap<Value,String>();
  
  protected HashMap<Expr,String> _exprMap
    = new HashMap<Expr,String>();
  
  protected HashMap<Expr[],String> _exprArrayMap
    = new HashMap<Expr[],String>();
  
  protected HashMap<QuercusModule,String> _moduleMap
    = new HashMap<QuercusModule,String>();

  protected ArrayList<InterpretedClassDef> _classList
    = new ArrayList<InterpretedClassDef>();

  protected HashMap<String,String> _charArrayMap
    = new HashMap<String,String>();

  protected HashMap<String,String> _functionIdMap
    = new HashMap<String,String>();

  protected HashMap<String,String> _classIdMap
    = new HashMap<String,String>();

  protected ArrayList<String> _staticVarList = new ArrayList<String>();

  private ArrayList<String> _loopLabelList
    = new ArrayList<String>();
  
  public PhpWriter(JavaWriter writer, QuercusProgram program)
  {
    super(writer);

    _program = program;
  }
  
  /**
   * Returns the engine.
   */
  public Quercus getPhp()
  {
    return _program.getPhp();
  }
  
  /**
   * Returns the program
   */
  public QuercusProgram getProgram()
  {
    return _program;
  }
  
  /*
   * Returns the number of labels on the stack.
   */
  public int getLoopLabelSize()
  {
    return _loopLabelList.size();
  }

  /*
   * Returns the prefix used for switch statement labels.
   */
  public static String getSwitchLabelPrefix()
  {
    return SWITCH_LABEL_PREFIX;
  }
  
  /*
   * Returns the prefix used for switch statement labels.
   */
  public static String getForeachLabelPrefix()
  {
    return FOREACH_LABEL_PREFIX;
  }
  
  /*
   * Creates a loop label.
   */
  public String createWhileLabel()
  {
    return "while_" + generateId();
  }
  
  /*
   * Creates a loop label.
   */
  public String createForLabel()
  {
    return "for_" + generateId();
  }
  
  /*
   * Creates a loop label.
   */
  public String createForeachLabel()
  {
    return FOREACH_LABEL_PREFIX + generateId();
  }
  
  /*
   * Creates a loop label.
   */
  public String createDoLabel()
  {
    return "do_" + generateId();
  }
  
  /*
   * Creates a loop label.
   */
  public String createSwitchLabel()
  {
    return SWITCH_LABEL_PREFIX + generateId();
  }
  
  /*
   * Returns true if this is a switch label.
   */
  public static boolean isSwitchLabel(String label)
  {
    return label != null && label.startsWith(SWITCH_LABEL_PREFIX);
  }
  
  /*
   * Returns the label at specified index.
   * @param i index of label in array
   */
  public String getLoopLabel(int i)
  {
    int size = _loopLabelList.size();
    
    if (0 <= i && i < size)
      return _loopLabelList.get(i);
    else
      return null;
  }
  
  /*
   * Pushes the label onto the stack.
   * @param label to push onto the stack
   */
  public void pushLoopLabel(String label)
  {
    _loopLabelList.add(label);
  }
  
  /*
   * Pops a label from the stack;
   */
  public String popLoopLabel()
  {
    int size = _loopLabelList.size();
    
    if (size == 0)
      return null;
    else
      return _loopLabelList.remove(size - 1);
  }
  
  /*
   * Peeks at the top of the stack.
   */
  public String peekLoopLabel()
  {
    int size = _loopLabelList.size();
    
    if (size == 0)
      return null;
    else
      return _loopLabelList.get(size - 1);
  }

  /**
   * Prints a contstant.
   */
  public void print(Value value)
    throws IOException
  {
    print(addValue(value));
  }

  /**
   * Adds a constant value.
   *
   * @return the generated id for the value
   */
  public String addValue(Value value)
  {
    String var = _valueMap.get(value);

    if (var == null) {
      String sValue = value.toString();

      StringBuilder name = new StringBuilder();
      name.append("qv_");

      for (int i = 0; i < 16 && i < sValue.length(); i++) {
	char ch = sValue.charAt(i);
	
	if (Character.isJavaIdentifierPart(ch))
	  name.append(ch);
      }

      name.append("_");
      name.append(generateId());
      
      var = name.toString();

      _valueMap.put(value, var);
    }

    return var;
  }

  /**
   * Adds a constant value.
   *
   * @return the generated id for the value
   */
  public void addClass(InterpretedClassDef cl)
  {
    _classList.add(cl);
  }

  /**
   * Adds an expression
   *
   * @return the generated id for the expression
   */
  public String addExpr(Expr expr)
  {
    String var = _exprMap.get(expr);

    if (var == null) {
      var = "_quercus_expr_" + generateId();

      _exprMap.put(expr, var);
    }

    return var;
  }

  /**
   * Adds an expression
   *
   * @return the generated id for the expression
   */
  public String addExprArray(Expr []exprArray)
  {
    String var = _exprArrayMap.get(exprArray);

    if (var == null) {
      var = "_quercus_expr_" + generateId();

      _exprArrayMap.put(exprArray, var);
    }

    return var;
  }

  /**
   * Adds a module
   *
   * @return the generated id for the expression
   */
  public String addModule(QuercusModule module)
  {
    String var = _moduleMap.get(module);

    if (var == null) {
      var = "_quercus_module_" + generateId();

      _moduleMap.put(module, var);
    }

    return var;
  }

  /**
   * Returns a static variable name.
   */
  public String createStaticVar()
  {
    String varName = "__quercus_static_" + _staticVarList.size();
    
    _staticVarList.add(varName);

    return varName;
  }

  /**
   * Adds a function id
   */
  public String addFunctionId(String name)
  {
    if (! getPhp().isStrict())
      name = name.toLowerCase();

    String id = "f_" + name;

    _functionIdMap.put(name, id);

    return id;
  }

  /**
   * Adds a class id
   */
  public String addClassId(String name)
  {
    name = name.toLowerCase();

    String id = "cl_" + name;
    
    _classIdMap.put(name, id);

    return id;
  }

  /**
   * Adds an expression
   *
   * @return the generated id for the expression
   */
  public String addCharArray(String string)
  {
    String varName = _charArrayMap.get(string);

    if (varName != null)
      return varName;
    
    StringBuilder sb = new StringBuilder("_qc_");

    for (int i = 0; i < string.length(); i++) {
      char ch = string.charAt(i);
      
      if (Character.isJavaIdentifierPart(ch))
        sb.append(ch);
    }

    varName = sb.toString();

    if (! _charArrayMap.values().contains(varName)) {
      _charArrayMap.put(string, varName);
      return varName;
    }

    String baseName = varName;

    int i = 0;
    do {
      varName = baseName + "_" + i++;
    } while (_charArrayMap.values().contains(varName));

    _charArrayMap.put(string, varName);

    return varName;
  }

  /**
   * Generates the tail.
   */
  public void generateCoda()
    throws IOException
  {
  }
}

