/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.page;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;
import java.util.logging.Level;

import com.caucho.vfs.*;

import com.caucho.util.L10N;
import com.caucho.util.ThreadPool;

import com.caucho.java.JavaCompileException;
import com.caucho.java.JavaCompiler;

import com.caucho.quercus.Quercus;
import com.caucho.quercus.ProQuercus;
import com.caucho.quercus.QuercusException;
import com.caucho.quercus.gen.QuercusGenerator;
import com.caucho.quercus.program.QuercusProgram;

/**
 * Each "page" refers to a quercus file.
 */
public class ProPageManager extends PageManager {
  private static final Logger log
    = Logger.getLogger(PageManager.class.getName());

  protected static final L10N L = new L10N(ProPageManager.class);
  
  private ArrayList<Path> _pendingPages = new ArrayList<Path>();
  
  private boolean _isCompilerActive;
  private boolean _isRemoveClassOnError;

  private boolean _isClosed;
  
  /**
   * Constructor.
   */ 
  public ProPageManager(Quercus quercus)
  {
    super(quercus);
  }

  /**
   * Returns the relative path.
   */
  public String getClassName(Path path)
  {
    if (path == null)
      return "tmp.eval";
    
    String pathName = path.getFullPath();
    String pwdName = getPwd().getFullPath();

    String relPath;

    if (pathName.startsWith(pwdName))
      relPath = pathName.substring(pwdName.length());
    else
      relPath = pathName;

    return "_quercus." + JavaCompiler.mangleName(relPath);
  }

  @Override
  protected QuercusPage compilePage(QuercusProgram program, Path path)
  {
    try {
      if (isCompile()) {
        QuercusGenerator gen = new QuercusGenerator((ProQuercus) getQuercus());

        Class pageClass;

        synchronized (_pendingPages) {
          pageClass = gen.generate(program);
        }

        return createPage(path, program, pageClass);
      }
      else if (isLazyCompile()) {
        boolean spawn = false;
	
        QuercusGenerator gen = new QuercusGenerator((ProQuercus) getQuercus());

        synchronized (program) {
          if (program.isCompiling()) {
            if (log.isLoggable(Level.FINE))
              log.log(Level.FINE, L.l("Quercus[{0}] is already being compiled", path));
          }
          else {
            try {
              Class pageClass = gen.preload(program);

              if (pageClass != null) {
                QuercusPage page = createPage(path, program, pageClass);
          
                program.setCompiledPage(page);

                if (page != null) {
                  if (log.isLoggable(Level.FINE))
                    log.log(Level.FINE, L.l("Quercus[{0}] loading precompiled page", path));
              
                  return page;
                }
              }
            } catch (Exception e) {
              log.log(Level.FINER, e.toString(), e);
            }
          }
        }

        synchronized (_pendingPages) {
          if (! _pendingPages.contains(path)) {
            _pendingPages.add(path);
            spawn = ! _isCompilerActive;
            _isCompilerActive = true;
          }
        }

        if (spawn)
          ThreadPool.getThreadPool().schedule(new CompileThread());

        return new InterpretedPage(program);
      }
      else
        return new InterpretedPage(program);
    } catch (JavaCompileException e) {
      log.log(Level.FINE, L.l("Quercus[{0}] loading interpreted page instead because of compile error",
                              path), e);
      
      program.setCompilable(false);

      return new InterpretedPage(program);
    } catch (Exception e) {
      throw new QuercusException(e);
    }
  }

  @Override
  protected void clearProgram(Path path, QuercusProgram program)
  {
    super.clearProgram(path, program);

    ((ProQuercus) getQuercus()).setCompileClassLoader(null);
  }

  private QuercusPage createPage(Path path,
				 QuercusProgram program,
				 Class pageClass)
  {
    try {
      QuercusPage page = (QuercusPage) pageClass.newInstance();

      page.init(getQuercus());
      
      Method selfPath = pageClass.getMethod("quercus_setSelfPath",
                                            new Class[] { Path.class });
      
      selfPath.invoke(null, path);

      program.setCompiledPage(page);
      
      return page;
    } catch (RuntimeException e) {
      throw e;
    } catch (ClassFormatError e) {
      throw e;
    } catch (Exception e) {
      throw new QuercusException(e);
    }
  }

  public void close()
  {
    _isClosed = true;
  }

  class CompileThread implements Runnable {
    public void run()
    {
      while (isActive()) {
        Path path = null;

        synchronized (_pendingPages) {
          if (_pendingPages.size() == 0) {
            _isCompilerActive = false;
            return;
          }
          
          path = _pendingPages.remove(0);
        }

        QuercusProgram program = _programCache.get(path);

        if (program == null
            || (program.getCompiledPage() != null
                && ! program.isModified())) {
          continue;
        }

	boolean isCompile = false;

        synchronized (program) {
          if (! program.isCompiling()) {
	    isCompile = true;
	    program.setCompiling(true);
	  }
	}

	if (isCompile) {
	  try {
	    compileCode(path, program);
	  } finally {
	    synchronized (program) {
	      program.setCompiling(false);
	    }
	  }
	}
      }
    }

    private void compileCode(Path path, QuercusProgram program)
    {
      QuercusGenerator gen
	= new QuercusGenerator((ProQuercus) getQuercus());
	    
      try {
	program.setCompiling(true);

	Class pageClass = gen.generate(program);

	createPage(path, program, pageClass);
      } catch (JavaCompileException e) {
	log.log(Level.FINE, L.l("Quercus[{0}] loading interpreted page instead because of compile error",
				path), e);
              
	program.setCompilable(false);
      } catch (ClassFormatError e) {
	log.log(Level.WARNING, e.toString(), e);

	try {
	  if (_isRemoveClassOnError)
	    gen.getClassFilePath(program).remove();
	} catch (Exception e1) {
	}
      } catch (Exception e) {
	log.log(Level.FINE, e.toString(), e);
      } finally {
	program.setCompiling(false);
      }
    }
  }
}
