/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * This file is part of Resin(R) Open Source
 *
 * Each copy or derived work must preserve the copyright notice and this
 * notice unmodified.
 *
 * Resin Open Source is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Resin Open Source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, or any warranty
 * of NON-INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Resin Open Source; if not, write to the
 *
 *   Free Software Foundation, Inc.
 *   59 Temple Place, Suite 330
 *   Boston, MA 02111-1307  USA
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.program;

import com.caucho.quercus.expr.CompilingExpr;
import com.caucho.quercus.expr.Expr;
import com.caucho.quercus.expr.GenerateExpr;
import com.caucho.quercus.env.StringValue;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.quercus.Location;

import java.io.IOException;
import java.util.Map;

/**
 * Represents an interpreted PHP class definition.
 */
public class ProClassDef extends InterpretedClassDef
{
  public ProClassDef(Location location,
                     String name,
		     String parentName,
		     String []ifaceList)
  {
    super(location, name, parentName, ifaceList);
  }

  /**
   * Analyzes the class
   */
  public void analyze()
  {
    for (AbstractFunction fun : _functionMap.values()) {
      FunctionGenerator funGen = ((CompilingFunction) fun).getGenerator();
      
      funGen.analyze();
    }
  }

  /**
   * Generates the class.
   */
  public void generate(PhpWriter out)
    throws IOException
  {
    out.addClass(this);

    String implClass;

    if (isInterface())
      implClass = "CompiledInterfaceDef";
    else
      implClass = isAbstract() ? "CompiledAbstractClassDef" : "CompiledClassDef";

    out.println();
    out.print("private ClassDef q_cl_" + getName() + " = ");
    out.print("new " + implClass + "(");

    Location location = getLocation();
    out.print("new com.caucho.quercus.Location(");
    out.printQuotedJavaString(location.getFileName());
    out.print(",");
    out.print(location.getLineNumber());
    out.print(",");
    out.printQuotedJavaString(location.getClassName());
    out.print(",");
    out.printQuotedJavaString(location.getFunctionName());
    out.print("), ");

    out.print("\"" + getName() + "\", ");

    if (getParentName() != null)
      out.print("\"" + getParentName() + "\"");
    else
      out.print("null");

    out.print(", new String[] {");
    String []ifaceList = getInterfaces();
    for (int i = 0; i < ifaceList.length; i++) {
      if (i != 0)
	out.print(", ");

      out.print("\"");
      out.print(ifaceList[i]);
      out.print("\"");
    }
    
    out.print("}, quercus_" + getName() + ".class");
    
    if (isFinal())
      out.print(", true");
    
    out.print(");");
    
    out.println();
    out.println("public static class quercus_" + getName() + " implements InstanceInitializer, CompiledClass {");
    out.pushDepth();

    if (false) {
      out.println();
      for (StringValue key : _fieldMap.keySet()) {
	out.println("private final int _f_" + key + ";");
      }
    }

    out.println();
    // out.println("public quercus_" + getName() + "(QuercusClass cl)");
    out.println("public void init(QuercusClass cl)");
    out.println("{");
    out.pushDepth();

    out.println("cl.addInitializer(this);");
    out.println();
    
    if (_constructor != null) {
      // php/393g
      // php/393i
      out.println("cl.setConstructor(fun_" + _constructor.getName() + ");");
      out.println("cl.addMethod(\"__construct\", fun_" + _constructor.getName() + ");");
      out.println();
    }
    
    if (_getField != null) {
      out.println("cl.setFieldGet(fun_" + _getField.getName() + ");");
      out.println();
    }
    
    if (_setField != null) {
      out.println("cl.setFieldSet(fun_" + _setField.getName() + ");");
      out.println();
    }
    
    if (_call != null) {
      out.println("cl.setCall(fun_" + _call.getName() + ");");
      out.println();
    }

    if (false) {
      for (StringValue key : _fieldMap.keySet()) {
	out.print("_f_" + key + " = ");
	out.println("cl.addFieldIndex(\"" + key + "\");");
      }
    }

    out.println();

    for (String key : _functionMap.keySet()) {
      AbstractFunction fun = _functionMap.get(key);

      /* XXX: abstract methods need to be added to QuercusClass
              php/393g, php/393i, php/39j2
      if (fun instanceof MethodDeclaration)
        continue;
        */

      out.print("cl.addMethod(\"");
      out.printJavaString(key);
      out.println("\", fun_" + key + ");");
    }

    for (Map.Entry<StringValue,Expr> entry : _fieldMap.entrySet()) {
      GenerateExpr exprGen = ((CompilingExpr) entry.getValue()).getGenerator();
      
      out.print("cl.addField(");
      out.print(entry.getKey());
      out.print(", ");
      out.print("0, ");
      exprGen.generateExpr(out);
      out.println(");");
    }

    for (Map.Entry<String,Expr> entry : _staticFieldMap.entrySet()) {
      GenerateExpr exprGen = ((CompilingExpr) entry.getValue()).getGenerator();
      
      out.print("cl.addStaticFieldExpr(\"");
      out.printJavaString(getName());
      out.print("\", \"");
      out.printJavaString(entry.getKey());
      out.print("\", ");
      exprGen.generateExpr(out);
      out.println(");");
    }

    for (Map.Entry<String,Expr> entry : _constMap.entrySet()) {
      GenerateExpr exprGen = ((CompilingExpr) entry.getValue()).getGenerator();
      
      out.print("cl.addConstant(\"");
      out.printJavaString(entry.getKey());
      out.print("\", ");
      exprGen.generateExpr(out);
      out.println(");");
    }
    
    out.popDepth();
    out.println("}");

    /*
    out.println();
    out.println("public void init(QuercusClass cl)");
    out.println("{");
    out.pushDepth();
    out.println("quercus_" + getName() + " def = new quercus_" + getName() +
		"(cl);");
    out.popDepth();
    out.println("}");
    */

    out.println();
    out.println("public void initInstance(Env env, Value valueArg)");
    //out.println("   throws Throwable");
    out.println("{");
    out.pushDepth();

    //out.println("CompiledObjectValue value = (CompiledObjectValue) valueArg;");
    out.println("ObjectValue value = (ObjectValue) valueArg;");

    for (Map.Entry<StringValue,Expr> entry : _fieldMap.entrySet()) {
      StringValue key = entry.getKey();
      Expr value = entry.getValue();
      
      GenerateExpr valueGen = ((CompilingExpr) value).getGenerator();

      if (false) {
	out.print("value._fields[_f_" + key + "] = ");
	valueGen.generate(out);
	out.println(";");
      }
      else {
	out.print("value.initField(");
	out.print(key);
	out.print(", ");
	valueGen.generate(out);
	out.println(");");
      }
	
    }

    /*
    for (AbstractFunction fun : _functionMap.values()) {
      fun.generateInit(out);
    }
    */
    
    out.popDepth();
    out.println("}");

    for (AbstractFunction fun : _functionMap.values()) {
      FunctionGenerator funGen = ((CompilingFunction) fun).getGenerator();

      funGen.generate(out);
    }

    // XXX:
    out.println();
    out.println("public void init(Env env)");
    out.println("{");
    out.pushDepth();

    for (Map.Entry<String,Expr> var : _staticFieldMap.entrySet()) {
      GenerateExpr varGen = ((CompilingExpr) var.getValue()).getGenerator();

      out.print("env.setGlobalValue(\"" + getName() + "::" + var.getKey() + "\", ");
      varGen.generate(out);
      out.println(");");
    }
    
    out.popDepth();
    out.println("}");

    out.popDepth();
    out.println("}");
  }

  /**
   * Generates the class initialization.
   */
  public void generateInit(PhpWriter out)
    throws IOException
  {
    out.print("addClass(\"");
    out.printJavaString(getName());
    out.print("\", new CompiledClassDef(\"");
    out.printJavaString(getName());
    out.print("\", ");
    if (getParentName() != null) {
      out.print("\"");
      out.printJavaString(getParentName());
      out.print("\"");
    }
    else
      out.print("null");

    out.print(", new String[] {");
    String []ifaceList = getInterfaces();
    for (int i = 0; i < ifaceList.length; i++) {
      if (i != 0)
	out.print(", ");

      out.print("\"");
      out.print(ifaceList[i]);
      out.print("\"");
    }
    
    out.println("}, quercus_" + getName() + ".class));");
  }
}

