/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.program;

import com.caucho.quercus.Location;
import com.caucho.quercus.expr.*;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.quercus.statement.CompilingStatement;
import com.caucho.quercus.statement.StatementGenerator;
import com.caucho.util.L10N;

import java.io.IOException;
import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * Represents sequence of statements.
 */
public class ProFunction extends Function
  implements CompilingFunction
{
  private static final Logger log
    = Logger.getLogger(ProFunction.class.getName());
  private static final L10N L = new L10N(ProFunction.class);

  public ProFunction(ExprFactory exprFactory,
                     Location location,
                     String name,
                     FunctionInfo info,
                     Arg []argList,
                     Statement []statementList)
  {
    super(exprFactory, location, name, info, argList, statementList);
  }

  public FunctionGenerator getGenerator()
  {
    return GENERATOR;
  }

  private FunctionGenerator GENERATOR = new FunctionGenerator() {
      private StatementGenerator getStatement()
      {
	return ((CompilingStatement) _statement).getGenerator();
      }
      
      /**
       * Analyzes the function.
       */
      public void analyze()
      {
	AnalyzeInfo info = new AnalyzeInfo(_info);

	for (int i = 0; i < _args.length; i++) {
	  Arg arg = _args[i];

	  ProVarExpr var = new ProVarExpr(_info.createVar(arg.getName()));
	  var.setVarState(VarState.VALID);

	  info.addVar(var);
	}

	_hasReturn = ! getStatement().analyze(info);
      }

      /**
       * Returns true if the function can generate the call directly.
       */
      public boolean canGenerateCall(Expr []args)
      {
	// can only handle case where the arg length matches
	return args.length <= _args.length;
      }

      /**
       * Generates code to calluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generate(PhpWriter out, GenerateExpr funExpr, Expr []args)
	throws IOException
      {
	generateImpl(out, funExpr, args, false);
      }

      /**
       * Generates code to calluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      public void generateRef(PhpWriter out, GenerateExpr funExpr, Expr []args)
	throws IOException
      {
	generateImpl(out, funExpr, args, true);
      }

      /**
       * Generates code to calluate the expression.
       *
       * @param out the writer to the Java source code.
       */
      private void generateImpl(PhpWriter out,
				GenerateExpr funExpr,
				Expr []args, boolean isRef)
	throws IOException
      {
	String ref;

	if (! _isReturnsReference)
	  ref = "";
	else
	  ref = "Ref";
    
	out.print("fun_" + _name + ".call" + ref + "(env");

	if (! isVariableArgs()) {
	}
	else if (args.length == 0 && _args.length == 0)
	  out.print(", Value.NULL_ARGS");
	else
	  out.print(", new Value[] {");

	for (int i = 0; i < _args.length; i++) {
	  if (i != 0 || ! isVariableArgs())
	    out.print(", ");

	  if (i < args.length) {
	    CompilingExpr arg = (CompilingExpr) args[i];
	    GenerateExpr argGen = arg.getGenerator();

	    if (_args[i].isReference()) {
	      argGen.generateRef(out);
	    }
	    else {
	      argGen.generateArg(out);
	    }
	  }
	  else if (_args[i].getDefault() != null) {
	    CompilingExpr arg = (CompilingExpr) _args[i].getDefault();
	    
	    arg.getGenerator().generateArg(out);
	  }
	  else
	    out.print("NullValue.NULL");
	}

	for (int i = _args.length; i < args.length; i++) {
	  CompilingExpr arg = (CompilingExpr) args[i];
	  
	  if (i != 0)
	    out.print(", ");
	  
	  arg.getGenerator().generateArg(out);
	}

	if (isVariableArgs() && (_args.length > 0 || args.length > 0))
	  out.print("}");

	out.print(")");

	//if (_isReturnsReference && ! isRef)
	//  out.print(".toValue()");
    
	/* php/3a70 - should only be generated from generateCopy
	   if (! (isRef && _isReturnsReference))
	   out.print(".copyReturn()");
	*/
      }

      private boolean isVariableArgs()
      {
	return _info.isVariableArgs() || _args.length > 5;
      }

      private boolean isVariableMap()
      {
	// return _info.isVariableVar();
	// php/3254
	return _info.isUsesSymbolTable() || _info.isVariableVar();
      }

      /**
       * Generates the code for the class component.
       *
       * @param out the writer to the output stream.
       */
      @Override
      public void generate(PhpWriter out)
	throws IOException
      {
	if (isVariableArgs())
	  generateVariableArgs(out);
	else
	  generateFixedArgs(out);
      }

      /**
       * Generates the code for the class component.
       *
       * @param out the writer to the output stream.
       */
      public void generateFixedArgs(PhpWriter out)
	throws IOException
      {
	out.println();
	out.print("public static final fun_" + _name + " fun_" + _name);
	out.println(" = new fun_" + _name + "();");

	String ref = _isReturnsReference ? "Ref" : "";

	out.println();
	out.print("final static class fun_" + _name + " extends ");
    
	if (isStatic())
	  out.print("CompiledFunction" + ref + "_" + _args.length);
	else
	  out.print("CompiledMethod" + ref + "_" + _args.length);

	out.println(" {");
	out.pushDepth();
    
	out.println("fun_" + _name + "()");
	out.println("{");
	out.print("  super(\"");
	out.printJavaString(_name);
	out.print("\"");

	for (int i = 0; i < _args.length; i++) {
	  out.print(", ");

	  CompilingExpr defaultExpr = (CompilingExpr) _args[i].getDefault();

	  defaultExpr.getGenerator().generateExpr(out);
	}
	out.println(");");

	out.println("}");
        
	out.println();
	if (isStatic())
	  out.print("public final Value call" + ref + "(Env env");
	else
	  out.print("public final Value callMethod" + ref + "(Env env, Value q_this");

	for (int i = 0; i < _args.length; i++) {
	  out.print(", ");

	  out.print("Value v_");
	  out.print(_args[i].getName());
	}
	out.println(")");

	out.println("{");

	out.pushDepth();

	generateBody(out);

	out.popDepth();

	out.println("}");
	out.popDepth();
	out.println("}");

        if (! isGlobal() || isFinal() || ! isPublic() || getDeclaringClassName() != null) {
          out.println();
          out.println("static {");
          out.pushDepth();
          
          if (! isGlobal())
            out.println("fun_" + _name + ".setGlobal(false);");
          
          if (isFinal())
            out.println("fun_" + _name + ".setFinal(true);");
          
          if (isProtected())
            out.println("fun_" + _name + ".setVisibility(Visibility.PROTECTED);");
          else if (isPrivate())
            out.println("fun_" + _name + ".setVisibility(Visibility.PRIVATE);");
          
          out.println("fun_" + _name + ".setDeclaringClassName(\"" + getDeclaringClassName() +  "\");");
          
          out.popDepth();
          
          out.println("}");
        }
      }

      /**
       * Generates the code for the class component.
       *
       * @param out the writer to the output stream.
       */
      public void generateVariableArgs(PhpWriter out)
	throws IOException
      {
	out.println();
	out.print("public static final fun_" + _name + " fun_" + _name);
	out.println(" = new fun_" + _name + "();");
    
	out.println();
	out.print("public static final class fun_" + _name + " extends ");

	String ref = _isReturnsReference ? "Ref" : "";

	if (isStatic())
	  out.println(" CompiledFunction" + ref + "_N {");
	else
	  out.println(" CompiledMethod" + ref + "_N {");

	out.pushDepth();
	out.println("fun_" + _name + "()");
	out.println("{");
	out.print("  super(\"");
	out.printJavaString(_name);
	out.print("\", new Expr[] {");

	for (int i = 0; i < _args.length; i++) {
	  if (i != 0)
	    out.print(", ");

	  CompilingExpr defaultExpr = (CompilingExpr) _args[i].getDefault();

	  if (defaultExpr != null)
	    defaultExpr.getGenerator().generateExpr(out);
	  else
	    out.print("null");
	}
	out.println("});");
	out.println("}");
    

	out.println();
	if (isStatic())
	  out.println("public Value call" + ref + "Impl(Env env, Value []args)");
	else
	  out.println("public Value callMethod" + ref + "Impl(Env env, Value q_this, Value []args)");

	out.println("{");
	out.pushDepth();

	if (_info.isVariableArgs()) {
	  out.println("Value []quercus_oldArgs = env.setFunctionArgs(args);");
	  out.println("try {");
	  out.pushDepth();
	}

	generateBody(out);

	if (_info.isVariableArgs()) {
	  out.popDepth();
	  out.println("} finally {");
	  out.pushDepth();
	  out.println("env.restoreFunctionArgs(quercus_oldArgs);");
	  out.popDepth();
	  out.println("}");
	}

	out.popDepth();
	out.println("}");

	out.popDepth();
	out.println("}");
      }

      private void generateBody(PhpWriter out)
	throws IOException
      {
	out.println("env.checkTimeout();");
	out.println();

	for (VarInfo var : _info.getVariables()) {
	  String varName = "v_" + var.getName();
	  String argName = varName;

	  if (! var.isArgument())
	    out.print("Value ");
	  else if (isVariableArgs()) {
	    out.print("Value ");

	    argName = "args[" + var.getArgumentIndex() + "]";
	  }

	  if (var.isArgument()) {
	    if (isVariableMap()) {
	      if (var.isRefArgument()) {
		// php/3214
		out.println(varName + " = " + argName + ".toRefVar();");
	      }
	      else
		out.println(varName + " = " + argName + ".toVar();");
	    }
	    else if (var.isReadOnly() && var.isValue()) {
	      // php/3a70, php/343k
	      out.println(varName + " = " + argName + ".toArgValueReadOnly();");
	    }
	    /*
	      else if (var.isReadOnly()) {
	      // php/3783
	      out.println(varName + " = " + argName + ".toRefValue();");
	      }
	    */
	    else if (var.isRefArgument()) {
	      // php/344r, 3a57
	      out.println(varName + " = " + argName + ".toRefVar();");
	    }
	    else if (var.isAssigned() && var.isReference()) {
	      out.println(varName + " = " + argName + ".toVar();");
	    }
	    else if (var.isReference() && ! var.isReadOnly()) {
	      // php/3a6k
          out.println(varName + " = " + argName + ".toVar();");
	    }
	    else {
	      // php/399j
	      out.println(varName + " = " + argName + ".toArgValue();");
	    }
	  }
	  else {
	    // local variable, i.e. not argument

	    if (var.isValue())
	      out.println(varName + " = NullValue.NULL;");
	    else
	      out.println(varName + " = null;");
	  }
	}

	for (String var : _info.getTempVariables()) {
	  out.println("Value " + var + ";");
	}

	//generateVisibilityCheck(out);
	
    if (isVariableMap()) {
      out.println("java.util.HashMap<String,Var> _quercus_map = new java.util.HashMap<String,Var>();");
      out.println("java.util.HashMap<String,Var> _quercus_oldMap = env.pushEnv(_quercus_map);");

      if (! isStatic())
        out.println("Value q_oldThis = env.setThis(q_this);");
      
      out.println("try {");
      out.pushDepth();

      for (int i = 0; i < _args.length; i++) {
        out.print("_quercus_map.put(\"");
        out.printJavaString(_args[i].getName());
        out.println("\", (Var) v_" + _args[i].getName() + ");");
      }
    }

	CompilingStatement statement = (CompilingStatement) _statement;
	statement.getGenerator().generate(out);

    if (isVariableMap()) {
      out.popDepth();
      out.println("} finally {");
      out.pushDepth();
      // out.println("env.setFunctionArgs(_quercus_oldArgs);");
      out.println("env.popEnv(_quercus_oldMap);");

      if (! isStatic())
        out.println("env.setThis(q_oldThis);");
      
      out.popDepth();
      out.println("}");
    }


	if (statement.getGenerator().fallThrough() != Statement.RETURN)
	  out.println("return com.caucho.quercus.env.NullValue.NULL;");
      }

      private void generateVisibilityCheck(PhpWriter out)
        throws IOException
      {
        if (isPublic()) {
        }
        else if (isProtected()) {
          out.println("if (q_oldThis != null)");
          out.println("  q_oldThis.checkProtected(env, getDeclaringClassName());");
        }
        else {
          // private
          
          out.println("if (q_oldThis != null)");
          out.println("  q_oldThis.checkPrivate(env, getDeclaringClassName());");
        }
      }
      
      /**
       * Generates the code for the initialization component.
       *
       * @param out the writer to the output stream.
       */
      public void generateInit(PhpWriter out)
	throws IOException
      {
	out.println();
	out.print("addFunction(\"");
	out.printJavaString(_name.toLowerCase());
	out.println("\", fun_" + _name + ");");
      }
    };
}

