/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.quercus.statement;

import com.caucho.quercus.Location;
import com.caucho.quercus.expr.CompilingExpr;
import com.caucho.quercus.expr.Expr;
import com.caucho.quercus.expr.GenerateExpr;
import com.caucho.quercus.gen.AnalyzeInfo;
import com.caucho.quercus.gen.PhpWriter;
import com.caucho.quercus.program.BreakStatement;
import com.caucho.quercus.program.Statement;

import java.io.IOException;

/**
 * Represents a break expression statement in a PHP program.
 */
public class ProBreakStatement extends BreakStatement
  implements CompilingStatement
{
  //public static final BreakStatement BREAK = new ProBreakStatement();
  
  /**
   * Creates the break statement.
   */
  private ProBreakStatement()
  {
  }
  
  public ProBreakStatement(Location location, Expr target)
  {
    super(location, target);
  }


  public StatementGenerator getGenerator()
  {
    return GENERATOR;
  }

  private StatementGenerator GENERATOR = new StatementGenerator() {
      protected Location getLocation()
      {
	return Location.UNKNOWN;
      }
      
      private GenerateExpr getTarget()
      {
        if (_target != null)
          return ((CompilingExpr) _target).getGenerator();
        else
          return null;
      }
      
      /**
       * Analyze the statement
       *
       * @return true if the following statement can be executed
       */
      public boolean analyze(AnalyzeInfo info)
      {
        if (getTarget() != null)
          getTarget().analyze(info);
        
	info.mergeLoopBreakInfo();

	// quercus/067i
	return true;
      }

      /**
       * Returns true if the statement can fallthrough.
       */
      @Override
      public int fallThrough()
      {
	return BREAK_FALL_THROUGH;
      }

      /**
       * Generates the Java code for the statement.
       *
       * @param out the writer to the generated Java source.
       */
      protected void generateImpl(PhpWriter out)
        throws IOException
      {
        if (out.getLoopLabelSize() == 0) {
          // need to send a php error instead of a compile error for buggy code
          // required for Wordpress OpenID plugin (and others)
          out.println("env.error(\"No loop/switch statement to break out of\");");
        }
        else if (getTarget() == null) {
          out.println("if (true) break " + out.peekLoopLabel() + ";");
        }
        else {
          out.print("switch (");
          getTarget().generateInt(out);
          out.println(") {");
          
          out.pushDepth();
          
          int labels = out.getLoopLabelSize();
          
          String parentLoopLabel = out.getLoopLabel(labels - 1);
          
          for (int i = labels; i > 0; i--) {
            out.println("case " + i
                        + ": break " + out.getLoopLabel(labels - i) + ";");
          }
          
          if (parentLoopLabel != null) {
            out.println("default: if (true) break " + parentLoopLabel + ";");
          }

          out.popDepth();
          out.println("}");
        }
      }
    };
}

