/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.admin;

import com.caucho.config.ConfigException;
import com.caucho.db.jdbc.DataSourceImpl;
import com.caucho.jmx.Jmx;
import com.caucho.lifecycle.Lifecycle;
import com.caucho.loader.EnvironmentLocal;
import com.caucho.management.server.AbstractNotificationObject;
import com.caucho.management.server.LogMessage;
import com.caucho.management.server.LogNotification;
import com.caucho.management.server.LogServiceMXBean;
import com.caucho.util.L10N;
import com.caucho.util.LruCache;
import com.caucho.vfs.Path;

import javax.sql.DataSource;
import javax.management.MBeanNotificationInfo;
import javax.management.ObjectName;
import javax.annotation.PostConstruct;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Persistent logging.
 */
public class LogService
{
  private static Logger log
    = Logger.getLogger(LogService.class.getName());
  private static L10N L = new L10N(LogService.class);

  private static final MBeanNotificationInfo []_notificationInfo;
  private static ObjectName _objectName;

  private static final EnvironmentLocal<LogService> _local
    = new EnvironmentLocal<LogService>();

  private final ProManagement _management;

  private Admin _admin = new Admin();
  private volatile long _seq;

  private String _path = "log";
  private Level _level = Level.INFO;
  private boolean _isEnable = true;

  private LruCache<String,Integer> _nameCache
    = new LruCache<String,Integer>(8192);

  private LruCache<Integer,String> _revNameCache
    = new LruCache<Integer,String>(8192);

  private DataSource _database;
  private String _logTable;
  private String _logNameTable;

  private Lifecycle _lifecycle = new Lifecycle();

  private Connection _conn;
  private PreparedStatement _selectNameStmt;
  private PreparedStatement _revSelectNameStmt;
  private PreparedStatement _insertNameStmt;
  private PreparedStatement _insertStmt;
  private PreparedStatement _selectStmt;

  LogService(ProManagement management)
  {
    _management = management;

    _local.set(this, ClassLoader.getSystemClassLoader());
  }

  /**
   * Enables the service.
   */
  public void setEnable(boolean isEnable)
  {
    _isEnable = isEnable;
  }

  /**
   * Sets the filesystem path for storing the log recoods relative
   * to the {@link Management} path, default is "log".
   */
  public void setPath(String path)
  {
    _path = path;
  }

   /**
    * Sets the level of records to capture, default is "info".
    */
  public void setLevel(String level)
    throws ConfigException
  {
    if (level.equals("off"))
      _level = Level.OFF;
    else if (level.equals("severe"))
      _level = Level.SEVERE;
    else if (level.equals("warning"))
      _level = Level.WARNING;
    else if (level.equals("info"))
      _level = Level.INFO;
    else if (level.equals("config"))
      _level = Level.CONFIG;
    else if (level.equals("fine"))
      _level = Level.FINE;
    else if (level.equals("finer"))
      _level = Level.FINER;
    else if (level.equals("finest"))
      _level = Level.FINEST;
    else if (level.equals("all"))
      _level = Level.ALL;
    else
      throw new ConfigException(L.l("`{0}' is an unknown log level.  Log levels are:\noff - disable logging\nsevere - severe errors only\nwarning - warnings\ninfo - information\nconfig - configuration\nfine - fine debugging\nfiner - finer debugging\nfinest - finest debugging\nall - all debugging", level));
  }

  void init()
  {
    if (! _lifecycle.toActive())
      return;

    try {
      if (_management.getPath() == null)
        return;

      Path path = _management.getPath().lookup(_path);
      String serverId = _management.getServerId();

      if (serverId == null)
        serverId = "";

      DataSourceImpl dataSource = new DataSourceImpl();
      dataSource.setPath(path);
      dataSource.setRemoveOnError(true);
      dataSource.init();

      _database = dataSource;

      _logTable = escapeName("log_" + serverId);
      _logNameTable = escapeName("log_name_" + serverId);

      _conn = _database.getConnection();

      createDatabases();

      String sql;

      sql = ("select id from " + _logNameTable + " where name=?");

      _selectNameStmt = _conn.prepareStatement(sql);

      sql = ("select name from " + _logNameTable + " where id=?");

      _revSelectNameStmt = _conn.prepareStatement(sql);

      sql = ("insert into " + _logNameTable + " (name) VALUES (?)");

      _insertNameStmt = _conn.prepareStatement(sql, Statement.RETURN_GENERATED_KEYS);

      sql = ("insert into " + _logTable
             + " (log_time,name_id,level,server_id,thread,message)"
             + " VALUES (?,?,?,?,?,?)");

      _insertStmt = _conn.prepareStatement(sql);

      sql = ("select log_time,name_id,level,server_id,thread,message"
             + " from " + _logTable
             + " where ? <= log_time and log_time <= ?"
             + " order by log_time");

      _selectStmt = _conn.prepareStatement(sql);
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw ConfigException.create(e);
    }

    try {
      _objectName = new ObjectName("resin:type=" + getClass().getSimpleName());

      Jmx.register(_admin, _objectName);
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw ConfigException.create(e);
    }

    PersistentLogHandler handler = new PersistentLogHandler();
    handler.setLevel(_level);

    Logger.getLogger("").addHandler(handler);
  }

  public static LogService getLocal()
  {
    return _local.get();
  }

  private static String escapeName(String name)
  {
    StringBuilder cleanName = new StringBuilder();

    for (int i = 0; i < name.length(); i++) {
      char ch = name.charAt(i);

      if ('a' <= ch && ch <= 'z'
          || 'A' <= ch && ch <= 'Z'
          || '0' <= ch && ch <= '9'
          || ch == '_') {
        cleanName.append(ch);
      }
      else
        cleanName.append('_');
    }

    return cleanName.toString();
  }

  public void log(long timestamp, String name, Level level, String server,
                  String thread, String message)
  {
    int nameId = getNameId(name);
    int levelId = getNameId(level.toString());
    int serverId = getNameId(server);

    synchronized (this) {
      try {
        _insertStmt.setLong(1, timestamp);
        _insertStmt.setInt(2, nameId);
        _insertStmt.setInt(3, level.intValue());
        _insertStmt.setInt(4, serverId);
        _insertStmt.setString(5, thread);
        _insertStmt.setString(6, message);

        _insertStmt.execute();
      } catch (SQLException e) {
        log.log(Level.FINER, e.toString(), e);
      }
    }

    if (Level.WARNING.intValue() <= level.intValue()) {
      String type = "caucho.log.broadcast";

      LogNotification notif
        = new LogNotification(type, _objectName,
                              _seq++, timestamp, message);

      notif.setName(name);
      notif.setLevel(level.toString());
      notif.setServer(server);
      notif.setThread(thread);

      _admin.sendNotification(notif);
      //msg.setTimestamp(timestamp);
    }
  }

  public LogEntry []findRecentLogEntries(int min)
  {
    return new LogEntry[0];
  }

  private int getNameId(String name)
  {
    synchronized (_nameCache) {
      Integer id = _nameCache.get(name);

      if (id != null)
        return id;
    }

    synchronized (this) {
      try {
        _selectNameStmt.setString(1, name);

        ResultSet rs = _selectNameStmt.executeQuery();

        if (rs.next()) {
          int id = rs.getInt(1);
          rs.close();

          synchronized (_nameCache) {
            _nameCache.put(name, id);
            _revNameCache.put(id, name);
          }

          return id;
        }

        rs.close();
      } catch (SQLException e) {
        log.log(Level.FINER, e.toString(), e);
      }

      try {
        _insertNameStmt.setString(1, name);

        _insertNameStmt.execute();

        ResultSet rs = _insertNameStmt.getGeneratedKeys();

        if (rs.next()) {
          int id = rs.getInt(1);
          rs.close();

          synchronized (_nameCache) {
            _nameCache.put(name, id);
          }

          return id;
        }

        rs.close();
      } catch (SQLException e) {
        log.log(Level.FINER, e.toString(), e);
      }

      throw new RuntimeException(L.l("Can't find log key"));
    }
  }

  private String getName(int id)
  {
    synchronized (_nameCache) {
      String key = _revNameCache.get(id);

      if (key != null)
        return key;
    }

    synchronized (this) {
      try {
        _revSelectNameStmt.setInt(1, id);

        ResultSet rs = _revSelectNameStmt.executeQuery();

        if (rs.next()) {
          String name = rs.getString(1);
          rs.close();

          synchronized (_nameCache) {
            _nameCache.put(name, id);
            _revNameCache.put(id, name);

            return name;
          }
        }

        rs.close();
      } catch (SQLException e) {
        log.log(Level.FINER, e.toString(), e);
      }

      return null;
    }
  }

  private void createDatabases()
    throws SQLException
  {
    Statement stmt = _conn.createStatement();

    boolean hasLogNames = false;
    try {
      String sql = ("select id, name from " + _logNameTable + " WHERE id=1");

      ResultSet rs = stmt.executeQuery(sql);

      hasLogNames = true;
    } catch (Exception e) {
      log.log(Level.FINEST, e.toString(), e);
    }

    if (! hasLogNames) {
      try {
        String sql = ("drop table " + _logNameTable);

        ResultSet rs = stmt.executeQuery(sql);
      } catch (Exception e) {
        log.log(Level.FINER, e.toString(), e);
      }

      try {
        StringBuilder sb = new StringBuilder();
        sb.append("create table " + _logNameTable + " (");
        sb.append("  id bigint primary key auto_increment,\n");
        sb.append("  name VARCHAR(512)");
        sb.append(")");

        stmt.executeUpdate(sb.toString());
      } catch (Exception e) {
        log.log(Level.FINER, e.toString(), e);
      }
    }

    boolean hasLog = false;
    try {
      String sql = ("select time, name_id, level, server_id, thread, message"
                    + " from " + _logTable + " WHERE log_time=1");

      ResultSet rs = stmt.executeQuery(sql);
      rs.close();

      hasLog = true;
    } catch (Exception e) {
      log.log(Level.FINEST, e.toString(), e);
    }

    if (! hasLog) {
      try {
        String sql = ("drop table " + _logTable);

        ResultSet rs = stmt.executeQuery(sql);
      } catch (Exception e) {
        log.log(Level.FINER, e.toString(), e);
      }

      try {
        StringBuilder sb = new StringBuilder();
        sb.append("create table " + _logTable + " (");
        sb.append("  log_time bigint,\n");
        sb.append("  name_id integer,\n");
        sb.append("  level integer,\n");
        sb.append("  server_id integer,\n");
        sb.append("  thread varchar(64),\n");
        sb.append("  message clob");
        sb.append(")");

        stmt.executeUpdate(sb.toString());
      } catch (Exception e) {
        log.log(Level.FINER, e.toString(), e);
      }
    }
  }

  public LogMessage []findMessages(long minTime, long maxTime)
  {
    try {
      ArrayList<LogMessage> values = new ArrayList<LogMessage>();

      _selectStmt.setLong(1, minTime);
      _selectStmt.setLong(2, maxTime);

      ResultSet rs = _selectStmt.executeQuery();

      while (rs.next()) {
        long timestamp = rs.getLong(1);
        int nameId = rs.getInt(2);
        int level = rs.getInt(3);
        int serverId = rs.getInt(4);
        String thread = rs.getString(5);
        String message = rs.getString(6);

        LogMessage logMessage = new LogMessage();
        logMessage.setTimestamp(timestamp);

        if (Level.SEVERE.intValue() <= level)
          logMessage.setLevel("severe");
        else if (Level.WARNING.intValue() <= level)
          logMessage.setLevel("warning");
        else if (Level.INFO.intValue() <= level)
          logMessage.setLevel("info");
        else if (Level.CONFIG.intValue() <= level)
          logMessage.setLevel("config");
        else if (Level.FINE.intValue() <= level)
          logMessage.setLevel("fine");
        else if (Level.FINER.intValue() <= level)
          logMessage.setLevel("finer");
        else if (Level.FINEST.intValue() <= level)
          logMessage.setLevel("finest");
        else
          logMessage.setLevel("off");

        logMessage.setName(getName(nameId));
        logMessage.setServer(getName(serverId));

        logMessage.setThread(thread);
        logMessage.setMessage(message);

        values.add(logMessage);
      }
      rs.close();

      LogMessage []messages = new LogMessage[values.size()];
      values.toArray(messages);

      return messages;
    } catch (SQLException e) {
      log.log(Level.WARNING, e.toString(), e);

      return null;
    }
  }

  public class Admin
    extends AbstractNotificationObject
    implements LogServiceMXBean
  {
    public String getName()
    {
      return null;
    }

    public String getType()
    {
      return "LogService";
    }

    public LogMessage []findMessages(long minTime, long maxTime)
    {
      return LogService.this.findMessages(minTime, maxTime);
    }

    public MBeanNotificationInfo []getNotificationInfo()
    {
      return _notificationInfo;
    }
  }

  static {
    _notificationInfo = new MBeanNotificationInfo[] {
      new MBeanNotificationInfo(new String[] { "caucho.log.broadcast" },
                                LogNotification.class.getName(),
                                L.l("Emits logger warnings and severe messages"))
    };
  }
}
