/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */
package com.caucho.server.admin;

import com.caucho.config.types.Period;
import com.caucho.jca.AbstractResourceAdapter;
import com.caucho.loader.DynamicClassLoader;
import com.caucho.log.EnvironmentStream;
import com.caucho.util.Alarm;
import com.caucho.util.AlarmListener;
import com.caucho.util.CharBuffer;
import com.caucho.util.L10N;
import com.caucho.util.QDate;
import com.caucho.vfs.Path;
import com.caucho.vfs.ReadStream;
import com.caucho.vfs.Vfs;

import javax.annotation.PostConstruct;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.openmbean.CompositeData;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.resource.spi.BootstrapContext;
import java.io.StringWriter;
import java.lang.management.LockInfo;
import java.lang.management.MonitorInfo;
import java.lang.management.ThreadInfo;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * A thread to restart the Resin server if it fails for some reason.
 *
 * <p>This pinging only catches some problems because it's running in the
 * same JDK process as Resin itself.  If the entire JDK freezes, this
 * thread will freeze as well.  Assuming the JDK doesn't freeze, the
 * PingThread will catch errors like deadlocks.
 */
public class PingThread extends AbstractResourceAdapter
  implements Runnable, AlarmListener
{
  static final Logger log
    = Logger.getLogger(PingThread.class.getName());
  static final L10N L = new L10N(PingThread.class);

  private static final ObjectName _threadMBean;

  // The server's host
  private String _pingHost;
  
  // The server's port
  private int _pingPort;
  
  // List of URLs to check
  private ArrayList<String> _urlList = new ArrayList<String>();
  
  // List of URL Paths to check
  private ArrayList<Path> _pathList = new ArrayList<Path>();
  
  // How long to wait for the initial sleep
  private long _initialSleepTime = 15 * 60 * 1000L;
  
  // How long to sleep between checks
  private long _sleepTime = 15 * 60 * 1000L;
  
  // How long to wait to retry a failing URL
  private long _retryTime = 1000L;
  
  // How many times to try a failing URL before giving up
  private int _tryCount = 3;
  
  // How many milliseconds to wait for a response before giving up
  private long _socketTimeout = 10 * 1000L;
  
  // How long extra the thread should count a failed response
  private long _failureTime = 5 * 60 * 1000L;

  // The last time the ping was tried.
  private long _lastPingTime;

  // The alarm listener.
  private Alarm _alarm;

  // The class loader.
  private DynamicClassLoader _loader;

  // The MBean server for deadlock checking
  private MBeanServer _mbeanServer;

  private volatile long _okayTime;

  // True if the ping has been closed
  private volatile boolean _isClosed;
  
  // True if the ping has failed and forced an exit
  private volatile boolean _isFailed;

  /**
   * Returns the server's ping host
   */
  public String getPingHost()
  {
    return _pingHost;
  }

  /**
   * Sets the server's ping host
   */
  public void setPingHost(String pingHost)
  {
    _pingHost = pingHost;
  }

  /**
   * Returns the server's ping port
   */
  public int getPingPort()
  {
    return _pingPort;
  }

  /**
   * Sets the server's ping port
   */
  public void setPingPort(int pingPort)
  {
    _pingPort = pingPort;
  }

  /**
   * Adds a new URL to be tested.
   */
  public void addURL(String url)
  {
    _urlList.add(url);
  }

  /**
   * Gets the sleep time between pings
   */
  public long getSleepTime()
  {
    return _sleepTime;
  }

  /**
   * Sets the sleep time between pings
   */
  public void setSleepTime(Period sleepTime)
  {
    _sleepTime = sleepTime.getPeriod();
  }

  /**
   * Sets the sleep time on startup
   */
  public void setInitialSleepTime(Period sleepTime)
  {
    _initialSleepTime = sleepTime.getPeriod();
  }

  /**
   * Gets the retry time between failing pings
   */
  public long getRetryTime()
  {
    return _retryTime;
  }

  /**
   * Sets the retry time between failing pings
   */
  public void setRetryTime(Period retryTime)
  {
    _retryTime = retryTime.getPeriod();
  }

  /**
   * Gets the number of times to retry before giving up
   */
  public int getTryCount()
  {
    return _tryCount;
  }

  /**
   * Sets the number of times to retry before giving up
   */
  public void setTryCount(int tryCount)
  {
    _tryCount = tryCount;
  }

  /**
   * Gets the time to wait on a non-responsive socket before giving up.
   */
  public long getSocketTimeout()
  {
    return _socketTimeout;
  }

  /**
   * Sets the number of times to retry before giving up
   */
  public void setSocketTimeout(Period timeout)
  {
    _socketTimeout = timeout.getPeriod();
  }

  /**
   * Initialize the pinger.
   */
  @PostConstruct
  public void init()
  {
    for (int i = 0; i < _urlList.size(); i++) {
      String url = _urlList.get(i);

      if (url.startsWith("/")) {
        if (_pingPort > 0)
          url = "http://" + _pingHost + ":" + _pingPort + url;
        else
          url = "http://" + _pingHost + url;
      }
         
      _pathList.add(Vfs.lookup(url));
    }

    try {
      Context ic = new InitialContext();
      MBeanServer server;
      
      server = (MBeanServer) ic.lookup("java:comp/env/jmx/MBeanServer");

      if (server.isRegistered(_threadMBean)) {
	_mbeanServer = server;
      }
    } catch (Throwable e) {
      log.log(Level.FINER, e.toString(), e);
    }
    
    log.finer("PingThread[] initialized");
  }

  /**
   * Start the pinger.
   */
  public void start(BootstrapContext cxt)
  {
    log.info("PingThread[] starting, checking " + _urlList);
    
    _loader = (DynamicClassLoader) Thread.currentThread().getContextClassLoader();

    _okayTime = System.currentTimeMillis();
    _isClosed = false;
    
    _alarm = new Alarm("ping", this, _initialSleepTime);
    
    Thread thread = new Thread(this, "resin-ping");
    thread.setDaemon(true);
    thread.start();
  }

  /**
   * Periodically pings the server to check for anything going wrong.
   */
  public void handleAlarm(Alarm alarm)
  {
    if (_isClosed || _isFailed) {
      log.fine("ping alarm closed at " +
	       QDate.formatLocal(System.currentTimeMillis()));
      return;
    }
    
    log.fine("ping alarm checking: " +
	     QDate.formatLocal(System.currentTimeMillis()));

    try {
      boolean isValid = true;

      isValid = checkJVMDeadlock();

      if (isValid)
	isValid = pingURLs();

      if (isValid)
	_okayTime = System.currentTimeMillis();
    } catch (Throwable e) {
      log.log(Level.WARNING, e.toString(), e);
    } finally {
      if (! _isClosed && ! _isFailed)
	_alarm.queue(_sleepTime);
    }
  }

  /**
   * The ping thread checks to make sure the alarm has been successful.
   * PingThread uses this separate thread to minimize the possible
   * freezes.  run() can only freeze if the entire JVM is frozen
   * and in that case, there's nothing a Java-based ping can do about
   * the situation.
   */
  public void run()
  {
    Thread thread = Thread.currentThread();
    // Can't use Alarm.getCurrentTime() because that might be frozen
    
    long startTime = System.currentTimeMillis();
    long okayInterval = _sleepTime + _tryCount * _retryTime;

    boolean isFine = log.isLoggable(Level.FINE);

    Runtime runtime = Runtime.getRuntime();
    
    while (! _isClosed) {
      try {
	long now = System.currentTimeMillis();

	if (now < startTime + _initialSleepTime + _failureTime) {
	}

	else if (_okayTime + okayInterval + _failureTime < now) {
	  // can't send a message because that might be frozen

	  //if (isFine)
	  //  System.err.println("Closing because of ping delay\n");

	  EnvironmentStream.logStderr("Stopping Resin due to detected alarm freeze.");

          if (_mbeanServer != null) {
            try {
              EnvironmentStream.logStderr("calling a thread dump");

              Object data = (CompositeData[]) _mbeanServer.invoke(
                _threadMBean,
                "dumpAllThreads",
                new Object[]{true, true},
                new String[]{"boolean", "boolean"});
              StringWriter writer =  null;
              if (data.getClass().isArray() && CompositeData.class.isAssignableFrom(data.getClass().getComponentType())) {
                CompositeData []composites = (CompositeData[]) data;
                writer = new StringWriter();
                for (CompositeData composite : composites) {
                  writer.write(toString(ThreadInfo.from(composite)));
                  writer.write('\n');
                }
              } else if (data.getClass().isArray() && ThreadInfo.class.isAssignableFrom(data.getClass().getComponentType())) {

                ThreadInfo []threads = (ThreadInfo[]) data;
                writer = new StringWriter();
                for (ThreadInfo threadInfo : threads) {
                  writer.write(toString(threadInfo));
                  writer.write('\n');
                }
              } else {
                EnvironmentStream.logStderr("Can't handle type: " + data.getClass());
              }

              if (writer != null)
                EnvironmentStream.logStderr(writer.getBuffer().toString());


              EnvironmentStream.logStderr(writer.getBuffer().toString());
            } catch (Throwable t) {
              EnvironmentStream.logStderr(t.getMessage(), t);
            }
          }

          runtime.halt(1);
	}

	thread.interrupted();
	thread.sleep(60000);
      } catch (Throwable e) {
	// e.printStackTrace();
      }
    }
  }

  private String toString(ThreadInfo thread)
  {
    StringBuilder builder = new StringBuilder("\""
      + thread.getThreadName()
      + "\""
      +
      " Id="
      + thread.getThreadId()
      + " "
      +
      thread.getThreadState());
    if (thread.getLockName() != null) {
      builder.append(" on " + thread.getLockName());
    }
    if (thread.getLockOwnerName() != null) {
      builder.append(" owned by \"" + thread.getLockOwnerName() +
        "\" Id=" + thread.getLockOwnerId());
    }
    if (thread.isSuspended()) {
      builder.append(" (suspended)");
    }
    if (thread.isInNative()) {
      builder.append(" (in native)");
    }
    builder.append('\n');

    StackTraceElement[] stackTrace = thread.getStackTrace();
    for (int i = 0; i < stackTrace.length; i++) {
      StackTraceElement trace = stackTrace[i];
      builder.append("\tat " + trace.toString());
      builder.append('\n');
      if (i == 0 && thread.getLockInfo() != null) {
        Thread.State ts = thread.getThreadState();
        switch (ts) {
        case BLOCKED:
          builder.append("\t-  blocked on " + thread.getLockInfo());
          builder.append('\n');
          break;
        case WAITING:
          builder.append("\t-  waiting on " + thread.getLockInfo());
          builder.append('\n');
          break;
        case TIMED_WAITING:
          builder.append("\t-  waiting on " + thread.getLockInfo());
          builder.append('\n');
          break;
        default:
        }
      }

      for (MonitorInfo monitor : thread.getLockedMonitors()) {
        if (monitor.getLockedStackDepth() == i) {
          builder.append("\t-  locked " + monitor);
          builder.append('\n');
        }
      }
    }

    LockInfo[] locks = thread.getLockedSynchronizers();
    if (locks.length > 0) {
      builder.append("\n\tNumber of locked synchronizers = " + locks.length);
      builder.append('\n');
      for (LockInfo lock : locks) {
        builder.append("\t- " + lock);
        builder.append('\n');
      }
    }
    builder.append('\n');

    return builder.toString();
  }

  /**
   * Checks for a JVM deadlock (only available on JDK 1.5)
   */
  private boolean checkJVMDeadlock()
    throws Exception
  {
    if (_mbeanServer == null)
      return true;

    long []deadlockedThreads =
      (long []) _mbeanServer.invoke(_threadMBean,
				    "findMonitorDeadlockedThreads",
				    new Object[0], new String[0]);

    if (deadlockedThreads == null)
      return true;

    try {
      log.severe("JDK detected deadlock. Restarting Resin.");

      for (int i = 0; i < deadlockedThreads.length; i++) {
	Long id = new Long(deadlockedThreads[i]);
	Integer maxDepth = new Integer(Integer.MAX_VALUE);
	
	Object info = _mbeanServer.invoke(_threadMBean,
					  "getThreadInfo",
					  new Object[] { id, maxDepth },
					  new String[] {"long", "int"});

	// log.severe(threadInfoToString(info));
	log.severe(String.valueOf(info));
      }

      return true;
    } catch (Exception e) {
      e.printStackTrace();
    } finally {
      Runtime.getRuntime().halt(1);
    }

    return false;
  }

  /**
   * Prints the thread info.
   */
  private String threadInfoToString(Object info)
    throws NoSuchMethodException,
	   IllegalAccessException,
	   InvocationTargetException
  {
    if (info == null)
      return "null";

    String s = info.toString();
    
    // implemented by reflection since we can't assume JDK 1.5
    
    Class cl = info.getClass();

    Method m = cl.getMethod("getLockName", new Class[0]);
    String lockName = (String) m.invoke(info, new Object[0]);
    if (lockName != null)
      s += "\n  waiting for " + lockName;
    
    m = cl.getMethod("getLockOwnerName", new Class[0]);
    String lockOwnerName = (String) m.invoke(info, new Object[0]);
    if (lockOwnerName != null)
      s += "\n  owned by " + lockOwnerName;

    m = cl.getMethod("getStackTrace", new Class[0]);
    StackTraceElement[] stackTrace;
    stackTrace = (StackTraceElement []) m.invoke(info, new Object[0]);

    if (stackTrace != null) {
      for (int i = 0; i < stackTrace.length; i++) {
	s += "\n\t" + stackTrace[i];
      }
    }

    return s;
  }
  
  /**
   * Pings the specified URLs.
   */
  private boolean pingURLs()
    throws Exception
  {
    boolean isValid = true;
    
    for (int i = _pathList.size() - 1; i >= 0; i--) {
      Path url = _pathList.get(i);

      boolean isOkay = false;
      for (int j = 0; ! isOkay && j < _tryCount; j++) {
	String status = null;

	try {
	  isOkay = checkPing(url, j);
	} catch (Throwable e) {
	  log.log(Level.WARNING, e.toString(), e);
	}

	if (! isOkay) {
	  Thread.currentThread().sleep(_retryTime);
	}
      }

      if (! isOkay) {
	isValid = false;
	pingFailed(url);
      }
    }

    return isValid;
  }

  /**
   * Called to check if the connection is alive.  An custom
   * class might override checkPing if it wanted to handle a different
   * status.
   *
   * @param url the url to test
   * @param count the number of times the url has been checked without success.
   *
   * @return true if the ping succeeds
   */
  protected boolean checkPing(Path url, int count)
    throws Exception
  {

    ReadStream is = url.openRead();
    
    if (_socketTimeout > 0)
      is.setAttribute("socket-timeout",
		      new Integer((int) _socketTimeout));
    
    String status;
    try {
      status = (String) is.getAttribute("status");

      if (log.isLoggable(Level.FINE))
        log.fine("ping " + url + " " + status);

      if (status != null && status.equals("200"))
        return true;
      else {
        pingFailedOnce(url, count, is);
        return false;
      }
    } finally {
      is.close();
    }
  }

  /**
   * Called if the ping fails once.
   */
  protected void pingFailedOnce(Path url, int count, ReadStream is)
    throws Exception
  {
    String status = (String) is.getAttribute("status");

    CharBuffer cb = new CharBuffer();
    int ch;
    while ((ch = is.read()) >= 0) {
      cb.append((char) ch);
    }
    

    if (log.isLoggable(Level.FINE)) {
      log.fine(L.l("ping {0} failed ({1}) status: {2}\n{3}",
		      String.valueOf(url), String.valueOf(count),
		      status, cb));
    }
    log.warning(L.l("ping {0} failed ({1}) status: {2}",
		      String.valueOf(url), String.valueOf(count),
		      status));
  }

  /**
   * Called if the ping completely fails.
   */
  protected void pingFailed(Path url)
    throws Exception
  {
    _isFailed = true;
    
    log.severe(L.l("pinging {0} failed.  Shutting down.", url));

    ThreadDump.dumpThreads();

    // shutdown the server
    _loader.destroy();
    System.exit(1);
  }

  /**
   * Stop the pinger.
   */
  public void stop()
  {
    log.fine("ping stopping");
    if (! _isFailed)
      _isClosed = true;
    _alarm.dequeue();
  }

  public String toString()
  {
    return "PingThread" + _urlList;
  }

  static {
    try {
      _threadMBean = new ObjectName("java.lang:type=Threading");
    } catch (Throwable e) {
      throw new RuntimeException(e);
    }
  }
}
