/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.admin;

import java.util.logging.*;

import com.caucho.config.types.Period;
import com.caucho.lifecycle.*;
import com.caucho.util.*;
import com.caucho.server.util.*;

/**
 * statistics
 */
public class StatService implements Runnable
{
  private static Logger log
    = Logger.getLogger(StatService.class.getName());
  private static L10N L = new L10N(StatService.class);

  private ProManagement _management;
  private Lifecycle _lifecycle = new Lifecycle();
  private long _samplePeriod = 60 * 1000;

  private Object _lock = new Object();
  private Thread _thread;

  private double _cpuLoadLogInfoThreshold = 1.0;
  private double _cpuLoadLogWarningThreshold = 5.0;
  private double _cpuLoadThreadDumpThreshold = 5.0;
  private double _cpuLoadExitThreshold = 100.0;
  
  private long _threadDumpInterval = 15 * 60 * 1000;
  private long _lastThreadDump;

  StatService(ProManagement management)
  {
    _management = management;
  }

  /**
   * Sets the cpu load threshold to generate an info log entry
   */
  public void setCpuLoadLogInfoThreshold(double threshold)
  {
    _cpuLoadLogInfoThreshold = threshold;
  }

  /**
   * Sets the cpu load threshold to generate a warning log entry
   */
  public void setCpuLoadLogWarningThreshold(double threshold)
  {
    _cpuLoadLogWarningThreshold = threshold;
  }

  /**
   * Sets the cpu load threshold to generate a thread dump
   */
  public void setCpuLoadThreadDumpThreshold(double threshold)
  {
    _cpuLoadThreadDumpThreshold = threshold;
  }

  /**
   * Sets the cpu load threshold to exit the server
   */
  public void setCpuLoadExitThreshold(double threshold)
  {
    _cpuLoadExitThreshold = threshold;
  }

  /**
   * Sets the statistics threshold
   */
  public void setSamplePeriod(Period period)
  {
    _samplePeriod = period.getPeriod();
  }

  /**
   * Sets the minimum time between thread dumps
   */
  public void setThreadDumpInterval(Period period)
  {
    _threadDumpInterval = period.getPeriod();
  }

  void start()
  {
    if (! _lifecycle.toActive())
      return;

    Thread thread = new Thread(this, "resin-pro-statistics");
    thread.setDaemon(true);
    thread.start();
  }

  void stop()
  {
    if (! _lifecycle.toDestroy())
      return;

    _management = null;
  }

  public void run()
  {
    while (_lifecycle.isActive() && _samplePeriod > 0) {
      try {
	double cpuLoad = CauchoSystem.getLoadAvg();

	if (Alarm.isTest() || cpuLoad <= 0) {
	}
	else if (_cpuLoadLogWarningThreshold < cpuLoad)
	  log.warning(this + " cpu-load=" + cpuLoad);
	else if (_cpuLoadLogInfoThreshold < cpuLoad)
	  log.info(this + " cpu-load=" + cpuLoad);
	else if (log.isLoggable(Level.FINER))
	  log.finer(this + " cpu-load=" + cpuLoad);

	if (_cpuLoadExitThreshold < cpuLoad) {
	  log.warning(this + " Resin exiting due to CPU load=" + cpuLoad);
	  System.exit(1);
	}

	if (_cpuLoadThreadDumpThreshold < cpuLoad
	    && _lastThreadDump + _threadDumpInterval < System.currentTimeMillis()) {
	  _lastThreadDump = System.currentTimeMillis();
	  log.info(this + " dumping threads due to cpu load=" + cpuLoad);
	  ThreadDump.dumpThreads();
	}
	
	long now = System.currentTimeMillis();
	long mod = now % _samplePeriod;

	synchronized (_lifecycle) {
	  _lifecycle.wait(_samplePeriod - now % _samplePeriod);
	}
      } catch (Exception e) {
	log.log(Level.FINER, e.toString(), e);
      }
    }
  }

  public String toString()
  {
    return getClass().getSimpleName() + "[]";
  }
}
