/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cache;

import com.caucho.config.ConfigException;
import com.caucho.config.types.Bytes;
import com.caucho.db.Database;
import com.caucho.db.store.BlockManager;
import com.caucho.db.store.Store;
import com.caucho.license.LicenseCheck;
import com.caucho.loader.CloseListener;
import com.caucho.loader.Environment;
import com.caucho.loader.EnvironmentLocal;
import com.caucho.server.webapp.*;
import com.caucho.server.resin.Resin;
import com.caucho.server.dispatch.*;
import com.caucho.server.cluster.Server;
import com.caucho.management.server.CacheItem;
import com.caucho.util.L10N;
import com.caucho.util.Log;
import com.caucho.util.LruCache;
import com.caucho.vfs.MemoryPath;
import com.caucho.vfs.Path;
import com.caucho.vfs.Vfs;

import javax.annotation.PostConstruct;
import javax.servlet.FilterChain;
import java.io.IOException;
import java.util.*;
import java.util.regex.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Cached response.
 */
public class Cache extends AbstractCache
{
  private static final L10N L = new L10N(Cache.class);
  private static final Logger log
    = Logger.getLogger(Cache.class.getName());

  private static EnvironmentLocal<Cache> _cacheLocal =
    new EnvironmentLocal<Cache>("caucho.server.cache");

  private LruCache<CacheFilterChain,CacheFilterChain> _cacheLruEntries;

  private Path _path;
  private Database _database;
  private Store _store;
  private boolean _isEnabled = true;
  private boolean _enableRange = true;
  private boolean _isRewriteVaryAsPrivate = false;
  private long _memorySize = 10 * 1024 * 1024;
  private long _diskSize = 1024 * 1024 * 1024;

  private int _maxEntrySize = 1 * 1024 * 1024;
  private int _entries = 8192;


  private Resin _resin;
  private CacheAdmin _admin;

  private volatile long _hitCount;
  private volatile long _missCount;

  public Cache()
    throws ConfigException
  {
    try {
      Class cl = Class.forName("com.caucho.license.LicenseCheckImpl");
      LicenseCheck license = (LicenseCheck) cl.newInstance();

      license.validate(1);
    } catch (ConfigException e) {
      throw e;
    } catch (Exception e) {
      log.log(Level.FINE, e.toString(), e);
      
      throw new ConfigException(L.l("<cache> requires a valid Resin Professional license.\nSee http://www.caucho.com for information on licensing."));
    }

    _resin = Resin.getLocal();
    _cacheLocal.set(this);

    _admin = new CacheAdmin(this);

    Environment.addClassLoaderListener(new CloseListener(this));
  }

  /**
   * Returns the local cache.
   */
  public static Cache getLocalCache()
  {
    return _cacheLocal.get();
  }

  /**
   * Sets the path to the cache directory.
   */
  public void setPath(Path path)
  {
    _path = path;
  }

  /**
   * Returns the path from the cache directory.
   */
  public Path getPath()
  {
    return _path;
  }

  /**
   * Sets the memory size of the cache
   */
  public void setMemorySize(Bytes size)
  {
    _memorySize = size.getBytes();
  }

  /**
   * Sets the disk size of the cache
   */
  public void setDiskSize(Bytes size)
  {
    _diskSize = size.getBytes();
  }

  /**
   * Sets the maximum entry size for cache values.
   */
  public void setMaxEntrySize(Bytes size)
  {
    _maxEntrySize = (int) size.getBytes();
  }

  /**
   * Sets the maximum entry size for cache values.
   */
  public int getMaxEntrySize()
  {
    return _maxEntrySize;
  }

  /**
   * Set true if enabled.
   */
  public void setEnable(boolean isEnabled)
  {
    _isEnabled = isEnabled;
  }

  /**
   * Return true if enabled.
   */
  public boolean isEnable()
  {
    return _isEnabled;
  }

  /**
   * Sets the max number of entries.
   */
  public void setEntries(int entries)
  {
    _entries = entries;
  }

  /**
   * Sets the path to the cache directory (backwards compatibility).
   */
  public void setDir(Path path)
  {
    setPath(path);
  }

  /**
   * Sets the size of the the cache (backwards compatibility).
   */
  public void setSize(Bytes size)
  {
    setMemorySize(size);
  }

  /**
   * True if range handling is enabled.
   */
  public void setEnableRange(boolean enableRange)
  {
    _enableRange = enableRange;
  }

  /**
   * True if range handling is enabled.
   */
  public boolean isEnableRange()
  {
    return _enableRange;
  }

  /**
   * True if Vary headers should be rewritten as Cache-Control: private to
   * work around IE's Vary handling bug.
   */
  public void setRewriteVaryAsPrivate(boolean rewrite)
  {
    _isRewriteVaryAsPrivate = rewrite;
  }

  /**
   * True if Vary headers should be rewritten as Cache-Control: private to
   * work around IE's Vary handling bug.
   */
  public boolean isRewriteVaryAsPrivate()
  {
    return _isRewriteVaryAsPrivate;
  }

  /**
   * Initialize the cache.
   */
  @PostConstruct
  public void init()
    throws ConfigException, IOException, java.sql.SQLException
  {
    if (! _isEnabled || _entries <= 0)
      return;
    
    if (_path == null) {
      _path = Vfs.lookup("cache");

      // special cased for testing
      if (_path instanceof MemoryPath)
	_path = Vfs.lookup("file:/tmp/caucho/qa/cache");
    }

    try {
      _path.mkdirs();
    } catch (Throwable e) {
    }

    if (! _path.isDirectory())
      throw new ConfigException(L.l("Cache path '{0}' must be a directory.",
				    _path));

    String server = (String) Environment.getAttribute("caucho.server-id");

    if (server == null || server.equals(""))
      server = "cache";
      
    Path path = _path.lookup(server + ".db");

    try {
      if (path.exists())
	path.remove();
    } catch (Exception e) {
      log.log(Level.FINE, e.toString(), e);
    }

    if (path.exists()) {
      throw new ConfigException(L.l("Proxy Cache at '{0}' cannot be removed on startup.  The proxy cache must be removable on Resin start.",
				path.getNativePath()));
    }

    log.info("Proxy Cache disk-size=" + (_diskSize / (1024 * 1024)) + "M" +
	     " memory-size=" + (_memorySize / (1024 * 1024)) + "M");

    _cacheLruEntries
      = new LruCache<CacheFilterChain,CacheFilterChain>(_entries);

    _database = new Database();
    _database.ensureMemoryCapacity(_memorySize);
    _database.init();
    
    _store = new Store(_database, "proxy-cache", null, path);
    _store.setFlushDirtyBlocksOnCommit(false);
    _store.create();
  }

  /**
   * Creates the filter.
   */
  public FilterChain createFilterChain(FilterChain next, WebApp app)
  {
    if (isEnable())
      return new CacheFilterChain(this, next, app);
    else
      return next;
  }

  /**
   * Mark the cache filter chain as active
   */
  void activateEntry(CacheFilterChain chain)
  {
    _cacheLruEntries.putIfNew(chain, chain);

    // remove overfull entries
    for (int i = 64;
	 i > 0
	   && _diskSize <= _store.getTotalFragmentSize()
	   && _cacheLruEntries.removeLongestTail();
	 i--) {
    }
  }

  /**
   * Mark the cache filter chain as active
   */
  boolean useEntry(CacheFilterChain chain)
  {
    return _cacheLruEntries.get(chain) != null;
  }

  /**
   * Clears the cache.
   */
  public void clear()
  {
    if (_cacheLruEntries != null)
      _cacheLruEntries.clear();
  }

  /**
   * Clears the cache by a pattern.
   */
  void clearCacheByPattern(Pattern hostRegexp, Pattern urlRegexp)
  {
    ArrayList<CacheFilterChain> clearEntries
      = new ArrayList<CacheFilterChain>();

    synchronized (_cacheLruEntries) {
      Iterator<CacheFilterChain> iter = _cacheLruEntries.keys();
      while (iter.hasNext()) {
	CacheFilterChain entry = iter.next();
	
	boolean isMatch = true;
	
	if (hostRegexp != null) {
	  WebApp webApp = entry.getWebApp();
	  if (! hostRegexp.matcher(webApp.getHostName()).find())
	    isMatch = false;
	}

	if (urlRegexp != null) {
	  if (! urlRegexp.matcher(entry.getUri()).find())
	    isMatch = false;
	}

	if (isMatch)
	  clearEntries.add(entry);
      }
    }

    for (int i = 0; i < clearEntries.size(); i++)
      _cacheLruEntries.remove(clearEntries.get(i));
  }

  /**
   * Clears the cache expires
   */
  void clearExpires()
  {
    ArrayList<CacheFilterChain> clearEntries
      = new ArrayList<CacheFilterChain>();

    synchronized (_cacheLruEntries) {
      Iterator<CacheFilterChain> iter = _cacheLruEntries.keys();
      while (iter.hasNext()) {
	CacheFilterChain entry = iter.next();

	entry.clearExpires();
      }
    }
  }

  /**
   * Creates a new writing inode.
   */
  public CacheInode createInode()
  {
    return new CacheInode(_store);
  }

  /**
   * Returns the hit count.
   */
  public long getHitCount()
  {
    return _hitCount;
  }

  /**
   * Returns the miss count.
   */
  public long getMissCount()
  {
    return _missCount;
  }

  /**
   * Returns the block hit count.
   */
  public long getMemoryBlockHitCount()
  {
    return BlockManager.create(1).getHitCountTotal();
  }

  /**
   * Returns the miss count.
   */
  public long getMemoryBlockMissCount()
  {
    return BlockManager.create(1).getMissCountTotal();
  }

  /**
   * Return most used connections.
   */
  public CacheItem []getCacheableEntries(int max)
  {
    ArrayList<CacheItem> items = getItems();

    Collections.sort(items, new CacheHitCompare());

    ArrayList<CacheItem> cacheableItems = new ArrayList<CacheItem>();

    int i;
    for (i = 0; i < items.size(); i++) {
      CacheItem item = items.get(i);
      
      if (item.isCacheable()) {
	cacheableItems.add(item);
      }
    }

    CacheItem []itemArray = new CacheItem[cacheableItems.size()];
    cacheableItems.toArray(itemArray);

    return itemArray;
  }

  /**
   * Return most used connections.
   */
  public CacheItem []getUncacheableEntries(int max)
  {
    ArrayList<CacheItem> items = getItems();

    Collections.sort(items, new CacheMissCompare());

    ArrayList<CacheItem> uncacheableItems = new ArrayList<CacheItem>();

    int i;
    for (i = 0; i < items.size(); i++) {
      CacheItem item = items.get(i);
      
      if (! item.isCacheable()) {
	uncacheableItems.add(item);
      }
    }

    CacheItem []itemArray = new CacheItem[uncacheableItems.size()];
    uncacheableItems.toArray(itemArray);

    return itemArray;
  }

  Server getServer()
  {
    return _resin.getServer();
  }

  private ArrayList<CacheItem> getItems()
  {
    Server server = _resin.getServer();

    ArrayList<Invocation> invocations = server.getInvocations();

    HashMap<String,CacheItem> itemMap = new HashMap<String,CacheItem>();

    for (int i = 0; i < invocations.size(); i++) {
      Invocation inv = invocations.get(i);

      String url = inv.getURI();

      int p = url.indexOf('?');
      if (p > 0)
        url = url.substring(0, p);

      CacheItem item = itemMap.get(url);

      if (item == null) {
        item = new CacheItem();
        item.setUrl(url);

        itemMap.put(url, item);
      }

      boolean isCacheable = false;
      long hitCount = 0;
      
      FilterChain chain = inv.getFilterChain();
      if (chain instanceof WebAppFilterChain) {
        WebAppFilterChain webAppChain = (WebAppFilterChain) chain;
        FilterChain next = webAppChain.getNext();

        if (next instanceof CacheFilterChain) {
	  CacheFilterChain cacheChain = (CacheFilterChain) next;
	  
          isCacheable = cacheChain.isCacheable();
	  hitCount = cacheChain.getHitCount();
	}
      }

      item.setHitCount(hitCount);
      item.setMissCount(inv.getRequestCount() - hitCount);
      
      if (isCacheable) {
	item.setCacheable(true);
      }
    }

    return new ArrayList<CacheItem>(itemMap.values());
  }

  /**
   * Adds a hit.
   */
  final void hit()
  {
    _hitCount++;
  }

  /**
   * Adds a miss.
   */
  final void miss()
  {
    _missCount++;
  }

  /**
   * Closes the cache.
   */
  public void destroy()
  {
    Store store = _store;
    _store = null;

    Database database = _database;
    _database = null;

    if (store != null)
      store.close();

    if (database != null)
      database.close();
  }

  private static class CacheHitCompare implements Comparator<CacheItem>
  {
    public int compare(CacheItem a, CacheItem b)
    {
      long diff = b.getHitCount() - a.getHitCount();

      if (diff < 0)
        return -1;
      else if (diff == 0)
        return 0;
      else
        return 1;
    }
  }

  private static class CacheMissCompare implements Comparator<CacheItem>
  {
    public int compare(CacheItem a, CacheItem b)
    {
      long diff = b.getMissCount() - a.getMissCount();

      if (diff < 0)
        return -1;
      else if (diff == 0)
        return 0;
      else
        return 1;
    }
  }
}
