/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cache;

import com.caucho.util.Alarm;

import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import java.io.*;

/**
 * Cached response.
 */
public class CacheEntry extends AbstractCacheEntry {
  private static final Logger log
    = Logger.getLogger(CacheEntry.class.getName());

  Cache _cache;

  // Next link for the vary
  CacheEntry _next;

  // path to the stored data
  CacheInode _inode;
  
  // saved header keys
  ArrayList<String> _headerKeys;
  // saved header values
  ArrayList<String> _headerValues;
  // saved content type
  String _contentType;
  // saved content encoding
  String _charEncoding;
  // saved etag
  String _etag;
  // saved last-modified
  String _lastModified;
  // saved mime-type
  String _mimeType;
  // cache length
  long _contentLength;
  // true if it allows range
  boolean _allowRange = true;

  // vary headers
  ArrayList<String> _vary;
  // vary crc
  long _varyCrc;

  // true if the cache depends on cookies
  boolean _varyCookies;
  // the cookie the request depends on
  String _varyCookie;

  long _resinMaxAge;
  long _maxAge = Long.MIN_VALUE;
  long _sMaxAge = Long.MIN_VALUE;
  
  long _expireDate;
  String _expireString;
  // true if the user set the expires
  boolean _hasExpires;
  boolean _isChar;
    
  // true if the entry is complete and valid
  boolean _isValid;

  CacheEntry(Cache cache)
  {
    _cache = cache;

    _allowRange = cache.isEnableRange();
  }
  
  /**
   * If the response discovers that the page hasn't changed,
   * update the cache time.
   */
  public void updateExpiresDate()
  {
    _expireDate = Alarm.getCurrentTime() + 5000;
  }

  /**
   * Clears the expires time
   */
  public void clearExpires()
  {
    _expireDate = 0;
  }

  /**
   * Returns the current inode.
   */
  CacheInode getInode()
  {
    return _inode;
  }

  /**
   * Opens the inode for byte writing.
   */
  public OutputStream openOutputStream()
  {
    _isChar = false;

    return _inode.openOutputStream();
  }

  /**
   * Opens the inode for writing.
   */
  public Writer openWriter()
  {
    _isChar = true;

    return _inode.openWriter();
  }

  /**
   * Sets the current inode.
   */
  void setInode(CacheInode inode)
  {
    if (_inode != null)
      throw new IllegalStateException();
    
    _inode = inode;
  }

  void remove()
  {
    _isValid = false;

    CacheInode inode = _inode;
    _inode = null;

    if (inode != null) {
      try {
	inode.free();
      } catch (Exception e) {
	_cache.setEnable(false);
	
	log.log(Level.WARNING, e.toString(), e);
      }
    }
  }

  public void destroy()
  {
    remove();
  }
}
