/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cache;

import com.caucho.server.connection.AbstractHttpRequest;
import com.caucho.server.connection.AbstractHttpResponse;
import com.caucho.server.connection.CauchoRequest;
import com.caucho.server.dispatch.AbstractFilterChain;
import com.caucho.server.webapp.WebApp;
import com.caucho.servlet.comet.CometFilterChain;
import com.caucho.util.*;
import com.caucho.vfs.ClientDisconnectException;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Represents the final servlet in a filter chain.
 */
public class CacheFilterChain
  extends AbstractCacheFilterChain implements CacheListener
{
  private static final L10N L = new L10N(CacheFilterChain.class);
  private static final Logger log
    = Logger.getLogger(CacheFilterChain.class.getName());

  private static long DEFAULT_EXPIRES = 60000L;
  private static long MIN_EXPIRES = 5000;
  
  private static int MAX_VARY = 8;

  private static QDate _calendar = new QDate();

  private Cache _cache;

  private FilterChain _next;

  // is this page cacheable
  private boolean _isCacheable = true;
  // time the cache was last checked
  private long _lastCacheCheck;

  // the cached entry
  private CacheEntry _entry;

  // thread filling the cache
  private Thread _cacheThread = null;

  private WebApp _webApp;

  // uri/query string for logging
  private String _uri;
  private String _queryString;

  private boolean _isTop = true;
  private boolean _isClosed;

  private long _hitCount;

  /**
   * Create the filter chain servlet.
   */
  public CacheFilterChain(FilterChain next,
			  WebApp webApp)
  {
    this(Cache.getLocalCache(), next, webApp);
  }
  
  /**
   * Create the filter chain servlet.
   */
  public CacheFilterChain(Cache cache,
			  FilterChain next,
			  WebApp webApp)
  {
    _cache = cache;
    
    _next = next;
    _webApp = webApp;
  }

  /**
   * True for top-level requests.
   */
  public void setTop(boolean isTop)
  {
    _isTop = isTop;
  }

  /**
   * True for top-level requests.
   */
  public boolean isTop(boolean isTop)
  {
    return _isTop;
  }
  
  public long getHitCount()
  {
    return _hitCount;
  }

  public WebApp getWebApp()
  {
    return _webApp;
  }

  public String getUri()
  {
    return _uri;
  }
  
  /**
   * Clears the expires timers
   */
  public void clearExpires()
  {
    synchronized (this) {
      for (CacheEntry entry = _entry; entry != null; entry = entry._next) {
	entry.clearExpires();
      }
    }
  }
  
  /**
   * Handles caching for the filter.
   *
   * @param req the servlet request
   * @param res the servlet response
   */
  public void doFilter(ServletRequest request, ServletResponse response)
    throws ServletException, IOException
  {
    if (! isCacheable()) {
      _next.doFilter(request, response);
      return;
    }
    else if (! (request instanceof CauchoRequest)
	     || ! (response instanceof AbstractHttpResponse)) {
      _next.doFilter(request, response);
      return;
    }
    

    CauchoRequest req = (CauchoRequest) request;
    
    if (! req.getMethod().equals("GET")
	&& ! req.getMethod().equals("HEAD")) {
      _next.doFilter(request, response);
      return;
    }
    
    AbstractHttpResponse res = (AbstractHttpResponse) response;

    long now = Alarm.getCurrentTime();
    boolean isCaching = false;

    // If we've got a cached value, try to use it
    if (_entry != null && useCache(req, res, _isTop)) {
      _cache.hit();
      return;
    }
    else if (req != res.getOriginalRequest()) {
    }
    else if (req.getWebApp() == null || ! req.getWebApp().isActive() ) {
      // web-app starting: #2474
    }
    else if (_isCacheable || _lastCacheCheck + 60000 < now) {
      isCaching = true;
      res.setCacheInvocation(this);
    }
    
    _cache.miss();
    
    Thread currentThread = Thread.currentThread();

    try {
      // If the request caches its results, it will call startCaching.
      _next.doFilter(req, res);

      // If we're caching, force a flush
      res.finish();
    } finally {
      if (_cacheThread == currentThread) {
	_cacheThread = null;
	res.killCaching();
      }

      if (isCaching)
	res.setCacheInvocation(null);
    }
  }
  
  /**
   * Resumes the request.
   *
   * @param request the servlet request
   * @param response the servlet response
   *
   * @since Resin 3.1.3
   */
  @Override
  public boolean doResume(ServletRequest request,
			  ServletResponse response)
    throws ServletException, IOException
  {
    if (_next instanceof CometFilterChain) {
      CometFilterChain next = (CometFilterChain) _next;

      return next.doResume(request, response);
    }
    else
      return false;
  }

  /**
   * Returns true if the entry is cacheable.
   */
  final boolean isCacheable()
  {
    if (! _cache.isEnable())
      return false;
    else if (_isCacheable)
      return true;
    else if (_lastCacheCheck + 60000 < Alarm.getCurrentTime())
      return true;
    else {
      _cache.miss();
      
      return false;
    }
  }

  /**
   * Tries to use the cache.  If the response hasn't expired,
   * send it directly to the user.  Otherwise, set If-Modified-Since and
   * If-None-Match so the servlet can bypass processing.
   *
   * @param req the servlet request
   * @param res the servlet response
   *
   * @return true if the request has completed because the cache has filled
   *   the request
   */
  boolean useCache(CauchoRequest req,
                   AbstractHttpResponse res,
                   boolean isTop)
    throws IOException, ServletException
  {
    CacheEntry entry = _entry;

    // If the entry isn't valid, don't use it
    if (entry == null)
      return false;

    // Mark this entry as used for the LRU
    if (! _cache.useEntry(this))
      return false;
    
    entry = getVaryEntry(entry, req);

    // If the entry isn't valid, don't use it
    if (entry == null || ! entry._isValid)
      return false;
    
    WebApp webApp = _webApp;

    boolean varyCookies = entry._varyCookies;
    String varyCookie = entry._varyCookie;

    // Check if anonymous caching is allowed
    if (entry._varyCookies &&
	(req.getHeader("Cookie") != null ||  
	 req.isRequestedSessionIdFromURL())) {
      return false;
    }

    long now = Alarm.getCurrentTime();

    // If the webApp is reloading, can't use the cache
    if (_webApp.isModified())
      return false;

    // If the entry is still good, send it back to the browser
    Thread cacheThread = _cacheThread;
    if ((now <= entry._expireDate
	 || cacheThread != null && cacheThread.isAlive())
	&& fillFromCache(req, res, entry, isTop)) {
      
      return true;
    }

    if (! entry._isValid)
      return false;
    
    // If the user set the expires, we need to go back so it can be
    // reset
    if (entry._hasExpires)
      return false;
    
    // If it's a range request, then must go through
    if (! entry._allowRange)
      return false;

    // If the servlet set the caching tags on its cached response,
    // set them for our request.
    String oldIfMatch = null;
      
    // DispatchRequest must ignore the If-None-Match header
    // in the original request
    if (entry._etag != null) {
      oldIfMatch = req.getHeader("If-None-Match");
      req.setHeader("If-None-Match", entry._etag);

      // If the user's request set the headers, set the response so
      // it can send the result directly
      if (oldIfMatch != null) {
        int p = oldIfMatch.indexOf(';');
        if (p > 0)
          oldIfMatch = oldIfMatch.substring(0, p);
        if (isTop && oldIfMatch.equals(entry._etag))
          res.setTopCache(true);
      }
    }

    // DispatchRequest must ignore the If-Modified-Since header
    // in the original request
    if (oldIfMatch == null && entry._lastModified != null) {
      String oldIfModifiedSince = req.getHeader("If-Modified-Since");
      req.setHeader("If-Modified-Since", entry._lastModified);

      // If the user's request set the headers, set the response so
      // it can send the result directly
      if (oldIfModifiedSince != null) {
        int plus = oldIfModifiedSince.indexOf('+');
        int semicolon = oldIfModifiedSince.indexOf(';');

	if (plus > 0)
          oldIfModifiedSince = oldIfModifiedSince.substring(0, plus);
        else if (semicolon > 0)
          oldIfModifiedSince = oldIfModifiedSince.substring(0, semicolon);
	
        if (isTop && entry._lastModified.startsWith(oldIfModifiedSince))
          res.setTopCache(true);
      }
    }

    // Point the response to the cache entry, so if the servlet returns
    // not modified, we can use the cache
    res.setMatchCacheEntry(entry);

    return false;
  }

  /**
   * fillFromCache is called when the client needs the entire result, and
   * the result is already in the cache.
   *
   * @param req the servlet request trying to get data from the cache
   * @param response the servlet response which will receive data
   * @param entry the cache entry to use
   * @param isTop if true, the not-modified should be sent to the browser
   */
  public boolean fillFromCache(CauchoRequest req,
                               AbstractHttpResponse response,
                               AbstractCacheEntry abstractEntry,
                               boolean isTop)
    throws IOException
  {
    CacheEntry entry = (CacheEntry) abstractEntry;
    
    if (! entry._isValid)
      return false;

    try {
      response.reset();
    } catch (Throwable e) {
      return false;
    }
    
    response.disableCaching(true);

    boolean notModified = false;

    // top requests can return NOT_MODIFIED.  Included requests must
    // return the full response
    if (isTop) {
      String key;
      if (entry._etag != null) {
        key = req.getHeader("if-none-match");
        if (key != null) {
          int i = key.indexOf(';');
          if (i >= 0)
            key = key.substring(0, i);
          if (key != null && key.equals(entry._etag)) {
            response.setStatus(response.SC_NOT_MODIFIED);
            notModified = true;
          }
        }
      }
      
      if (! notModified && entry._lastModified != null) {
        key = req.getHeader("if-modified-since");

	if (key != null) {
	  int semicolon = key.indexOf(';');
	  int plus = key.indexOf('+');
	  
	  if (plus > 0)
	    key = key.substring(0, plus);
	  else if (semicolon > 0)
	    key = key.substring(0, semicolon);

	  if (entry._lastModified.startsWith(key)) {
	    response.setStatus(response.SC_NOT_MODIFIED);
	    notModified = true;
	  }
	}
      }
    }

    String range = req.getHeader("Range");

    if (! entry._allowRange && range != null)
      return false;

    _hitCount++;

    if (log.isLoggable(Level.FINE)) {
      String debugURI = _uri;
      
      if (_queryString != null)
	debugURI = debugURI + "?" + _queryString;
      
      if (notModified)
        log.fine("not-modified: " + debugURI);
      else
        log.fine("using cache: " + debugURI);
    }

    long now = Alarm.getCurrentTime();

    /*
    if (entry._maxAge <= 0) {
      fillMaxAge(entry, _uri, req, response);
    }
    */

    if (notModified) {
      if (entry._etag != null)
	response.setHeader("ETag", entry._etag);
      
      if (entry._maxAge > 0) {
        response.setDateHeader("Expires", now + entry._maxAge);
	entry._expireDate = now + entry._maxAge;
      }
      
      return true;
    }

    if (entry._contentType != null)
      response.setContentType(entry._contentType);

    if (entry._charEncoding != null)
      response.setCharacterEncoding(entry._charEncoding);

    ArrayList<String> headerKeys = entry._headerKeys;
    ArrayList<String> headerValues = entry._headerValues;
    for (int i = 0; headerKeys != null && i < headerKeys.size(); i++) {
      response.addHeader(headerKeys.get(i), headerValues.get(i));
    }

    /*
    if (entry._varyCookies)
      response.setNoCache(true);
      else */
    if (entry._hasExpires)
      response.addHeader("Expires", entry._expireString);
    else if (entry._maxAge > 0) {
      // XXX: is this necessary? since the Cache-control already exists?
      // HTTP/1.0 browsers should be gone
      response.addDateHeader("Expires", now + entry._maxAge);
    }
      
    String method = req.getMethod();
    if (method != null && method.equals("HEAD")) {
      response.setContentLength((int) entry._contentLength);
      
      return true;
    }

    CacheInode inode = entry.getInode();
    if (inode == null || ! inode.allocate())
      return false;
    
    try {
      if (range != null) {
	String ifRange = req.getHeader("If-Range");

	if (ifRange != null && ! ifRange.equals(entry._etag)) {
	}
	else if (handleRange(req, response,
			     entry, inode, range,
			     entry._mimeType)) {
      
	  return true;
	}
      }

      if (entry._isChar)
	inode.writeToWriter(response.getWriter());
      else {
	response.setContentLength((int) entry._contentLength);

	inode.writeToStream(response.getResponseStream());
      }
    } catch (ClientDisconnectException e) {
      throw e;
    } catch (InterruptedIOException e) {
      throw e;
    } catch (Throwable e) {
      log.log(Level.WARNING, e.toString(), e);

      log.warning("cache disabled due to " + e.toString());
      _cache.setEnable(false);
      
      entry._isValid = false;
      
      try {
	req.killKeepalive();
        // response.reset();
      } catch (Throwable e1) {
        log.log(Level.FINE, e.toString(), e);
      }
      
      return true;
    } finally {
      inode.free();
    }

    return true;
  }

  /**
   * This is a duplicate of FileServlet, unfortunately.
   */
  private boolean handleRange(HttpServletRequest req,
                              AbstractHttpResponse res,
                              CacheEntry cache,
			      CacheInode inode,
			      String range,
			      String mime)
    throws IOException
  {
    // This is duplicated in CacheInvocation.  Possibly, it should be
    // completely removed although it's useful even without caching.
    int length = range.length();

    boolean hasMore = range.indexOf(',') > 0;

    int head = 0;
    ServletOutputStream os = res.getOutputStream();
    boolean isFirstChunk = true;
    String boundary = null;
    int off = range.indexOf("bytes=", head);

    if (off < 0)
      return false;

    off += 6;

    while (off > 0 && off < length) {
      boolean hasFirst = false;
      long first = 0;
      boolean hasLast = false;
      long last = 0;
      int ch = -1;

      // Skip whitespace
      for (; off < length && (ch = range.charAt(off)) == ' '; off++) {
      }

      // read range start (before '-')
      for (;
	   off < length && (ch = range.charAt(off)) >= '0' && ch <= '9';
	   off++) {
	first = 10 * first + ch - '0';
	hasFirst = true;
      }

      if (length <= off && ! isFirstChunk)
	break;
      else if (ch != '-')
	return false;

      // read range end (before '-')
      for (off++;
	   off < length && (ch = range.charAt(off)) >= '0' && ch <= '9';
	   off++) {
	last = 10 * last + ch - '0';
	hasLast = true;
      }

      // Skip whitespace
      for (; off < length && (ch = range.charAt(off)) == ' '; off++) {
      }

      head = off;

      long cacheLength = cache._contentLength;

      if (! hasLast) {
	if (first == 0)
	  return false;
	
	last = cacheLength - 1;
      }

      // suffix
      if (! hasFirst) {
	first = cacheLength - last;
	last = cacheLength - 1;
      }

      if (last < first)
	return false;
    
      if (cacheLength <= last) {
	// XXX: actually, an error
	return false;
      }
    
      res.setStatus(HttpServletResponse.SC_PARTIAL_CONTENT);

      CharBuffer cb = new CharBuffer();
      cb.append("bytes ");
      cb.append(first);
      cb.append('-');
      cb.append(last);
      cb.append('/');
      cb.append(cacheLength);
      String chunkRange = cb.toString();

      if (hasMore) {
	if (isFirstChunk) {
	  CharBuffer cb1 = new CharBuffer();

	  cb1.append("--");
	  Base64.encode(cb1, RandomUtil.getRandomLong());
	  boundary = cb1.toString();

	  res.setContentType("multipart/byteranges; boundary=" + boundary);
	}
	else {
	  os.write('\r');
	  os.write('\n');
	}

	isFirstChunk = false;
	
	os.write('-');
	os.write('-');
	os.print(boundary);
	os.print("\r\nContent-Type: ");
	os.print(mime);
	os.print("\r\nContent-Range: ");
	os.print(chunkRange);
	os.write('\r');
	os.write('\n');
	os.write('\r');
	os.write('\n');
      }
      else {
	res.setContentLength((int) (last - first + 1));
      
	res.addHeader("Content-Range", chunkRange);
      }

      inode.writeToStream(res.getResponseStream(), first, last - first + 1);

      for (off--; off < length && range.charAt(off) != ','; off++) {
      }

      off++;
    }

    if (hasMore) {
      os.write('\r');
      os.write('\n');
      os.write('-');
      os.write('-');
      os.print(boundary);
      os.write('-');
      os.write('-');
      os.write('\r');
      os.write('\n');
    }

    return true;
  }

  /**
   * Starts the caching after the headers have been sent.
   *
   * @param req the servlet request
   * @param req the servlet response
   * @param keys the saved header keys
   * @param values the saved header values
   * @param contentType the response content type
   * @param charEncoding the response character encoding
   *
   * @return the inode to store the cache value or null if
   *         uncacheable.
   */
  public AbstractCacheEntry startCaching(CauchoRequest req,
					 AbstractHttpResponse res,
					 ArrayList<String> keys,
					 ArrayList<String> values,
					 String contentType,
					 String charEncoding,
					 long contentLength)
  {
    long now = Alarm.getCurrentTime();

    synchronized (this) {
      if (! _isCacheable && now < _lastCacheCheck + 60000)
	return null;

      Thread cacheThread = _cacheThread;
      if (cacheThread != null && cacheThread.isAlive())
        return null;

      _lastCacheCheck = now;
      _cacheThread = Thread.currentThread();
    }

    CacheEntry oldEntry = _entry;
    
    boolean killCache = false;

    if (! "GET".equals(req.getMethod())) {
      if ("HEAD".equals(req.getMethod())) {
        _cacheThread = null;
        return null;
      }
      killCache = true;
    }

    else if (_cache.getMaxEntrySize() < contentLength)
      killCache = true;
      
    else if (res.isNoCache())
      killCache = true;

    else if (oldEntry != null
	     && oldEntry._varyCookies
	     && req.getHeader("Cookie") != null) {
      // server/13wi
      _cacheThread = null;
      return null;
    }
    else if (req.getHeader("Authorization") != null) {
      // server/2h15
      _cacheThread = null;
      return null;
    }
      
    CacheEntry entry = new CacheEntry(_cache);

    _uri = req.getPageURI();
    _queryString = req.getPageQueryString();

    boolean anonymousCaching = false;
    boolean hasCacheControl = false;
    long internalCacheTimeout = 5000L;
    long maxAge = Long.MIN_VALUE;
    long sMaxAge = Long.MIN_VALUE;

    ArrayList<String> vary = null;

    for (int i = 0; i < keys.size(); i++) {
      String key = keys.get(i);
      String value = values.get(i);

      if (killCache) {
      }
      else if (key.equalsIgnoreCase("cache-control")) {
	int length = value.length();
	int head = 0;

	while (head < length) {
	  char ch;
	  
	  for (;
	       head < length &&
		 ((ch = value.charAt(head)) == ' ' || ch == ',' || ch == ';');
	       head++) {
	  }

	  if (value.startsWith("public", head)) {
	    hasCacheControl = true;
	  }

	  else if (value.startsWith("must-revalidate", head)
		   || value.startsWith("proxy-revalidate", head)) {
	    // internalCacheTimeout = -1;
	  }
        
	  else if (value.startsWith("max-age", head)) {
	    int p = value.indexOf('=', head);

	    if (p > 0) {
	      int sign = 1;
	      int age = 0;
	      for (p++; p < value.length(); p++) {
		ch = value.charAt(p);

		if (ch == '-') {
		  sign = -1;
		}
		else if ('0' <= ch && ch <= '9') {
		  age = age * 10 + ch - '0';
		}
		else
		  break;
	      }

	      maxAge = 1000L * sign * age;

	      if (maxAge < 0)
		killCache = true;
	    }
	  }
	  else if (value.startsWith("s-maxage", head)) {
	    int p = value.indexOf('=', head);

	    if (p > 0) {
	      int sign = 1;
	      int age = 0;
	      for (p++; p < value.length(); p++) {
		ch = value.charAt(p);

		if (ch == '-') {
		  sign = -1;
		}
		else if (ch >= '0' && ch <= '9') {
		  age = age * 10 + ch - '0';
		}
		else
		  break;
	      }

	      sMaxAge = 1000L * sign * age;

	      if (sMaxAge < 0)
		killCache = true;
	    }
	  }
        
	  else {
	    hasCacheControl = true;
	    killCache = true;
	    break;
	  }

	  for (;
	       head < length
		 && (ch = value.charAt(head)) != ' ' && ch != ',' && ch != ';';
	       head++) {
	  }
	}
      }
      else if (key.equalsIgnoreCase("Vary")) {
	if (vary == null)
	  vary = new ArrayList<String>();

	int length = value.length();
	int head = 0;
	char ch = ' ';

	while (head < length) {
	  for (; head < length &&
		 ((ch = value.charAt(head)) == ' ' || ch == ',' || ch == ';');
	       head++) {
	  }
	  
	  if (length <= head)
	    break;

	  int tail = head;
	  for (; tail < length &&
		 ((ch = value.charAt(tail)) != ' ' && ch != ',' && ch != ';'
		  && ch != '=');
	       tail++) {
	  }

	  if (head < tail) {
	    String varyValue = value.substring(head, tail);

	    if (varyValue.equalsIgnoreCase("cookie")) {
	      entry._varyCookies = true;
	      
	      if (req.getHeader("Cookie") != null) {
		// server/13wj
		_cacheThread = null;
		return null;
	      }
	    }
	    else
	      vary.add(varyValue);
	  }

	  if (ch == '=') {
	    for (; tail < length &&
		   ((ch = value.charAt(tail)) != ' ' && ch != ',' && ch != ';');
		 tail++) {
	    }
	  }

	  head = tail;
	}

	if (_cache.isRewriteVaryAsPrivate() && ! entry._varyCookies) {
	  keys.set(i, "Cache-Control");
	  values.set(i, "private");
	}
      }
      else if (key.equalsIgnoreCase("Content-Encoding")) {
	entry._allowRange = false;
      }
      else if (value.equalsIgnoreCase("Set-Cookie")
	       || value.equalsIgnoreCase("Set-Cookie2")) {
	keys.remove(i);
	values.remove(i);
	i--;
      }
    }

    if (killCache) {
      synchronized (this) {
	_isCacheable = false;
	oldEntry = _entry;
	_entry = null;
	_cacheThread = null;
      }

      if (oldEntry != null)
        oldEntry.remove();

      return null;
    }

    if (anonymousCaching) {
      entry._varyCookies = req.getVaryCookies();
      entry._varyCookie = req.getVaryCookie();

      // If there's a cookie, this page doesn't cache, but let the
      // anonymous cache continue
      if (req.getHasCookie()) {
	// RSN-175
	if (! hasCacheControl)
	  res.setNoCache(true);

        _cacheThread = null;
        
        return null;
      }
    }

    entry._vary = vary;

    long crc = getVaryCRC(vary, req);

    entry._varyCrc = crc;

    long expireDate = -1;
    if (sMaxAge > 0) {
      entry._maxAge = sMaxAge;
      entry._resinMaxAge = sMaxAge;

      entry._expireDate = now + sMaxAge;
      // XXX: if adding Expires, need separate date
    }
    else if (maxAge > 0) {
      entry._maxAge = maxAge;
      entry._resinMaxAge = maxAge;
      entry._expireDate = now + maxAge;
    }

    String expires = res.getHeader("Expires");
    if (expires != null) {
      synchronized (_calendar) {
	try {
	  expireDate = _calendar.parseDate(expires);
	  entry._hasExpires = true;
	  entry._expireString = expires;
	  entry._expireDate = expireDate;

	  if (expireDate < now) {
	    synchronized (this) {
	      _isCacheable = false;
	      oldEntry = _entry;
	      _entry = null;
	      _cacheThread = null;
	    }
            
	    if (oldEntry != null)
	      oldEntry.remove();

	    return null;
	  }
	} catch (Exception e) {
	  log.log(Level.FINER, e.toString(), e);
	}
      }
      entry._expireDate = expireDate;
    }
    else if (entry._maxAge > 0)
      res.setDateHeader("Expires", entry._expireDate);

    String lastModified = res.getHeader("Last-Modified");
    entry._lastModified = lastModified;
      
    String etag = res.getHeader("ETag");
    entry._etag = etag;

    entry._mimeType = res.getHeader("Content-Type");

    // If we're caching, cache without checking for 5 seconds
    if (expireDate == -1 && (etag != null || lastModified != null)) {
      if (entry._maxAge <= 0)
	fillMaxAge(entry, _uri, req, res);

      // server/1319
      if (entry._maxAge > 0) // && ! req.getVaryCookies())
        res.setDateHeader("Expires", now + entry._maxAge);
        
      // Only cache internally for 5 seconds so we see updates
      if (entry._resinMaxAge > 0)
	entry._expireDate = now + entry._resinMaxAge;
      else if (entry._sMaxAge > 0)
	entry._expireDate = now + entry._sMaxAge;
      else if (entry._maxAge > 0)
	entry._expireDate = now + entry._maxAge;
    }

    if (entry._expireDate - 1000 < now
	&& entry._lastModified == null && entry._etag == null) {
      synchronized (this) {
	_isCacheable = false;
	oldEntry = _entry;
	_entry = null;
	_cacheThread = null;
      }
      
      if (oldEntry != null)
        oldEntry.remove();
      
      return null;
    }

    /*server/131z
    if (entry._varyCookies)
      res.setNoCache(true);
    */

    // Store the headers so cached entries will be correct
    entry._headerKeys = new ArrayList<String>();
    entry._headerValues = new ArrayList<String>();

    for (int i = 0; i < keys.size(); i++) {
      String key = keys.get(i);
      
      if (key.equalsIgnoreCase("expires")
	  || key.equalsIgnoreCase("set-cookie")
	  || key.equalsIgnoreCase("set-cookie2")) {
	// hide cookies and expires
      }
      else {
        entry._headerKeys.add(key);
        entry._headerValues.add(values.get(i));
      }
    }

    if (contentType != null)
      entry._contentType = contentType;
    else if (charEncoding != null)
      entry._contentType = "text/html; charset=" + charEncoding;

    if (charEncoding != null)
      entry._charEncoding = charEncoding;
    
    entry._inode = _cache.createInode();
    _isCacheable = true;

    return entry;
  }

  private void fillMaxAge(CacheEntry entry, String uri,
			  CauchoRequest req,
			  AbstractHttpResponse res)
  {
    WebApp app = req.getWebApp();

    if (app != null && uri != null) {
      entry._maxAge = app.getMaxAge(uri);
      entry._sMaxAge = app.getSMaxAge(uri);
    }

    if (entry._maxAge > 0 && entry._maxAge < MIN_EXPIRES)
      entry._maxAge = MIN_EXPIRES;

    long internalCacheTimeout = 5000L;

    /*
    if (entry._expireString == null &&
	entry._maxAge == Long.MIN_VALUE && entry._sMaxAge == Long.MIN_VALUE)
      entry._maxAge = internalCacheTimeout;
    */

    if (internalCacheTimeout <= 0) {
    }
    else if (entry._maxAge > 0 && entry._sMaxAge > 0) {
      res.addHeader("Cache-Control",
		    "max-age=" + entry._maxAge / 1000L +
		    ", s-maxage=" + entry._sMaxAge / 1000L);
    }
    else if (entry._maxAge > 0) {
      res.addHeader("Cache-Control", "max-age=" + entry._maxAge / 1000L);
    }
    else if (entry._sMaxAge > 0) {
      res.addHeader("Cache-Control", "s-maxage=" + entry._sMaxAge / 1000L);
    }

    if (entry._sMaxAge > 0)
      entry._resinMaxAge = entry._sMaxAge;
    else if (entry._maxAge > 0)
      entry._resinMaxAge = entry._maxAge;
    else
      entry._resinMaxAge = internalCacheTimeout;
  }

  void killCache()
  {
    _isCacheable = false;
  }

  void disableCache()
  {
    Cache cache = _cache;

    if (cache != null) {
      cache.setEnable(false);
      cache.clear();
    }
  }

  /**
   * Returns the vary entry matching the crc.
   */
  private CacheEntry getVaryEntry(CacheEntry entry, HttpServletRequest req)
  {
    ArrayList<String> vary = entry._vary;
    
    if (vary == null)
      return entry;

    long crc = getVaryCRC(vary, req);

    for (; entry != null; entry = entry._next) {
      if (entry._varyCrc == crc)
	return entry;
    }

    return null;
  }

  /**
   * Returns the vary crc.
   */
  private long getVaryCRC(ArrayList<String> vary, HttpServletRequest req)
  {
    if (vary == null)
      return 0;

    long crc = 0;

    if (req instanceof AbstractHttpRequest) {
      AbstractHttpRequest cauchoReq = (AbstractHttpRequest) req;
      
      for (int i = 0; i < vary.size(); i++) {
	String key = vary.get(i);
	CharSegment value = cauchoReq.getHeaderBuffer(key);

	if (value != null)
	  crc = Crc64.generate(crc,
			       value.getBuffer(),
			       value.getOffset(),
			       value.getLength());
      }
    }
    else {
      for (int i = 0; i < vary.size(); i++) {
	String key = vary.get(i);
	String value = req.getHeader(key);

	if (value != null)
	  crc = Crc64.generate(crc, value);
      }
    }

    return crc;
  }
  
  /**
   * Update the headers when the caching has finished.
   *
   * @param okay if true, the cache if valid
   */
  public void finishCaching(AbstractCacheEntry requestEntry)
  {
    CacheEntry loadingEntry = (CacheEntry) requestEntry;
    CacheInode inode;
    
    if ((inode = loadingEntry._inode) != null) {
      long contentLength = inode.getLength();
      
      loadingEntry._contentLength = contentLength;
      loadingEntry._isValid = true;

      if (contentLength <= 0) {
	log.fine(this + " tried to cache blank page");
	
	loadingEntry._isValid = false;
	_isCacheable = false;
      }
      else if (contentLength < _cache.getMaxEntrySize()) {
	synchronized (this) {
	  _entry = replaceEntry(_entry, loadingEntry);
	}
	
	loadingEntry = null;
	
	// Mark this entry as used for the LRU
	_cache.activateEntry(this);
      }
      else
	_isCacheable = false;
    }

    if (loadingEntry != null) {
      try {
	loadingEntry.remove();
      } catch (Throwable e) {
	log.log(Level.WARNING, e.toString(), e);
      }
    }

    if (log.isLoggable(Level.FINE) && _entry != null)
      logCache(_entry);
  }
  
  /**
   * Update the headers when the caching has finished.
   *
   * @param okay if true, the cache if valid
   */
  public void killCaching(AbstractCacheEntry loadingEntry)
  {
    _isCacheable = false;

    if (loadingEntry != null) {
      try {
	loadingEntry.destroy();
      } catch (Throwable e) {
	log.log(Level.WARNING, e.toString(), e);
      }
    }
  }

  /**
   * Replaces the entry with the matching CRC.
   */
  private CacheEntry replaceEntry(CacheEntry head, CacheEntry next)
  {
    int count = 0;
    
    CacheEntry prev = null;
    for (CacheEntry ptr = head; ptr != null; ptr = ptr._next) {
      if (ptr._varyCrc == next._varyCrc) {
	next._next = ptr._next;

	ptr.remove();
	
	if (prev != null) {
	  prev._next = next;
	  return head;
	}
	else
	  return next;
      }

      count++;
      prev = ptr;
    }

    next._next = head;

    if (MAX_VARY < count) {
      CacheEntry ptr = head;
      for (; ptr != null && count > 0; ptr = ptr._next) {
	count--;
      }

      if (ptr != null)
	ptr._next = null;
    }

    return next;
  }

  /**
   * Callback when the item has been removed from the cache, usually
   * because of an LRU event.
   */
  public void removeEvent()
  {
    CacheEntry entry;

    synchronized (this) {
      entry = _entry;
      _entry = null;

      _isClosed = true;
    }

    if (log.isLoggable(Level.FINE) && entry != null) {
      String debugURI = _uri;
      if (_queryString != null)
	debugURI += "?" + _queryString;
      
      log.fine("lru removing cache: " + debugURI);
    }
    
    for (; entry != null; entry = entry._next) {
      try {
	entry.remove();
      } catch (Throwable e) {
	log.log(Level.FINE, e.toString(), e);
      }
    }
  }

  /**
   * Logs the cache activity.
   */
  private void logCache(CacheEntry entry)
  {
    CharBuffer extra = new CharBuffer();

    extra.append(_uri);
    if (_queryString != null) {
      extra.append("?");
      extra.append(_queryString);
    }

    if (entry._varyCrc != 0) {
      extra.append(" crc=");
      extra.append(entry._varyCrc);
    }
    
    if (entry._varyCookies) {
      extra.append(" anonymous");
      if (entry._varyCookie != null) {
        extra.append("(");
        extra.append(entry._varyCookie);
        extra.append(")");
      }
    }
        
    if (entry._etag != null) {
      extra.append(" etag=");
      extra.append(entry._etag);
    }
    else if (entry._lastModified != null) {
      extra.append(" last-modified=");
      extra.append(entry._lastModified);
    }
    else if (Alarm.getCurrentTime() < entry._expireDate) {
      extra.append(" expires=");
      extra.append(QDate.formatGMT(entry._expireDate));
    }

    if (entry._isValid && entry._inode != null) {
      long length = entry._inode.getLength();
      extra.append(" length=" + length);
    }

    if (entry != null)
      log.fine("caching: " + extra.close());
    else {
      log.fine("removing cache: " + extra.close());
    }
  }

  public String toString()
  {
    return getClass().getSimpleName() + "[" + _uri + "?" + _queryString + ", next=" + _next + "]";
  }
}
