/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cache;

import com.caucho.db.store.RawTransaction;
import com.caucho.db.store.Store;
import com.caucho.db.store.StoreTransaction;
import com.caucho.util.L10N;
import com.caucho.vfs.OutputStreamWithBuffer;
import com.caucho.vfs.TempCharBuffer;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Represents the final servlet in a filter chain.
 */
class CacheInode {
  private static final L10N L = new L10N(CacheInode.class);
  private static final Logger log
    = Logger.getLogger(CacheInode.class.getName());

  private final Store _store;
  private int _useCount = 1;

  private ArrayList<Long> _fragmentList = new ArrayList<Long>();
  private long []_fragmentArray;
  private long _length;

  CacheInode(Store store)
  {
    _store = store;
  }

  public long getLength()
  {
    return _length;
  }
  
  /**
   * Allocates access to the inode.
   */
  public boolean allocate()
  {
    synchronized (this) {
      if (_useCount > 0) {
	_useCount++;
	return true;
      }
      else
	return false;
    }
  }

  public OutputStream openOutputStream()
  {
    return new CacheOutputStream();
  }

  public Writer openWriter()
  {
    return new CacheWriter();
  }

  /**
   * Writes the inode value to a stream.
   */
  public void writeToStream(OutputStreamWithBuffer os)
    throws IOException
  {
    writeToStream(os, 0, _length);
  }

  /**
   * Writes the inode value to a stream.
   */
  public void writeToStream(OutputStreamWithBuffer os,
			    long offset, long length)
    throws IOException
  {
    if (_length < length)
      length = _length;

    byte []buffer = os.getBuffer();
    int writeLength = buffer.length;
    int writeOffset = os.getBufferOffset();
    long []fragmentArray = _fragmentArray;
    
    try {
    while (length > 0) {
      int sublen = writeLength - writeOffset;

      if (sublen == 0) {
	buffer = os.nextBuffer(writeOffset);
	writeOffset = os.getBufferOffset();
	sublen = writeLength - writeOffset;
      }

      if (length < sublen)
	sublen = (int) length;

      long fragmentAddress = fragmentArray[(int) (offset / Store.FRAGMENT_SIZE)];
      int fragmentOffset = (int) (offset % Store.FRAGMENT_SIZE);

      if (Store.FRAGMENT_SIZE - fragmentOffset < sublen)
	sublen = Store.FRAGMENT_SIZE - fragmentOffset;

      int len = _store.readFragment(fragmentAddress, fragmentOffset,
				    buffer, writeOffset, sublen);

      if (len <= 0) {
	break;
      }

      writeOffset += len;
      offset += len;
      length -= len;
    }
    } catch (RuntimeException e) {
      e.printStackTrace();
      throw e;
    } catch (IOException e) {
      e.printStackTrace();
      throw e;
    }

    os.setBufferOffset(writeOffset);

    /*
    if (_useCount < 2)
      System.out.println("USE_COUNT: " + _useCount);
    */

    if (_useCount <= 0)
      throw new IllegalStateException(L.l("Unexpected close of cache inode"));
  }

  /**
   * Writes the inode value to a stream.
   */
  public void writeToWriter(Writer out)
    throws IOException
  {
    TempCharBuffer charBuffer = TempCharBuffer.allocate();
    char []buffer = charBuffer.getBuffer();

    long offset = 0;
    
    long length = _length;
    
    while (length > 0) {
      long fragmentAddress
	= _fragmentArray[(int) (offset / Store.FRAGMENT_SIZE)];
      int fragmentOffset = (int) (offset % Store.FRAGMENT_SIZE);

      int sublen = (Store.FRAGMENT_SIZE - fragmentOffset) / 2;

      if (buffer.length < sublen)
	sublen = buffer.length;

      if (length < 2 * sublen)
	sublen = (int) (length / 2);

      int len = _store.readFragment(fragmentAddress, fragmentOffset,
				    buffer, 0, sublen);

      if (len <= 0)
	break;

      out.write(buffer, 0, len);

      offset += 2 * len;
      length -= 2 * len;
    }

    TempCharBuffer.free(charBuffer);

    if (_useCount <= 0)
      throw new IllegalStateException(L.l("Unexpected close of cache inode"));
  }

  /**
   * Allocates access to the inode.
   */
  public void free()
  {
    int useCount;
    
    synchronized (this) {
      useCount = --_useCount;
    }

    if (useCount == 0) {
      remove();
    }
    else if (useCount < 0) {
      //System.out.println("BAD: " + useCount);
      throw new IllegalStateException();
    }
  }

  private void remove()
  {
    ArrayList<Long> fragmentList;
    long []fragmentArray;
    
    synchronized (this) {
      fragmentList = _fragmentList;
      _fragmentList = null;

      fragmentArray = _fragmentArray;
      _fragmentArray = null;
    }

    StoreTransaction xa = RawTransaction.create();

    if (fragmentArray != null) {
      if (_useCount > 0)
	Thread.dumpStack();
      for (long fragment : fragmentArray) {
	try {
	  _store.deleteFragment(xa, fragment);
	} catch (IOException e) {
	  log.log(Level.WARNING, e.toString(), e);
	}
      }
      if (_useCount > 0)
	Thread.dumpStack();
    }
    else if (fragmentList != null) {
      //System.out.println("FRAGMENT-LIST: " + fragmentList);
      
      for (long fragment : fragmentList) {
	try {
	  _store.deleteFragment(xa, fragment);
	} catch (IOException e) {
	  log.log(Level.WARNING, e.toString(), e);
	}
      }
    }
  }

  class CacheOutputStream extends OutputStream {
    private final StoreTransaction _xa = RawTransaction.create();
    private final byte []_tempBuffer = new byte[8];

    public void write(int ch)
      throws IOException
    {
      _tempBuffer[0] = (byte) ch;

      write(_tempBuffer, 0, 1);
    }

    public void write(byte []buffer, int offset, int length)
      throws IOException
    {
      long inodeOffset = _length;

      while (length > 0) {
	while (_fragmentList.size() <= _length / Store.FRAGMENT_SIZE) {
	  long fragment = _store.allocateFragment(_xa);

	  _fragmentList.add(fragment);
	}

	int fragmentOffset = (int) (_length % Store.FRAGMENT_SIZE);
	long fragmentAddress = _fragmentList.get(_fragmentList.size() - 1);

	int sublen = Store.FRAGMENT_SIZE - fragmentOffset;
	if (length < sublen)
	  sublen = length;

	_length += sublen;
	_store.writeFragment(_xa,
			     fragmentAddress, fragmentOffset,
			     buffer, offset, sublen);

	length -= sublen;
	offset += sublen;
      }
    }

    public void flush()
    {
    }

    public void close()
    {
      if (_fragmentList == null)
	return;
      
      _fragmentArray = new long[_fragmentList.size()];
      
      for (int i = 0; i < _fragmentList.size(); i++) {
	_fragmentArray[i] = _fragmentList.get(i);
      }
    }
  }

  class CacheWriter extends Writer {
    private final StoreTransaction _xa = RawTransaction.create();
    private final char []_tempBuffer = new char[8];

    public void write(char ch)
      throws IOException
    {
      _tempBuffer[0] = ch;

      write(_tempBuffer, 0, 1);
    }

    public void write(char []buffer, int offset, int length)
      throws IOException
    {
      while (length > 0) {
	while (_fragmentList.size() <= _length / Store.FRAGMENT_SIZE) {
	  long fragment = _store.allocateFragment(_xa);

	  _fragmentList.add(fragment);
	}

	int fragmentOffset = (int) (_length % Store.FRAGMENT_SIZE);
	long fragmentAddress = _fragmentList.get(_fragmentList.size() - 1);

	int sublen = (Store.FRAGMENT_SIZE - fragmentOffset) / 2;
	if (length < sublen)
	  sublen = length;

	_length += 2 * sublen;
	_store.writeFragment(_xa,
			     fragmentAddress, fragmentOffset,
			     buffer, offset, sublen);

	length -= sublen;
	offset += sublen;
      }
    }

    public void flush()
    {
    }

    public void close()
    {
      synchronized (this) {
	if (_fragmentList == null)
	  return;
      
	_fragmentArray = new long[_fragmentList.size()];
      
	for (int i = 0; i < _fragmentList.size(); i++) {
	  _fragmentArray[i] = _fragmentList.get(i);
	}

	_fragmentList = null;
      }
    }
  }
}
