/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cluster;

import com.caucho.config.ConfigException;
import com.caucho.management.server.PersistentStoreMXBean;
import com.caucho.naming.Jndi;
import com.caucho.server.hmux.HmuxClusterRequest;
import com.caucho.server.hmux.HmuxRequest;
import com.caucho.util.CharBuffer;
import com.caucho.util.FreeList;
import com.caucho.vfs.Path;
import com.caucho.vfs.ReadStream;
import com.caucho.vfs.TempBuffer;
import com.caucho.vfs.TempStream;
import com.caucho.vfs.TempOutputStream;
import com.caucho.vfs.WriteStream;

import javax.sql.DataSource;
import java.io.IOException;
import java.io.InputStream;
import java.net.SocketException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.logging.Level;

/**
 * The ClusterStore manages a cluster of servers all backing up their
 * sessions to the next server.
 */
public class ClusterStore extends ClusterStoreManager {
  private static final int LOAD_OK = 1;
  private static final int LOAD_FAIL = 2;
  private static final int LOAD_EOF = 3;

  private FreeList<ClusterConnection> _freeConn
    = new FreeList<ClusterConnection>(32);
  
  private Cluster _cluster;

  private String _name;
  
  private String _serverId;

  private FileBacking _backing;

  private ClusterStoreAdmin _admin;

  private DataSource _dataSource;
  private boolean _isTriplicate = true;
  private boolean _isWaitForAcknowledge = false;

  private String _tableName;
  private String _loadQuery;
  private String _updateQuery;
  private String _accessQuery;
  private String _insertQuery;
  private String _invalidateQuery;
  private String _timeoutQuery;
  private String _dumpQuery;
  
  private ServerConnector []_serverList = new ServerConnector[0];

  // statistics
  private long _loadRemoteCount;
  
  private long _saveRemoteCount;
  
  public ClusterStore()
  {
    _cluster = Cluster.getLocal();

    _admin = new ClusterStoreAdmin(this);
    
    _backing = new FileBacking();
  }

  public void setCluster(Cluster cluster)
  {
    _cluster = cluster;
  }

  public Path getPath()
  {
    return _backing.getPath();
  }

  public void setPath(Path path)
  {
    _backing.setPath(path);
  }

  public void setJndiName(String name)
  {
    _name = name;
  }

  /**
   * Set true if the data should be stored in triplicate.
   */
  public void setTriplicate(boolean isTriplicate)
  {
    _isTriplicate = isTriplicate;
  }

  /**
   * Set true if the requests to wait for an ack.
   */
  public void setWaitForAcknowledge(boolean isWait)
  {
    _isWaitForAcknowledge = isWait;
  }

  public PersistentStoreMXBean getAdmin()
  {
    return _admin;
  }

  public boolean init()
    throws Exception
  {
    if (! super.init()) {
      return false;
    }

    if (_cluster == null)
      throw new ConfigException(L.l("Cluster is not properly initialized"));

    _serverId = _cluster.getServerId();

    ClusterServer server = _cluster.getSelfServer();
    _selfIndex = server.getIndex();

    _tableName = _backing.serverNameToTableName(_serverId);
    _backing.setTableName(_tableName);

    int length = _cluster.getServerList().length;

    if (length <= 0)
      length = 1;

    int backupLength = length - 1;
    if (backupLength <= 0)
      backupLength = 1;
    
    _loadQuery = "SELECT data FROM " + _tableName + " WHERE id=? AND is_valid=1";
    _insertQuery = ("INSERT into " + _tableName
		    + " (id,is_valid,data,mod_time,access_time,expire_interval,server1,server2,server3) "
		    + "VALUES(?,1,?,?,?,?,?,?)");
    _updateQuery = "UPDATE " + _tableName + " SET is_valid=1, data=?, mod_time=?, access_time=? WHERE id=?";
    _accessQuery = "UPDATE " + _tableName + " SET access_time=? WHERE id=?";
    _invalidateQuery = "UPDATE " + _tableName + " SET is_valid=0 WHERE id=?";
    _timeoutQuery = "DELETE FROM " + _tableName + " WHERE access_time + expire_interval < ?";

    if (_isTriplicate) {
      _dumpQuery = ("SELECT id, is_valid, expire_interval, data FROM " + _tableName +
		    " WHERE ? <= mod_time AND " +
		    "   (?=server1 OR ?=server2 OR ?=server3)");
    }
    else {
      _dumpQuery = ("SELECT id, is_valid, expire_interval, data FROM " + _tableName +
		    " WHERE ? <= mod_time AND " +
		    "   (?=server1 OR ?=server2)");
    }

    _backing.init(length);

    _dataSource = _backing.getDataSource();
    
    ClusterServer []serverList = _cluster.getServerList();
    _serverList = new ServerConnector[serverList.length];
    
    for (int i = 0; i < serverList.length; i++) {
      _serverList[i] = serverList[i].getServerConnector();

      if (_serverList[i] == null && i != _selfIndex) {
	log.warning(L.l("{0}: cluster storage requires <cluster-port> for id='{1}'.",
			this, serverList[i].getId()));
      }
    }

    _admin.register();
    
    if (_name != null)
      Jndi.bindDeepShort(_name, this);

    return true;
  }

  public boolean start()
    throws Exception
  {
    if (! super.start())
      return false;

    long delta = _backing.start();
      
    int count = _serverList.length;

    for (int i = 1; i < count; i++) {
      try {
        int nextIndex = (_selfIndex + i) % _serverList.length;

        loadDump(_serverId, _serverList[nextIndex], delta);
      } catch (Exception e) {
        log.log(Level.WARNING, e.toString(), e);
      }
    }

    return true;
  }

  /**
   * On initialization, loads all objects from the sibling.
   */
  private void loadDump(String ownerId, ServerConnector sibling, long delta)
    throws Exception
  {
    if (sibling == null)
      return;

    ClusterStream stream = sibling.openIfLive();
    
    if (stream == null) {
      log.fine(this + " tcp-store can't connect to server " + sibling);
      return;
    }
    
    if (log.isLoggable(Level.FINE)) {
      log.fine(this + " loading sessions from " + sibling);
    }

    ReadStream rs = stream.getReadStream();
    WriteStream ws = stream.getWriteStream();
    CharBuffer cb = new CharBuffer();

    int code = 0;
    TempBuffer tempBuffer = TempBuffer.allocate();
    byte []buf = tempBuffer.getBuffer();


    try {
      writeChannel(ws, 2);
      writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_CLUSTER_PROTOCOL);
      writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
                  String.valueOf(_selfIndex));
      writeString(ws, HmuxClusterRequest.HMUX_LAST_UPDATE,
		  String.valueOf(delta));
      writeString(ws, HmuxClusterRequest.HMUX_DUMP_SESSION, "default");
      // XXX: need to give time
      ws.write(HmuxRequest.HMUX_QUIT);

      String serverId = Cluster.getServerId();

      code = rs.read();
      loop:
      while (code == HmuxClusterRequest.HMUX_DUMP_SESSION) {
        int ch1 = rs.read();
        int ch2 = rs.read();
        int len = ((ch1 << 8) + ch2);

        if (ch2 < 0)
          break loop;

	// XXX:
	long expire = 120 * 60000L;
	boolean isValid = true;

        rs.readAll(buf, 0, len);
	String mangledId = new String(buf, 0, len);

	TempOutputStream os = new TempOutputStream();

	if (log.isLoggable(Level.FINE))
	  log.fine(this + " load dump: " + serverId + " " + mangledId);
	code = rs.read();

	if (code == HmuxClusterRequest.HMUX_SESSION_EXPIRE) {
	  len = (rs.read() << 8) + rs.read();
	  cb.clear();
	  rs.readAll(cb, len);

	  expire = 0;
	  for (int i = 0; i < len; i++) {
	    char ch = cb.charAt(i);
	    if ('0' <= ch && ch <= '9')
	      expire = 10 * expire + ch - '0';
	  }
	
	  code = rs.read();
	}
	  
	if (code == HmuxClusterRequest.HMUX_SESSION_INVALID) {
	  len = (rs.read() << 8) + rs.read();
	  cb.clear();
	  rs.readAll(cb, len);

	  isValid = false;
	}
	  
	for (;
	     code == HmuxClusterRequest.HMUX_SESSION_DATA;
	     code = rs.read()) {
	  ch1 = rs.read();
	  ch2 = rs.read();
	  len = ((ch1 << 8) + ch2);

	  if (ch2 < 0)
	    break loop;

	  if (os != null)
	    rs.writeToStream(os, len);
	  else
	    rs.skip(len);
	}

	if (isValid) {
	  ReadStream is = os.openRead();

	  try {
	    storeSelf(mangledId, is, os.getLength(), expire);
	  } finally {
	    is.close();
	  }
	}
	else {
	  _backing.remove(mangledId);
	}
      }      
    } finally {
      // if the load completed successfully, reuse the connection
      if (code == HmuxRequest.HMUX_QUIT) {
        stream.free();
      } else {
	stream.close();
      }
      
      TempBuffer.free(tempBuffer);
      tempBuffer = null;
    }
  }

  /**
   * Clears the old objects.
   */
  @Override
  public void clearOldObjects()
    throws SQLException
  {
    _backing.clearOldObjects(getMaxIdleTime());
  }
  
  /**
   * Creates the cluster object.
   */
  ClusterObject create(Store store, String id)
  {
    return new ClusterObject(this, store, id);
  }

  /**
   * Loads a distributed object from the backing store
   *
   * @param clusterObject the cluster object to load
   * @param obj the underlying object to load
   *
   * @return true if the load succeeded
   */
  public boolean load(ClusterObject clusterObject, Object obj)
    throws Exception
  {
    String objectId = clusterObject.getObjectId();

    ServerConnector primaryServer = getPrimaryServer(clusterObject);

    // servers A and B don't need auto-reloading, but server C does.
    // XXX: only primary can avoid auto-reload

    if (load(clusterObject, obj, primaryServer))
      return true;
    else if (primaryServer == null)
      return false;
    
    ServerConnector secondaryServer = getSecondaryServer(clusterObject);
    
    if (load(clusterObject, obj, secondaryServer))
      return true;
    else if (secondaryServer == null)
      return false;

    if (! _isTriplicate)
      return false;
    
    ServerConnector tertiaryServer = getTertiaryServer(clusterObject);
    
    if (load(clusterObject, obj, tertiaryServer))
      return true;
    else
      return false;
  }

  public boolean load(ClusterObject clusterObject,
		      Object obj,
		      ServerConnector server)
    throws Exception
  {
    if (server != null) {
      clusterObject.update();
      
      try {
	if (loadObject(clusterObject, obj, server))
	  return true;
      } catch (Exception e) {
	_loadFailCount++;
      
	log.log(Level.FINE, e.toString(), e);
      }

      return false;
    }
    else if (clusterObject.isValid()) {
      return true;
    }
    else if (_backing.loadSelf(clusterObject, obj)) {
      _loadCount++;
      
      if (log.isLoggable(Level.FINE)) {
        log.fine(this + " self-load " + clusterObject.getUniqueId());
      }
      
      return true;
    }
    else {
      return false;
    }
  }

  /**
   * Try to load the session object from the backup server.
   *
   * @param clusterObject the object handle
   * @param obj the object itself
   * @param server the client interface to the backup
   */
  boolean loadObject(ClusterObject clusterObject,
		     Object obj,
		     ServerConnector server)
    throws IOException
  {
    ClusterStream stream = server.openIfLive();

    _loadCount++;
    _loadRemoteCount++;

    if (stream != null) {
      // If server goes down and up, might have dead socket
      // server/01f1
      for (int i = 1; i >= 0 && stream != null; i--) {
	try {
	  int value = loadObject(clusterObject, obj, server, stream, -1);

	  if (value == LOAD_OK)
	    return true;
	  else if (value == LOAD_FAIL) {
	    _loadFailCount++;
	    
	    return false;
	  }
	  else {
	    _loadFailCount++;
	    
	    return false;
	  }
	} catch (SocketException e) {
	  server.clearRecycle();
	  log.fine(this + " " + e.toString());
	  log.log(Level.FINEST, e.toString(), e);
	} catch (Throwable e) {
	  server.clearRecycle();
	  log.log(Level.FINER, e.toString(), e);
	}

	if (i > 0)
	  stream = server.openIfLive();
      }
    }

    _loadFailCount++;

    return false;
  }

  /**
   * @param clusterObject the object to load.
   * @param obj the underlying object to load
   * @param srun the client to the target server
   * @param stream the write stream
   * @param updateCount the update count of the session
   *
   * @return true if the backing was successfully contacted.
   */
  int loadObject(ClusterObject clusterObject,
		 Object obj,
		 ServerConnector srun,
		 ClusterStream stream,
		 int updateCount)
    throws Exception
  {
    int code;
    TempStream tempStream = new TempStream();
    
    try {
      String uniqueId = clusterObject.getUniqueId();

      ReadStream rs = stream.getReadStream();
      WriteStream ws = stream.getWriteStream();

      writeChannel(ws, 2);
      writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_CLUSTER_PROTOCOL);
      writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
		  String.valueOf(_selfIndex));
      writeString(ws, HmuxClusterRequest.HMUX_LOAD_OBJECT, uniqueId);
      ws.write(HmuxRequest.HMUX_QUIT);
    
      WriteStream os = new WriteStream(tempStream);
      
      boolean hasData = false;
      while ((code = rs.read()) == HmuxClusterRequest.HMUX_SESSION_DATA) {
	int ch1 = rs.read();
	int ch2 = rs.read();
	int len = ((ch1 << 8) + ch2);

	if (ch2 < 0)
	  break;

	hasData = true;

	os.writeStream(rs, len);
      }
    
      os.close();

      if (code == HmuxRequest.HMUX_QUIT)
	stream.free();
      else
	stream.close();

      stream = null;

      if (code != HmuxRequest.HMUX_EXIT && code != HmuxRequest.HMUX_QUIT ||
	  ! hasData) {
	tempStream.destroy();

	return LOAD_EOF;
      }
      else {
	ReadStream is = tempStream.openRead();
	boolean isValid = false;
      
	try {
	  isValid = clusterObject.load(is, obj);
	} finally {
	  is.close();
	}

	if (log.isLoggable(Level.FINE))
	  log.fine(this + " dist-load(" +
		   srun.getIndex() + ") " + isValid + " " + uniqueId);
        
	return isValid ? LOAD_OK : LOAD_FAIL;
      }
    } finally {
      if (stream != null)
	stream.close();
    }
  }

  /**
   * The ring store has already stored the object, so this is unneeded.
   */
  public void storeOnShutdown(DistributedObject object)
    throws Exception
  {
  }

  /**
   * Stores the object in the backup.
   *
   * @param object the object to store.
   */
  public void store(ClusterObject object,
		    TempStream tempStream,
		    long crc)
    throws Exception
  {
    if (crc == 0)
      return;

    _saveCount++;
    
    ServerConnector primaryServer = getPrimaryServer(object);

    boolean hasSelf = false;

    if (primaryServer != null) {
      saveServer(tempStream, object, primaryServer);
    }
    else
      hasSelf = true;

    ServerConnector secondaryServer = getSecondaryServer(object);
    
    if (secondaryServer != null)
      saveServer(tempStream, object, secondaryServer);
    else
      hasSelf = true;

    if (_isTriplicate) {
      ServerConnector tertiaryServer = getTertiaryServer(object);
    
      if (tertiaryServer != null)
	saveServer(tempStream, object, tertiaryServer);
      else
	hasSelf = true;
    }

    if (hasSelf) {
      String objectId = object.getObjectId();
      int primary = getPrimaryIndex(objectId, 0);
      int secondary = getSecondaryIndex(objectId, 0);
      int tertiary = getTertiaryIndex(objectId, 0);

      int length = tempStream.getLength();
      ReadStream is = tempStream.openReadAndSaveBuffer();
      try {
	_backing.storeSelf(object.getUniqueId(),
			   is, length,
			   object.getMaxIdleTime(),
			   primary, secondary, tertiary);
      } finally {
	is.close();
      }
        
      if (log.isLoggable(Level.FINE)) {
	log.fine(this + " self-store " +
		 object.getObjectId() +
                 " length=" + length);
      }
    }
  }

  /**
   * Saves the session to a particular server.
   */
  void saveServer(TempStream tempStream, ClusterObject obj,
		  ServerConnector srun)
    throws IOException
  {
    _saveRemoteCount++;
    
    // server/01b1
    // loop is for the case where the recycle socket is to a closed
    // server
    for (int i = 0; i < 2; i++) {
      ClusterStream stream = srun.openIfLive();
      if (stream == null)
	return;

      try {
        saveObject(tempStream, obj, srun, stream);

	// saveObject handles the close()

	return;
      } catch (IOException e) {
	_saveFailCount++;
	
	log.log(Level.FINE, e.toString(), e);
      }

      srun.clearRecycle();
    }
  }

  /**
   * Save the object at the srun specified by the index.
   *
   * @param tempStream buffer containing the serialized object
   * @param obj the cluster object
   * @param srun the cluster client
   * @param stream connection to the cluster client
   *
   * @return true if the save was successful
   */
  boolean saveObject(TempStream tempStream, ClusterObject obj,
		     ServerConnector srun, ClusterStream stream)
    throws IOException
  {
    try {
      String mangledId = obj.getUniqueId();
    
      ReadStream rs = stream.getReadStream();
      WriteStream ws = stream.getWriteStream();

      writeChannel(ws, 2);
      writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_CLUSTER_PROTOCOL);
      writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
		  String.valueOf(srun.getIndex()));
    
      writeString(ws, HmuxClusterRequest.HMUX_SAVE_OBJECT, mangledId);
    
      writeString(ws, HmuxClusterRequest.HMUX_SESSION_EXPIRE,
		  String.valueOf(obj.getExpireInterval()));

      int length = tempStream.getLength();
      ReadStream ts = tempStream.openReadAndSaveBuffer();
      TempBuffer tb = TempBuffer.allocate();

      byte []buffer = tb.getBuffer();
      int len;

      while ((len = ts.read(buffer, 0, buffer.length)) > 0) {
	ws.write(HmuxClusterRequest.HMUX_SESSION_DATA);
	ws.write(len >> 8);
	ws.write(len);
	ws.write(buffer, 0, len);
      }

      TempBuffer.free(tb);
      tb = null;
      ts.close();

      if (_isWaitForAcknowledge || stream.isLongIdle()) {
	ws.write(HmuxRequest.HMUX_QUIT);

	int code = rs.read();

	if (code == HmuxRequest.HMUX_QUIT) {
	  stream.free();
	  stream = null;
	}
	else {
	  stream.close();
	  stream = null;

	  if (code < 0) {
	    if (log.isLoggable(Level.FINE)) {
	      log.fine(this + " failed-store(" +
		       srun.getId() + ") " + mangledId +
		       " length=" + length);
	    }
        
	    return false;
	  }
	}
      }
      else {
	ws.write(HmuxRequest.HMUX_YIELD);
	ws.flush();
	stream.free();
	stream = null;
      }
        
      if (log.isLoggable(Level.FINE)) {
	log.fine(this + " dist-store(" +
		 srun.getId() + ") " + mangledId +
		 " length=" + length);
      }
    } finally {
      if (stream != null)
	stream.close();
    }
    
    return true;
  }

  /**
   * Updates the object's access time.
   *
   * @param obj the object to store.
   */
  public void updateAccess(String uniqueId)
    throws Exception
  {
    _backing.updateAccess(uniqueId);
  }

  /**
   * Updates the object's access time.
   *
   * @param obj the object to store.
   */
  public void accessImpl(String objectId, String uniqueId)
    throws Exception
  {
    int p = uniqueId.lastIndexOf(';');
    
    ServerConnector primaryServer = getPrimaryServer(uniqueId, p + 1);

    if (primaryServer != null)
      accessServer(uniqueId, primaryServer);
    else
      _backing.updateAccess(uniqueId);
    
    ServerConnector secondaryServer = getSecondaryServer(uniqueId, p + 1);
    if (secondaryServer != null)
      accessServer(uniqueId, secondaryServer);
    else
      _backing.updateAccess(uniqueId);

    if (_isTriplicate) {
      ServerConnector tertiaryServer = getTertiaryServer(uniqueId, p + 1);
      if (tertiaryServer != null)
	accessServer(uniqueId, tertiaryServer);
      else
	_backing.updateAccess(uniqueId);
    }
  }

  void accessServer(String uniqueId, ServerConnector srun)
  {
    ClusterStream stream = srun.openIfLive();
    
    if (stream != null) {
      try {
	updateAccess(srun, stream, uniqueId);

	// the close is handled by updateAccess
      } catch (Exception e) {
	log.log(Level.WARNING, e.toString(), e);
      }
    }
  }

  /**
   * Invalidates a remote session.
   *
   * @param srun client to the remote machine
   * @param stream connection to the remote machine
   * @param uniqueId the object's unique id.
   */
  private void updateAccess(ServerConnector srun,
			    ClusterStream stream,
			    String uniqueId)
    throws Exception
  {
    ReadStream rs = stream.getReadStream();
    WriteStream ws = stream.getWriteStream();

    writeChannel(ws, 2);
    writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_CLUSTER_PROTOCOL);
    writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
                String.valueOf(_selfIndex));
    writeString(ws, HmuxClusterRequest.HMUX_ACCESS_OBJECT, uniqueId);

    if (_isWaitForAcknowledge || stream.isLongIdle()) {
      ws.write(HmuxRequest.HMUX_QUIT);
        
      int code = rs.read();

      if (code == HmuxRequest.HMUX_QUIT)
	stream.free();
      else
	stream.close();
    }
    else {
      ws.write(HmuxRequest.HMUX_YIELD);
      ws.flush();
      stream.free();
    }
        
    if (log.isLoggable(Level.FINE)) {
      log.fine(this + " dist-access " + srun.getId() + ", " + uniqueId);
    }
  }

  /**
   * When the object is no longer valid, remove it from the backing store.
   *
   * @param storeId the identifier of the storeage group
   * @param objectId the identifier of the object to remove
   */
  public void remove(String storeId, String objectId)
    throws Exception
  {
    ClusterObject obj = getClusterObject(storeId, objectId);

    removeClusterObject(storeId, objectId);

    if (obj != null)
      obj.getStore().notifyRemove(objectId);

    if (obj == null) {
      // creating the object so it can be removed
      Store store = getStore(storeId);
      obj = getClusterObject(store, objectId);
    }

    if (obj != null)
      obj.removeImpl();

    _backing.remove(storeId + ';' + objectId);
  }

  /**
   * When the session is no longer valid, remove it from the backing store.
   */
  public void remove(ClusterObject object)
    throws Exception
  {
    String storeId = object.getStoreId();
    String objectId = object.getObjectId();
    
    ServerConnector primaryServer = getPrimaryServer(object);

    removeSelf(object);
    // object.removeImpl();

    removeClusterObject(storeId, objectId);

    if (primaryServer != null)
      invalidateServer(storeId, objectId, primaryServer);
    else {
      if (log.isLoggable(Level.FINE)) {
        log.fine(this + " self-invalidate()" + object);
      }
    }
    
    ServerConnector secondaryServer = getSecondaryServer(object);
    if (secondaryServer != null)
      invalidateServer(storeId, objectId, secondaryServer);
    else {
      if (log.isLoggable(Level.FINE)) {
        log.fine(this + " self-invalidate()" + object);
      }
    }

    if (_isTriplicate) {
      ServerConnector tertiaryServer = getTertiaryServer(object);
      if (tertiaryServer != null) {
	invalidateServer(storeId, objectId, tertiaryServer);
      }
      else {
	if (log.isLoggable(Level.FINE)) {
	  log.fine(this + " self-invalidate()" + object);
	}
      }
    }
  }

  /**
   * Removes the named object from the store.
   */
  private void removeSelf(ClusterObject obj)
    throws Exception
  {
    String objectId = obj.getObjectId();

    removeClusterObject(obj.getStoreId(), objectId);

    _backing.remove(obj.getUniqueId());
  }

  void invalidateServer(String storeId, String objectId,
			ServerConnector srun)
  {
    ClusterStream stream = srun.openIfLive();

    if (stream != null) {
      try {
        invalidate(srun, stream, storeId, objectId);
      } catch (Exception e) {
        log.log(Level.FINE, e.toString(), e);
	
        stream = null;
      }
    }
  }

  /**
   * Invalidates a remote session.
   *
   * @param index the remote srun index.
   * @param rws read/write pair to the remove srun.
   * @param mangledId the remote server's id.
   */
  private void invalidate(ServerConnector srun,
                          ClusterStream stream,
                          String storeId, String objectId)
    throws Exception
  {
    ReadStream rs = stream.getReadStream();
    WriteStream ws = stream.getWriteStream();
    String mangledId = storeId + ';' + objectId;

    //writeChannel(ws, 2);
    writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_CLUSTER_PROTOCOL);
    writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
                String.valueOf(_selfIndex));
    writeString(ws, HmuxClusterRequest.HMUX_KILL_SESSION, mangledId);

    if (_isWaitForAcknowledge || stream.isLongIdle()) {
      ws.write(HmuxRequest.HMUX_QUIT);

      int code = rs.read();

      if (code == HmuxRequest.HMUX_QUIT)
	stream.free();
      else
	stream.close();
    }
    else {
      ws.write(HmuxRequest.HMUX_YIELD);
      ws.flush();
      stream.free();
    }
        
    if (log.isLoggable(Level.FINE)) {
      log.fine(this + " dist-invalidate [" + srun.getId() + "] " + mangledId);
    }
  }

  //
  // statistics
  //

  public long getObjectCount()
  {
    try {
      return _backing.getObjectCount();
    } catch (SQLException e) {
      log.log(Level.FINE, e.toString(), e);

      return -1;
    }
  }

  //
  // utilities
  //

  /**
   * Returns the object's secondary server.
   *
   * @param id the object id
   */
  private ServerConnector getPrimaryServer(ClusterObject obj)
  {
    return getPrimaryServer(obj.getObjectId(), 0);
  }

  /**
   * Returns the object's secondary server.
   *
   * @param id the object id
   */
  private ServerConnector getSecondaryServer(ClusterObject obj)
  {
    return getSecondaryServer(obj.getObjectId(), 0);
  }

  /**
   * Returns the object's tertiary server.
   *
   * @param id the object id
   */
  private ServerConnector getTertiaryServer(ClusterObject obj)
  {
    return getTertiaryServer(obj.getObjectId(), 0);
  }

  private void writeChannel(WriteStream ws, int channel)
    throws IOException
  {
    ws.write(HmuxRequest.HMUX_CHANNEL);
    ws.write(channel >> 8);
    ws.write(channel);
  }

  private void writeString(WriteStream ws, int code, String value)
    throws IOException
  {
    int len = value.length();

    ws.write(code);
    ws.write(len >> 8);
    ws.write(len);
    ws.print(value);
  }

  private void writeInt(WriteStream ws, int code, int value)
    throws IOException
  {
    ws.write(code);
    ws.write(0);
    ws.write(4);
    ws.write(value >> 24);
    ws.write(value >> 16);
    ws.write(value >> 8);
    ws.write(value);
  }

  /**
   * Updates the store.
   */
  public void update(String serverId, String uniqueId,
		     ReadStream is, long expireInterval)
    throws IOException
  {
    if (serverId.equals(_serverId)) {
      ClusterObject obj = getClusterObject(uniqueId);

      if (obj != null) {
	storeSelf(obj, is, -1);
	obj.update();
      }
      else
	storeSelf(uniqueId, is, -1, expireInterval);
    }
  }

  private void storeSelf(String uniqueId,
			 ReadStream is,
			 int length,
			 long expireInterval)
  {
    int p = uniqueId.indexOf(';');

    if (p < 0)
      p = -1;

    int primary = getPrimaryIndex(uniqueId, p + 1);
    int secondary = getSecondaryIndex(uniqueId, p + 1);
    int tertiary = getTertiaryIndex(uniqueId, p + 1);

    _backing.storeSelf(uniqueId, is, -1, expireInterval,
		       primary, secondary, tertiary);
  }
			 

  /**
   * Reads from the store.
   */
  public void read(String uniqueId, WriteStream os)
    throws IOException
  {
    Connection conn = null;
    try {
      conn = _dataSource.getConnection();

      PreparedStatement pstmt = conn.prepareStatement(_loadQuery);
      pstmt.setString(1, uniqueId);

      ResultSet rs = pstmt.executeQuery();
      if (rs.next()) {
	InputStream is = rs.getBinaryStream(1);

	os.writeStream(is);

	is.close();
      }
    } catch (SQLException e) {
      log.log(Level.FINE, e.toString(), e);
    } finally {
      try {
	if (conn != null)
	  conn.close();
      } catch (SQLException e) {
      }
    }
  }

  /**
   * Stores the cluster object on the local store.
   *
   * @param obj the cluster object
   * @param length the length of the serialized object
   */
  private void storeSelf(ClusterObject obj, ReadStream is, int length)
  {
    String uniqueId = obj.getUniqueId();

    String objectId = obj.getObjectId();
	
    int primary = getPrimaryIndex(objectId, 0);
    int secondary = getSecondaryIndex(objectId, 0);
    int tertiary = getTertiaryIndex(objectId, 0);
    
    _backing.storeSelf(uniqueId, is, length, obj.getExpireInterval(),
		       primary, secondary, tertiary);
  }

  public void dumpObjects(WriteStream os, int srun, long updateTime)
    throws IOException
  {
    Connection conn = null;
    TempBuffer tempBuffer = TempBuffer.allocate();
    
    try {
      // Contact from the target means it's alive
      ServerConnector server = _serverList[srun % _serverList.length];
      server.clearRecycle();
      server.wake();
      
      conn = _dataSource.getConnection();

      byte []buffer = tempBuffer.getBuffer();

      PreparedStatement pstmt = conn.prepareStatement(_dumpQuery);
      int updateMinutes = (int) (updateTime / 60000L);
      pstmt.setInt(1, updateMinutes);
      pstmt.setInt(2, srun);
      pstmt.setInt(3, srun);
      if (_isTriplicate)
	pstmt.setInt(4, srun);

      ResultSet rs = pstmt.executeQuery();

      while (rs.next()) {
	String uniqueId = rs.getString(1);
	boolean isValid = rs.getBoolean(2);

	writeString(os, HmuxClusterRequest.HMUX_DUMP_SESSION, uniqueId);

	String expireInterval = rs.getString(3);
	writeString(os, HmuxClusterRequest.HMUX_SESSION_EXPIRE,
		    expireInterval);

	if (isValid) {
	  try {
	    InputStream is = rs.getBinaryStream(4);

	    int len;
	    while ((len = is.read(buffer, 0, buffer.length)) > 0) {
	      os.write(HmuxClusterRequest.HMUX_SESSION_DATA);
	      os.write(len >> 8);
	      os.write(len);

	      os.write(buffer, 0, len);
	    }

	    is.close();
      
	    if (log.isLoggable(Level.FINE))
	      log.fine(this + " dump " + srun + " " + uniqueId);
	  } catch (Exception e) {
	    log.log(Level.FINE, e.toString(), e);
	  }
	}
	else {
	  os.write(HmuxClusterRequest.HMUX_SESSION_INVALID);
	  os.write(0);
	  os.write(0);
      
	  if (log.isLoggable(Level.FINE))
	    log.fine(this + " invalid " + srun + " " + uniqueId);
	}
      }
    } catch (Exception e) {
      log.log(Level.FINE, e.toString(), e);
    } finally {
      try {
	if (conn != null)
	  conn.close();
      } catch (SQLException e) {
      }
      
      TempBuffer.free(tempBuffer);
      tempBuffer = null;
    }
  }

  @Override
  public void destroy()
  {
    try {
      super.destroy();

      _backing.destroy();
      
      _admin.unregister();
    } finally {
      _dataSource = null;
      _freeConn = null;
    }
  }

  private ClusterConnection getConnection()
    throws SQLException
  {
    ClusterConnection cConn = _freeConn.allocate();

    if (cConn == null) {
      Connection conn = _dataSource.getConnection();
      cConn = new ClusterConnection(conn);
    }

    return cConn;
  }

  public String toString()
  {
    return getClass().getSimpleName() + "[" + _serverId + "," + _selfIndex + "]";
  }

  class ClusterConnection {
    private Connection _conn;
    
    private PreparedStatement _loadStatement;
    private PreparedStatement _updateStatement;
    private PreparedStatement _insertStatement;
    private PreparedStatement _accessStatement;
    private PreparedStatement _invalidateStatement;
    private PreparedStatement _timeoutStatement;

    ClusterConnection(Connection conn)
    {
      _conn = conn;
    }

    PreparedStatement prepareLoad()
      throws SQLException
    {
      if (_loadStatement == null)
	_loadStatement = _conn.prepareStatement(_loadQuery);

      return _loadStatement;
    }

    PreparedStatement prepareUpdate()
      throws SQLException
    {
      if (_updateStatement == null)
	_updateStatement = _conn.prepareStatement(_updateQuery);

      return _updateStatement;
    }

    PreparedStatement prepareInsert()
      throws SQLException
    {
      if (_insertStatement == null)
	_insertStatement = _conn.prepareStatement(_insertQuery);

      return _insertStatement;
    }

    PreparedStatement prepareAccess()
      throws SQLException
    {
      if (_accessStatement == null)
	_accessStatement = _conn.prepareStatement(_accessQuery);

      return _accessStatement;
    }

    PreparedStatement prepareInvalidate()
      throws SQLException
    {
      if (_invalidateStatement == null)
	_invalidateStatement = _conn.prepareStatement(_invalidateQuery);

      return _invalidateStatement;
    }

    void close()
    {
      _freeConn.free(this);
    }
  }

  enum LoadValue {
    OK,
    FAIL,
    UNKNOWN
  }
}
