/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cluster;

import com.caucho.log.Log;
import com.caucho.server.hmux.HmuxClusterRequest;
import com.caucho.server.hmux.HmuxRequest;
import com.caucho.util.L10N;
import com.caucho.vfs.ReadStream;
import com.caucho.vfs.WriteStream;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Base class for distributed stores.
 */
abstract public class ClusterStoreManager extends StoreManager {
  static protected final Logger log
    = Logger.getLogger(ClusterStoreManager.class.getName());
  static final L10N L = new L10N(ClusterStoreManager.class);

  /**
   * Updates the owner object.
   *
   * @param uniqueId the identifier of the storage group
   */
  public void updateOwner(String objectId, String uniqueId)
    throws Exception
  {
    ServerConnector owningServer = getOwningServer(objectId);

    if (owningServer == null)
      return;
    
    updateServer(uniqueId, owningServer);
  }

  /**
   * Updates the cached object value on the target server.
   */
  void updateServer(String uniqueId, ServerConnector server)
  {
    ClusterStream stream = server.openIfLive();

    if (stream != null) {
      try {
        update(server, stream, uniqueId);
      } catch (Throwable e) {
        log.log(Level.FINE, e.toString(), e);
        stream = null;
      }
    }
  }

  /**
   * Updates a remote object.
   *
   * @param index the remote srun index.
   * @param rws read/write pair to the remove srun.
   * @param mangledId the remote server's id.
   */
  private void update(ServerConnector server,
		      ClusterStream stream,
		      String uniqueId)
    throws Exception
  {
    try {
      ReadStream rs = stream.getReadStream();
      WriteStream ws = stream.getWriteStream();

      writeChannel(ws, 2);
      writeInt(ws, HmuxRequest.HMUX_PROTOCOL,
	       HmuxRequest.HMUX_CLUSTER_PROTOCOL);
      writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
		  String.valueOf(_selfIndex));
      writeString(ws, HmuxClusterRequest.HMUX_UPDATE_SESSION, uniqueId);
      ws.write(HmuxRequest.HMUX_QUIT);
        
      int code = rs.read();

      if (code == HmuxRequest.HMUX_QUIT) {
	stream.free();
	stream = null;
      }
    } finally {
      if (stream != null)
	stream.close();
    }
        
    if (log.isLoggable(Level.FINE)) {
      log.fine("[" + server.getId() + "] dist-update() " + uniqueId);
    }
  }

  /**
   * Returns the owner index for the id.
   */
  @Override
  public int getPrimaryIndex(String id, int offset)
  {
    ServerConnector []serverList = getServerList();

    if (serverList.length <= 1 || id.length() <= offset)
      return 0;
    else
      return decode(id.charAt(offset)) % serverList.length;
  }

  /**
   * Returns the primary server for the id.
   */
  public ServerConnector getPrimaryServer(String id, int offset)
  {
    ServerConnector []serverList = getServerList();

    if (serverList.length < 1 || id.length() <= offset)
      return null;
    else
      return serverList[decode(id.charAt(offset)) % serverList.length];
  }

  /**
   * Returns the secondary index for the id.
   */
  @Override
  public int getSecondaryIndex(String id, int offset)
  {
    ServerConnector []serverList = getServerList();
    
    if (serverList.length <= 2 || id.length() <= offset + 1)
      return 1;
    
    int index2 = decode(id.charAt(offset + 1));

    return index2 % serverList.length;
  }

  /**
   * Returns the secondary server for the id.
   */
  public ServerConnector getSecondaryServer(String id, int offset)
  {
    ServerConnector []serverList = getServerList();

    if (serverList.length < 2 || id.length() <= offset + 1)
      return null;
    else
      return serverList[decode(id.charAt(offset + 1)) % serverList.length];
  }

  /**
   * Returns the tertiary index for the id.
   */
  @Override
  public int getTertiaryIndex(String id, int offset)
  {
    ServerConnector []serverList = getServerList();
    
    if (serverList.length < 3 || id.length() <= offset + 2)
      return 2;
    
    int index3 = decode(id.charAt(offset + 2));

    return index3 % serverList.length;
  }

  /**
   * Returns the tertiary server for the id.
   */
  public ServerConnector getTertiaryServer(String id, int offset)
  {
    ServerConnector []serverList = getServerList();

    if (serverList.length < 3 || id.length() <= offset + 2)
      return null;
    else
      return serverList[decode(id.charAt(offset + 2)) % serverList.length];
  }

  /**
   * Updates the cached object value on the target server.
   */
  protected void ping(ServerConnector server)
  {
    ClusterStream stream = server.open();

    if (stream != null) {
      try {
        ping(server, stream);
      } catch (Throwable e) {
        log.log(Level.FINE, e.toString(), e);
        stream = null;
      }
    }
  }

  /**
   * Pings a remote server.
   *
   * @param server the server to ping.
   */
  private void ping(ServerConnector server, ClusterStream stream)
    throws Exception
  {
    if (log.isLoggable(Level.FINE)) {
      log.fine("[" + server.getId() + "] ping()");
    }
    
    try {
      ReadStream rs = stream.getReadStream();
      WriteStream ws = stream.getWriteStream();

      writeChannel(ws, 2);
      writeInt(ws, HmuxRequest.HMUX_PROTOCOL, HmuxRequest.HMUX_CLUSTER_PROTOCOL);
      writeString(ws, HmuxClusterRequest.HMUX_SESSION_SRUN,
		  String.valueOf(_selfIndex));
      ws.write(HmuxRequest.HMUX_QUIT);
        
      int code = rs.read();

      if (code == HmuxRequest.HMUX_QUIT) {
	stream.free();
	stream = null;
      }
    } finally {
      if (stream != null)
	stream.close();
    }
  }

  /**
   * Writes the hmux channel to the target.
   */
  private void writeChannel(WriteStream ws, int channel)
    throws IOException
  {
    ws.write(HmuxRequest.HMUX_CHANNEL);
    ws.write(channel >> 8);
    ws.write(channel);
  }

  /**
   * Writes a hmux string to the target.
   */
  private void writeString(WriteStream ws, int code, String value)
    throws IOException
  {
    int len = value.length();

    ws.write(code);
    ws.write(len >> 8);
    ws.write(len);
    ws.print(value);
  }

  /**
   * Writes a hmux int to the target.
   */
  private void writeInt(WriteStream ws, int code, int value)
    throws IOException
  {
    ws.write(code);
    ws.write(0);
    ws.write(4);
    ws.write(value >> 24);
    ws.write(value >> 16);
    ws.write(value >> 8);
    ws.write(value);
  }
}
