/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.cluster;

import com.caucho.config.ConfigException;
import com.caucho.jdbc.JdbcMetaData;
import com.caucho.util.Alarm;
import com.caucho.util.CharBuffer;
import com.caucho.vfs.TempStream;

import javax.sql.DataSource;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Level;

/**
 * Storing sessions using a JDBC database as a backing store.
 */
public class JdbcStore extends ClusterStoreManager {
  protected DataSource _dataSource;
  
  private String _loadQuery;
  private String _updateQuery;
  private String _accessQuery;
  private String _insertQuery;
  private String _invalidateQuery;
  private String _timeoutQuery;

  private String _tableName = "persistent_session";
  private String _blobType;

  /**
   * Create a new JDBC backing store.
   */
  public JdbcStore()
  {
  }

  public void setDataSource(DataSource dataSource)
  {
    _dataSource = dataSource;
  }

  public void setTableName(String tableName)
  {
    _tableName = tableName;
  }

  public void setBlobType(String blobType)
  {
    _blobType = blobType;
  }

  public boolean init()
    throws Exception
  {
    if (! super.init())
      return false;

    if (_dataSource == null)
      throw new ConfigException(L.l("JdbcStore requres a data-source configuration"));
    
    _loadQuery = "SELECT access_time, data FROM " + _tableName + " WHERE id=?";
    _insertQuery = ("INSERT into " + _tableName + " (id,data,access_time,expire_interval) " +
                   "VALUES(?,?,?,?)");
    _updateQuery = "UPDATE " + _tableName + " SET data=?, access_time=? WHERE id=?";
    _accessQuery = "UPDATE " + _tableName + " SET access_time=? WHERE id=?";
    _invalidateQuery = "DELETE FROM " + _tableName + " WHERE id=?";
    _timeoutQuery = "DELETE FROM " + _tableName + " WHERE access_time + expire_interval < ?";

    createTable();

    clearOldObjects();

    return true;
  }

  private void createTable()
    throws SQLException
  {
    // Create if necessary
    Connection conn = null;
    Statement stmt = null;
    try {
      String blob = _blobType;

      JdbcMetaData jdbcMetaData = JdbcMetaData.create(_dataSource);

      if (blob == null) {
	blob = jdbcMetaData.getBlobType();
      }

      if (blob == null)
        blob = "blob";
      
      conn = _dataSource.getConnection();

      stmt = conn.createStatement();

      try {
        ResultSet rs = stmt.executeQuery("SELECT expire_interval FROM " + _tableName + " WHERE 1=0");
	rs.close();
	stmt.close();
	return;
      } catch (SQLException e) {
	log.log(Level.FINEST, e.toString(), e);
      }

      try {
	stmt.executeUpdate("DROP TABLE " + _tableName);
      } catch (SQLException e) {
	log.log(Level.FINEST, e.toString(), e);
      }

      StringBuilder sb = new StringBuilder();
      sb.append("CREATE TABLE " + _tableName + " (");
      sb.append("  id varchar(64) NOT NULL,");
      sb.append("  data " + blob + ",");
      sb.append("  access_time INTEGER,");
      sb.append("  expire_interval INTEGER,");
      sb.append("  PRIMARY KEY(id)");
      sb.append(")");

      String sql = sb.toString();
      stmt = conn.createStatement();

      stmt.executeUpdate(sql);

      stmt.close();
    } finally {
      if (conn != null) conn.close();
    }
  }

  public boolean start()
    throws Exception
  {
    if (! super.start())
      return false;

    ServerConnector []servers = getServerList();

    for (int i = 0; i < servers.length; i++) {
      try {
	if (servers[i] != null)
	  ping(servers[i]);
      } catch (Exception e) {
        log.log(Level.WARNING, e.toString(), e);
      }
    }

    return true;
  }

  /**
   * Clears the old objects.
   */
  public void clearOldObjects()
    throws SQLException
  {
    Connection conn =  _dataSource.getConnection();

    try {
      // remove anything expired
      PreparedStatement pstmt = conn.prepareStatement(_timeoutQuery);
  
      long now = Alarm.getCurrentTime();
      
      int minute = (int) (now / 60000L);

      pstmt.setInt(1, minute);
      
      int count = pstmt.executeUpdate();

      if (log.isLoggable(Level.FINE))
	log.fine("clearing " + count + " expired objects");

      pstmt.close();
    } finally {
      conn.close();
    }
  }

  /**
   * Load the session from the jdbc store.
   *
   * @param clusterObj the cluster object to fill.
   *
   * @return true if the load was valid.
   */
  @Override
  protected boolean load(ClusterObject clusterObj, Object obj)
    throws Exception
  {
    String uniqueId = clusterObj.getUniqueId();
    
    Connection conn = _dataSource.getConnection();
    try {
      PreparedStatement stmt = conn.prepareStatement(_loadQuery);
      stmt.setString(1, uniqueId);
      
      ResultSet rs = stmt.executeQuery();
      boolean validLoad = false;

      if (rs.next()) {
	int accessTime = rs.getInt(1);
	
        InputStream is = rs.getBinaryStream(2);

        if (log.isLoggable(Level.FINE))
          log.fine("load jdbc: " + uniqueId);
      
        validLoad = clusterObj.load(is, obj);

	if (validLoad)
	  clusterObj.setAccessTime(accessTime * 60000L);

        is.close();
      }
      else if (log.isLoggable(Level.FINE))
        log.fine("no session loaded for " + uniqueId);

      rs.close();
      stmt.close();

      return validLoad;
    } finally {
      conn.close();
    }
  }

  /**
   * Save the object to the jdbc store.
   *
   * @param obj the object to store.
   * @param tempStream the stream containing the object's data
   */
  protected void store(ClusterObject obj, TempStream tempStream,
		       long crc)
    throws Exception
  {
    Connection conn = _dataSource.getConnection();

    try {
      int length = tempStream.getLength();

      String objectId = obj.getObjectId();
      String uniqueId = obj.getUniqueId();
	
      long now = Alarm.getCurrentTime();
      long expireInterval = obj.getExpireInterval();

      int nowMinutes = (int) (now / 60000L);
      int expireMinutes = (int) (expireInterval / 60000L);
      
      int count = 0;
      try {
        PreparedStatement stmt = conn.prepareStatement(_updateQuery);
        InputStream is = tempStream.openReadAndSaveBuffer();
        stmt.setBinaryStream(1, is, length);
	  
        stmt.setInt(2, nowMinutes);
        stmt.setString(3, uniqueId);

        count = stmt.executeUpdate();
        
        is.close();
        
        stmt.close();

        if (count > 0) {
          if (log.isLoggable(Level.FINE)) 
            log.fine("update jdbc: " + uniqueId + " length=" + length);

	  updateOwner(objectId, uniqueId);
	  
          return;
        }
      } catch (SQLException e) {
        log.log(Level.FINER, e.toString(), e);
      }

      try {
        PreparedStatement stmt = conn.prepareStatement(_insertQuery);
        
        stmt.setString(1, uniqueId);

        InputStream is = tempStream.openReadAndSaveBuffer();
        stmt.setBinaryStream(2, is, length);
        stmt.setInt(3, nowMinutes);
        stmt.setInt(4, expireMinutes);
      
        stmt.executeUpdate();
        is.close();
          
        stmt.close();
      } catch (SQLException e) {
        log.log(Level.WARNING, e.toString(), e);
      }
        
      if (log.isLoggable(Level.FINE))
        log.fine("insert jdbc: " + objectId + " length:" + length);

      updateOwner(objectId, uniqueId);
    } finally {
      conn.close();

      tempStream.destroy();
    }
  }

  /**
   * Updates the object's access time.
   *
   * @param uniqueId the object id to update to store.
   */
  @Override
  public void accessImpl(String objectId, String uniqueId)
    throws Exception
  {
    Connection conn = _dataSource.getConnection();

    try {
      PreparedStatement stmt = conn.prepareStatement(_accessQuery);

      long now = Alarm.getCurrentTime();
      int nowMinutes = (int) (now / 60000L);
      
      stmt.setInt(1, nowMinutes);
      stmt.setString(2, uniqueId);

      int count = stmt.executeUpdate();
        
      stmt.close();

      if (count > 0) {
	if (log.isLoggable(Level.FINE)) 
	  log.fine("access jdbc: " + uniqueId);
	return;
      }
    } finally {
      conn.close();
    }
  }

  /**
   * Removes the named object from the store.
   */
  public void remove(ClusterObject obj)
    throws Exception
  {
    Connection conn = _dataSource.getConnection();
    
    try {
      String objectId = obj.getObjectId();
      String uniqueId = obj.getUniqueId();

      removeClusterObject(obj.getStoreId(), objectId);

      PreparedStatement stmt = conn.prepareStatement(_invalidateQuery);
      stmt.setString(1, uniqueId);

      stmt.executeUpdate();

      stmt.close();
      
      if (log.isLoggable(Level.FINE))
        log.fine(this + " remove jdbc: " + uniqueId);
      
      updateOwner(objectId, uniqueId);
    } finally {
      conn.close();
    }
  }

  public String toString()
  {
    return "JdbcStore[" + _dataSource + "]";
  }
}
