/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.hmux;

import com.caucho.log.Log;
import com.caucho.server.cluster.Cluster;
import com.caucho.server.cluster.ClusterServer;
import com.caucho.server.cluster.ClusterStore;
import com.caucho.server.cluster.StoreManager;
import com.caucho.server.webapp.ErrorPageManager;
import com.caucho.util.Alarm;
import com.caucho.util.CharBuffer;
import com.caucho.vfs.ReadStream;
import com.caucho.vfs.TempBuffer;
import com.caucho.vfs.TempStream;
import com.caucho.vfs.TempOutputStream;
import com.caucho.vfs.WriteStream;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Handles cluster requests from a remote dispatcher.
 */
public class HmuxClusterRequest extends AbstractClusterRequest {
  private static final Logger log = Log.open(HmuxClusterRequest.class);

  // other, specialized protocols
  public static final int HMUX_QUERY =           'q';
  public static final int HMUX_PING =            'p';
  public static final int HMUX_SAVE_OBJECT =     'b';
  public static final int HMUX_LOAD_OBJECT =     'c';
  public static final int HMUX_SESSION_DATA =    'e';
  public static final int HMUX_KILL_SESSION =    'i';
  public static final int HMUX_SESSION_SRUN =    's';
  public static final int HMUX_SESSION_EXPIRE =  'E';
  public static final int HMUX_SESSION_INVALID = 'I';
  public static final int HMUX_UPDATE_SESSION =  'u';
  public static final int HMUX_ACCESS_OBJECT =   'a';
  public static final int HMUX_DUMP_SESSION =    'd';
  public static final int HMUX_LAST_UPDATE =     'l';

  private CharBuffer _cb = new CharBuffer();

  private HmuxRequest _request;

  private Cluster _cluster;

  private ErrorPageManager _errorManager = new ErrorPageManager();

  private WriteStream _clusterWriter = new WriteStream();

  private int _srunIndex;

  public HmuxClusterRequest()
  {
  }

  public void setRequest(HmuxRequest request)
  {
    _request = request;
  }

  public void setCluster(Cluster cluster)
  {
    _cluster = cluster;
  }

  /**
   * Handles a new request.  Initializes the protocol handler and
   * the request streams.
   *
   * <p>Note: ClientDisconnectException must be rethrown to
   * the caller.
   */
  public int handleRequest(ReadStream is, WriteStream os)
    throws IOException
  {
    CharBuffer cb = _cb;
    boolean isLoggable = log.isLoggable(Level.FINE);
    int code;
    int len;
    long lastUpdate = 0;

    Thread thread = Thread.currentThread();
    ClassLoader oldLoader = thread.getContextClassLoader();

    try {
      ClassLoader loader = _request.getDispatchServer().getClassLoader();
      if (loader != null)
	thread.setContextClassLoader(loader);

      while (true) {
	code = is.read();

	switch (code) {
	case -1:
	  if (isLoggable)
	    log.fine(dbgId() + "end of file");
	  return HmuxRequest.HMUX_EXIT;

	case HmuxRequest.HMUX_CHANNEL:
	  is.read();
	  is.read();
	  break;

	case HmuxRequest.HMUX_YIELD:
	  if (isLoggable) {
	    log.fine(dbgId() + (char) code + ": yield");
	  }
        
	  return HmuxRequest.HMUX_YIELD;

	case HmuxRequest.HMUX_QUIT:
	  if (isLoggable) {
	    log.fine(dbgId() + (char) code + ": end of request");
	  }
        
	  return HmuxRequest.HMUX_QUIT;

	case HmuxRequest.HMUX_EXIT:
	  if (isLoggable)
	    log.fine(dbgId() + (char) code + ": end of socket");

	  return HmuxRequest.HMUX_EXIT;

	case HMUX_SESSION_SRUN:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  _srunIndex = 0;
	  for (int i = 0; i < len; i++) {
	    char ch = _cb.charAt(i);
	    if (ch >= '0' && ch <= '9')
	      _srunIndex = 10 * _srunIndex + ch - '0';
	  }
        
	  if (isLoggable) {
	    log.fine(dbgId() + "srun: " + _srunIndex);
	  }

	  ClusterServer distServer;
	  distServer = _cluster.getServer(_srunIndex);
	  if (distServer != null && distServer.getServerConnector() != null)
	    distServer.getServerConnector().wake();
	  break;

	case HMUX_SAVE_OBJECT:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  if (isLoggable)
	    log.fine(dbgId() + "save-object " + _cb);

	  return saveObject(is, os, _cb);

	case HMUX_LOAD_OBJECT:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  if (! loadObject(os, _cb)) {
	    _request.killKeepalive();
	    return HmuxRequest.HMUX_EXIT;
	  }
	  break;
        
	case HMUX_DUMP_SESSION:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  if (! dumpObjects(os, _cb, lastUpdate)) {
	    _request.killKeepalive();
	    return HmuxRequest.HMUX_EXIT;
	  }
	  break;
        
	case HMUX_LAST_UPDATE:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  lastUpdate = Long.parseLong(_cb.toString()) + Alarm.getCurrentTime();
	  break;

	case HMUX_KILL_SESSION:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  killObject(_cb);
	  break;

	case HMUX_UPDATE_SESSION:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  updateObject(_cb);
	  break;

	case HMUX_ACCESS_OBJECT:
	  len = (is.read() << 8) + is.read();
	  _cb.clear();
	  is.readAll(_cb, len);

	  accessObject(_cb);
	  break;
        
	default:
	  len = (is.read() << 8) + is.read();

	  if (isLoggable)
	    log.fine(dbgId() + (char) code + " " + len + " (cluster)");
	  is.skip(len);
	  
	  _request.killKeepalive();
	  return HmuxRequest.HMUX_EXIT;
	}
      }
    } catch (Throwable e) {
      log.log(Level.FINER, e.toString(), e);
      
      return HmuxRequest.HMUX_EXIT;
    } finally {
      thread.setContextClassLoader(oldLoader);
    }

    // _filter.setClientClosed(true);

    // return false;
  }

  /**
   * Save the session
   *
   * @param cb char buffer containing the session id
   */
  private int saveObject(ReadStream is, WriteStream os, CharBuffer cb)
    throws IOException
  {
    String uniqueId = cb.toString();

    if (_cluster == null)
      return HmuxRequest.HMUX_EXIT;

    ClusterStore store = (ClusterStore) _cluster.getStore();
    if (store == null)
      return HmuxRequest.HMUX_EXIT;

    TempOutputStream out = new TempOutputStream();

    int code = HmuxRequest.HMUX_EXIT;

    long expire = 120 * 60000L;
    boolean hasData = false;

    code = is.read();

    if (code == HMUX_SESSION_EXPIRE) {
      int len = (is.read() << 8) + is.read();
      _cb.clear();
      is.readAll(_cb, len);

      expire = 0;
      for (int i = 0; i < len; i++) {
	char ch = _cb.charAt(i);
	if (ch >= '0' && ch <= '9')
	  expire = 10 * expire + ch - '0';
      }
	
      code = is.read();
    }
      
    for (; code == HMUX_SESSION_DATA; code = is.read()) {
      int c1 = is.read();
      int c2 = is.read();
      int len = (c1 << 8) + c2;
            
      is.writeToStream(out, len);

      if (c2 >= 0)
	hasData = true;
    }

    switch (code) {
    case HmuxRequest.HMUX_EXIT:
    case HmuxRequest.HMUX_QUIT:
    case HmuxRequest.HMUX_YIELD:
      break;
          
    default:
      // bad requests must not store data
      return HmuxRequest.HMUX_EXIT;
    }

    if (hasData) {
      ClusterServer distServer = _cluster.getServer(_srunIndex);

      is = out.openRead();
      try {
	store.update(distServer.getId(), uniqueId, is, expire);
      } finally {
	is.close();
      }
    }

    return code;
  }

  /**
   * Returns the serialized session based on a foreign request.
   */
  private boolean loadObject(WriteStream os, CharBuffer cb)
    throws IOException
  {
    if (_cluster == null)
      return false;

    ClusterStore store = (ClusterStore) _cluster.getStore();
    if (store == null)
      return false;

    String uniqueId = cb.toString();
    
    TempStream ts = new TempStream();
    WriteStream out = new WriteStream(ts);

    store.read(uniqueId, out);
    out.close();
    
    ReadStream is = ts.openRead();
    TempBuffer tb = TempBuffer.allocate();
    try {
      byte []buffer = tb.getBuffer();

      int len;
      while ((len = is.read(buffer, 0, buffer.length)) > 0) {
        os.write(HMUX_SESSION_DATA);
        os.write(len >> 8);
        os.write(len);

        os.write(buffer, 0, len);
      }
      
      if (log.isLoggable(Level.FINE))
        log.fine(dbgId() + "load object: " + uniqueId);
      
      return true;
    } finally {
      TempBuffer.free(tb);
      tb = null;
      is.close();
    }
  }

  /**
   * Returns the serialized session based on a foreign request.
   */
  private boolean dumpObjects(WriteStream os,
			      CharBuffer cb,
			      long lastUpdate)
    throws IOException
  {
    String runId = cb.toString();

    if (log.isLoggable(Level.FINE))
      log.fine(dbgId() + "session dump " + runId + " for " + _srunIndex);

    ClusterStore store = (ClusterStore) _cluster.getStore();

    if (store == null)
      return false;

    int callerSrunIndex = _srunIndex;

    store.dumpObjects(os, callerSrunIndex, lastUpdate);

    return true;
  }

  /**
   * Remove the named session from the store in response to an invalidate
   * request.
   *
   * @param cb the name of the session.
   */
  private boolean killObject(CharBuffer cb)
    throws IOException
  {
    int p = cb.lastIndexOf(';');

    if (p < 0)
      return false;

    String storeId = cb.substring(0, p);
    String objectId = cb.substring(p + 1);

    ClusterStore store = (ClusterStore) _cluster.getStore();

    if (store == null)
      return false;

    try {
      store.remove(storeId, objectId);

      if (log.isLoggable(Level.FINE))
	log.fine(dbgId() + "remote invalidate session " + cb);
    } catch (Exception e) {
      log.log(Level.WARNING, e.toString(), e);
    }

    return true;
  }

  /**
   * Remove any cached value of the named session.
   *
   * @param cb the name of the session.
   */
  private boolean updateObject(CharBuffer cb)
    throws IOException
  {
    int p = cb.lastIndexOf(';');

    if (p < 0)
      return false;

    String storeId = cb.substring(0, p);
    String objectId = cb.substring(p + 1);

    StoreManager store = _cluster.getStore();

    if (store == null)
      return false;

    try {
      store.update(storeId, objectId);

      if (log.isLoggable(Level.FINE))
	log.fine(dbgId() + "remote update session " + cb);
    } catch (Throwable e) {
      log.log(Level.WARNING, e.toString(), e);
    }

    return true;
  }

  /**
   * Remove any cached value of the named session.
   *
   * @param cb the name of the session.
   */
  private boolean accessObject(CharBuffer cb)
    throws IOException
  {
    StoreManager store = _cluster.getStore();

    if (! (store instanceof ClusterStore))
      return false;

    ClusterStore clusterStore = (ClusterStore) store;

    try {
      clusterStore.updateAccess(cb.toString());

      if (log.isLoggable(Level.FINE))
	log.fine(dbgId() + "remote access session " + cb);
    } catch (Exception e) {
      log.log(Level.WARNING, e.toString(), e);
    }

    return true;
  }

  void writeString(WriteStream os, int code, String value)
    throws IOException
  {
    int len = value.length();

    os.write(code);
    os.write(len >> 8);
    os.write(len);
    os.print(value);
    
    if (log.isLoggable(Level.FINE))
      log.fine(dbgId() + (char)code + " " + value);
  }

  private String dbgId()
  {
    return _request.dbgId();
  }
}
