/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.server.port;

import com.caucho.boot.JniProcess;
import com.caucho.config.ConfigException;
import com.caucho.license.LicenseCheck;
import com.caucho.lifecycle.Lifecycle;
import com.caucho.log.Log;
import com.caucho.util.Alarm;
import com.caucho.util.L10N;
import com.caucho.util.ThreadPool;
import com.caucho.vfs.JniSocketImpl;
import com.caucho.vfs.WriteStream;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Represents a protocol connection.
 */
public class JniSelectManager
  extends AbstractSelectManager implements Runnable
{
  private static final L10N L = new L10N(JniSelectManager.class);
  private static final Logger log = Log.open(JniSelectManager.class);

  private static boolean _hasJni;

  private long _maxSelectTime = 60000L;
  private int _selectMax;

  private Thread _thread;

  private TcpConnection []_connections = new TcpConnection[16 * 1024];
  private volatile int _maxConnection;
  private int _connectionCount;

  private boolean _isFiner;

  private long _fd;

  private final Lifecycle _lifecycle = new Lifecycle();

  private JniSelectManager()
  {
    int fd = 0;

    try {
      fd = JniProcess.getFdMax();
    } catch (Exception e) {
      log.log(Level.FINE, e.toString(), e);
    }

    if (fd > 256)
      _selectMax = fd - 256;
    else if (fd > 0)
      _selectMax = 512;
    else
      _selectMax = 8192;
  }

  /**
   * Returns a jni select manager.
   */
  public static JniSelectManager create()
  {
    try {
      Class cl = Class.forName("com.caucho.license.LicenseCheckImpl");

      LicenseCheck license = (LicenseCheck) cl.newInstance();

      license.validate(1);

      if (_hasJni)
        return new JniSelectManager();
    } catch (ConfigException e) {
      log.finer(e.toString());
    } catch (Throwable e) {
      log.log(Level.FINER, e.toString(), e);
    }

    return null;
  }

  /**
   * Sets the max select.
   */
  public void setSelectMax(int max)
  {
    _selectMax = max;
  }

  /**
   * Sets the max select.
   */
  public int getSelectMax()
  {
    return _selectMax;
  }

  /**
   * Sets the select timeout
   */
  public void setSelectTimeout(long timeout)
  {
    _maxSelectTime = timeout;
  }

  /**
   * Starts the manager.
   */
  @Override
  public boolean start()
  {
    if (! _lifecycle.toStarting())
      return false;

    _fd = createNative();
    if (_fd == 0) {
      _lifecycle.toDestroy();
      log.finer(this + " is not available on this system.");
      return false;
    }

    _isFiner = log.isLoggable(Level.FINER);

    String name = "resin-select-manager";
    _thread = new Thread(this, name);
    _thread.setDaemon(true);

    _thread.start();
    Thread.yield();

    synchronized (_thread) {
      try {
        if (! _lifecycle.isActive())
          _thread.wait(2000);
      } catch (Throwable e) {
      }
    }

    if (log.isLoggable(Level.FINER))
      log.finer(this + " active");

    return true;
  }

  /**
   * Returns the keepalive count.
   */
  public int getSelectCount()
  {
    return _connectionCount;
  }

  /**
   * Returns the available keepalive.
   */
  public int getFreeKeepalive()
  {
    return _selectMax - _connectionCount;
  }

  /**
   * Adds a keepalive connection.
   *
   * @param conn the connection to register as keepalive
   *
   * @return true if the connection changes to the keepalive state
   */
  public boolean keepalive(TcpConnection tcpConn)
  {
    try {
      if (! _lifecycle.isActive()) {
        log.fine(this + " inactive");
        return false;
      }
      else if (_selectMax <= _connectionCount) {
        log.warning(this + " keepalive overflow " + _connectionCount + " max=" + _selectMax);

        return false;
      }

      JniSocketImpl socket = (JniSocketImpl) tcpConn.getSocket();

      if (socket == null) {
        if (log.isLoggable(Level.FINE))
          log.fine(this + " socket empty for " + tcpConn);

        return false;
      }

      WriteStream os = tcpConn.getWriteStream();
      os.flush();

      int nativeFd = socket.getNativeFd();

      if (nativeFd >= 0 && nativeFd < _connections.length) {
        tcpConn.setAccessTime(Alarm.getCurrentTime());

        synchronized (this) {
          if (_maxConnection < nativeFd)
            _maxConnection = nativeFd;

          if (_connections[nativeFd] != null) {
            TcpConnection oldConn = _connections[nativeFd];

            tcpConn.destroy();
            if (oldConn != null)
              oldConn.destroy();

            throw new IllegalStateException("FD: " + nativeFd + " isAlreadyWaiting:" + _connections[nativeFd] + " ? " + tcpConn);
          }

          _connections[nativeFd] = tcpConn;

          _connectionCount++;

          tcpConn.setKeepalive();
        }

        int result = addNative(_fd, nativeFd);

        if (result == nativeFd) {
          tcpConn = null;
          return true;
        }
        else {
          tcpConn.clearKeepalive();

          log.warning("JniSelect: add-native failed " + nativeFd + " -> " + result + " for " + tcpConn);

          synchronized (this) {
            _connectionCount--;
            if (_connections[nativeFd] == tcpConn) {
              _connections[nativeFd] = null;
            }
          }
        }
      }
      else if (nativeFd >= 0) {
        log.warning(this + " file descriptor overflow fd=" + nativeFd
                    + " max=" + _connections.length + " " + tcpConn);
      }
      else {
        log.fine(this + " file descriptor closed fd=" + nativeFd + " " + tcpConn);
      }

      return false;
    } catch (Exception e) {
      log.log(Level.WARNING, e.toString(), e);
      _lifecycle.toError();

      return false;
    } finally {
      if (tcpConn != null) {
        tcpConn.destroy();
      }
    }
  }

  /**
   * Running process accepting connections.
   */
  public void run()
  {
    int interruptCount = 0;
    int exceptionCount = 0;

    initNative(_fd);

    _lifecycle.toActive();
    log.finer(this + " active");

    synchronized (_thread) {
      _thread.notify();
    }

    long lastCheckTime = 0;
    long checkInterval = 15000L;
    long maxSelectTime = _maxSelectTime;

    while (_lifecycle.isActive()) {
      try {
        int select = selectNative(_fd, 5000L);

        if (select >= 0) {
          TcpConnection conn;

          synchronized (this) {
            conn = _connections[select];
            _connections[select] = null;

            if (conn != null)
              _connectionCount--;
          }

          if (conn != null)
            ThreadPool.getThreadPool().schedule(conn);
        }

        long now = Alarm.getCurrentTime();

        if (lastCheckTime + checkInterval < now) {
          lastCheckTime = now;

          // XXX: xmpp issues
          boolean isFull = false;

          // adaptively set the keepalive time to keep free connections
          if (3 * _selectMax < _connectionCount * 4) {
            maxSelectTime /= 2;

            if (maxSelectTime <= 0)
              maxSelectTime = 1000;

            isFull = true;
          }
          else if (maxSelectTime < _maxSelectTime) {
            maxSelectTime += 100;

            if (_maxSelectTime < maxSelectTime)
              maxSelectTime = _maxSelectTime;
          }

          int maxConnection;

          synchronized (this) {
            maxConnection = _maxConnection;
            _maxConnection = -1;
          }

          for (int i = maxConnection; i >= 0; i--) {
            TcpConnection conn;

            synchronized (this) {
              conn = _connections[i];

              if (conn == null) {
              }
              // connection has timed out
              else if (isFull && conn.getAccessTime() + maxSelectTime < now) {
                if (log.isLoggable(Level.FINER)) {
                  log.finer(this + " connection throttle timeout time=" +
                            (now - conn.getAccessTime()) + " conn=" + conn);
                }

                _connections[i] = null;
                _connectionCount--;
              }
              // connection has timed out
              else if (! isFull && conn.getKeepaliveExpireTime() < now) {
                if (log.isLoggable(Level.FINER)) {
                  log.finer(this + " connection keepalive timeout time=" +
                            (now - conn.getAccessTime()) + " conn=" + conn);
                }

                _connections[i] = null;
                _connectionCount--;
              }
              else {
                conn = null;

                if (_maxConnection < i)
                  _maxConnection = i;
              }
            }

            if (conn != null) {
              remove(conn);
              conn.free();
            }
          }
        }

        interruptCount = 0;
        exceptionCount = 0;
      } catch (InterruptedIOException e) {
        log.log(Level.FINER, e.toString(), e);

        Thread.currentThread().interrupted();

        // If there's some sort of terminal exception, throw it
        if (interruptCount++ > 100) {
          log.fine("closing because too many interrupts");

          log.log(Level.FINE, e.toString(), e);
          close();
          break;
        }
      } catch (Throwable e) {
        log.log(Level.FINER, e.toString(), e);

        // If there's some sort of terminal exception, throw it
        if (exceptionCount++ > 100) {
          log.severe("closing because too many exceptions");
          log.log(Level.SEVERE, e.toString(), e);

          stop();
          System.exit(1);
          break;
        }
      }
    }

    _thread = null;

    stop();

    log.finer(this + " stopped");
  }

  /**
   * Closing the manager.
   */
  public boolean stop()
  {
    if (! _lifecycle.toDestroy())
      return false;

    log.finest(this + " stopping");

    for (int i = 0; i < _connections.length; i++) {
      TcpConnection tcpConn;

      synchronized (this) {
        tcpConn = (TcpConnection) _connections[i];
        _connections[i] = null;
       }

       if (tcpConn != null) {
        remove(tcpConn);

        try {
          tcpConn.close();
        } catch (Throwable e) {
          log.log(Level.WARNING, e.toString(), e);
        }
       }
     }

     long fd;

     synchronized (this) {
       fd = _fd;
       _fd = 0;
     }

     if (fd != 0)
       closeNative(fd);

     Thread.yield();

     return true;
   }

  /**
   * Removes the connection from the selection.
   */
  private void remove(TcpConnection conn)
  {
    if (conn == null)
      return;

    JniSocketImpl socket = (JniSocketImpl) conn.getSocket();
    int nativeFd = socket.getNativeFd();

    if (nativeFd > 0) {
      try {
        removeNative(_fd, nativeFd);
      } catch (Throwable e) {
        log.log(Level.WARNING, e.toString(), e);
        _lifecycle.toError();
      }
    }
  }

  public String toString()
  {
    return getClass().getSimpleName() + "[]";
  }

  public native long createNative();
  public native void initNative(long fd);
  public native int addNative(long manager, int fd) throws IOException;
  public native int removeNative(long manager, int fd) throws IOException;

  public native int selectNative(long manager, long timeout) throws IOException;
  public native int closeNative(long manager);

  protected void finalize()
  {
    close();
  }

  static {
    try {
      System.loadLibrary("resin");
      _hasJni = true;
    } catch (Throwable e) {
      // log.log(Level.WARNING, e.toString(), e);
    }
  }
}
