/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.servlets;

import com.caucho.config.ConfigException;
import com.caucho.server.admin.Management;
import com.caucho.server.cluster.Cluster;
import com.caucho.server.cluster.ClusterServer;
import com.caucho.server.cluster.ClusterStream;
import com.caucho.server.cluster.ServerConnector;
import com.caucho.server.connection.AbstractHttpRequest;
import com.caucho.server.hmux.HmuxRequest;
import com.caucho.server.resin.Resin;
import com.caucho.server.webapp.Application;
import com.caucho.util.*;
import com.caucho.vfs.ReadStream;
import com.caucho.vfs.TempBuffer;
import com.caucho.vfs.WriteStream;
import com.caucho.vfs.ClientDisconnectException;

import javax.servlet.GenericServlet;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Load balancing.
 *
 * <pre>
 * &lt;servlet-mapping url-pattern='/remote/*'>
 *   &lt;servlet-name>com.caucho.servlets.LoadBalanceServlet&lt;/servlet-name>
 *   &lt;init>
 *      &lt;cluster>app-tier&lt;/cluster>
 *   &lt;/init>
 * &lt;/servlet-mapping>
 * </pre>
 */
public class HmuxLoadBalanceServlet extends GenericServlet {
  static protected final Logger log
    = Logger.getLogger(HmuxLoadBalanceServlet.class.getName());
  static final L10N L = new L10N(HmuxLoadBalanceServlet.class);

  private static final String REQUEST_URI =
    "javax.servlet.include.request_uri";
  private static final String QUERY_STRING =
    "javax.servlet.include.query_string";

  private static final int EXIT_MASK = 0x1;
  private static final int QUIT = 0x0;
  private static final int EXIT = 0x1;

  private static final int STATUS_MASK = 0x6;
  private static final int OK = 0x0; // request succeeded
  private static final int BUSY = 0x2; // server sends busy (retry GET/POST)
  private static final int FAIL = 0x4; // server failed (retry GET)

  static int DECODE[];

  private String _clusterId;
  private Cluster _cluster;
  private ServerConnector []_srunClients;
  private boolean _isStickySessions = true;
  private LoadBalanceStrategy _strategy;
  private int _roundRobin;
  protected QDate _calendar = new QDate();

  private String _sessionPrefix;
  private String _altSessionPrefix;
  private boolean _isEscapedURL;

  /**
   * Set true if sticky-sessions should be enabled.
   */
  public void setStickySessions(boolean isStickySessions)
  {
    _isStickySessions = isStickySessions;
  }

  /**
   * Sets the cluster id
   */
  public void setCluster(String id)
  {
    _clusterId = id;
  }

  /**
   * Sets true for escaped URLs
   */
  public void setEscapedURL(boolean isEscaped)
  {
    _isEscapedURL = isEscaped;
  }

  /**
   * Sets the strategy.
   */
  public void setStrategy(String strategy)
    throws ConfigException
  {
    if ("round-robin".equals(strategy))
      _strategy = new RoundRobinBalanceStrategy();
    else if ("least-connection".equals(strategy))
      _strategy = new LeastConnectionBalanceStrategy();
    else
      throw new ConfigException(L.l("'{0}' is an unknown load-balance strategy.  'round-robin' and 'least-connection' are the known values.",
                                    strategy));
  }

  /**
   * Initialize the servlet with the server's sruns.
   */
  public void init()
    throws ServletException
  {
    Application app = (Application) getServletContext();
    //ServletServer server = app.getServer();

    _sessionPrefix = app.getSessionManager().getSessionPrefix();
    _altSessionPrefix = app.getSessionManager().getAlternateSessionPrefix();

    if (_clusterId == null)
      _clusterId = getInitParameter("cluster");

    if (_clusterId == null)
      throw new ServletException("LoadBalanceServlet needs valid `cluster'");

    String sticky = getInitParameter("sticky-sessions");
    if ("false".equals(sticky))
      _isStickySessions = false;

    Resin resin = Resin.getLocal();

    if (resin != null)
      _cluster = resin.findCluster(_clusterId);

    if (_cluster == null)
      throw new ServletException(L.l("'{0}' is not a known cluster. LoadBalanceServlet needs valid cluster.", _clusterId));

    ClusterServer []serverList = _cluster.getServerList();
    _srunClients = new ServerConnector[serverList.length];
    for (int i = 0; i < serverList.length; i++) {
      _srunClients[i] = serverList[i].getServerConnector();
    }

    if (_strategy == null)
      _strategy = new LeastConnectionBalanceStrategy();

    _strategy.setServerGroup(_srunClients);
  }

  /**
   * Handle the request.
   */
  public void service(ServletRequest request, ServletResponse response)
    throws ServletException, IOException
  {
    HttpServletRequest req = (HttpServletRequest) request;
    HttpServletResponse res = (HttpServletResponse) response;

    String sessionId;

    if (req instanceof AbstractHttpRequest)
      sessionId = ((AbstractHttpRequest) req).getRequestedSessionIdNoVary();
    else
      sessionId = req.getRequestedSessionId();

    TempBuffer tempBuf = TempBuffer.allocate();
    byte []buf = tempBuf.getBuffer();

    try {
      InputStream is = request.getInputStream();
      int len = is.read(buf, 0, buf.length);
      boolean isComplete;

      if (len < 0)
        isComplete = true;
      else if (len < buf.length) {
        int sublen = is.read(buf, len, 1);

        if (sublen < 0)
          isComplete = true;
        else {
          len += sublen;
          isComplete = false;
        }
      }
      else
        isComplete = false;

      ServerConnector client = null;
      int result = OK|EXIT;

      ClusterStream stream = openServer(sessionId, null);

      // If everything fails, return an error
      if (stream == null) {
        if (! res.isCommitted())
          res.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE);
        return;
      }

      client = stream.getServer();
      int index = client.getIndex();

      ReadStream rs = stream.getReadStream();
      WriteStream ws = stream.getWriteStream();
      boolean keepalive = false;
      result = OK|EXIT;
      
      long requestStart = Alarm.getCurrentTime();
      String failoverReason = null;

      try {
        result = handleRequest(req, res, stream, rs, ws, index,
                               buf, len, isComplete, isComplete);

        if ((result & STATUS_MASK) == OK) {
          return;
        } else if ((result & STATUS_MASK) == BUSY) {
          client.busy();
          failoverReason = "busy";
        } else {
          client.failSocket(requestStart);
          failoverReason = "socket failure";
        }
      } catch (ClientDisconnectException e) {
        log.fine(stream.getDebugId() + e.toString());
        log.log(Level.FINER, stream.getDebugId() + e.toString(), e);
        
        failoverReason = "socket disconnect";
        result = FAIL|EXIT;

        // #2236
        return;
      } catch (IOException e) {
        client.failSocket(requestStart);

        log.log(Level.WARNING, stream.getDebugId() + e.toString(), e);

        failoverReason = "socket failure";
        result = FAIL|EXIT;
      } finally {
        if ((result & EXIT_MASK) == QUIT)
          stream.free();
        else
          stream.close();
      }

      // server/2675
      if (isComplete && (result & STATUS_MASK) == BUSY
          || "GET".equals(req.getMethod())) {
        stream = openServer(sessionId, client);

        // If everything fails, return an error
        if (stream == null) {
          if (log.isLoggable(Level.INFO))
            log.info("load-balance failed" + (client != null ? (" for " + client.getDebugId()) : ""));

          if (! res.isCommitted())
            res.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE);
          return;
        }

        ServerConnector client2 = stream.getServer();
        index = client2.getIndex();

        if (log.isLoggable(Level.INFO))
          log.info("load-balance retrying "
                   + (sessionId != null ? "session " + sessionId : "new session ")  
                   + (client != null ? (" from " + client.getDebugId()) : "")
                   + " on " + client2.getDebugId()
                   + " due to "
                   + failoverReason);

        rs = stream.getReadStream();
        ws = stream.getWriteStream();
        result = OK|EXIT;
        
        requestStart = Alarm.getCurrentTime();

        try {
          result = handleRequest(req, res, stream, rs, ws, index,
                                 buf, len, isComplete, false);

          if ((result & STATUS_MASK) == OK)
            return;
          else if ((result & STATUS_MASK) == BUSY)
            client2.busy();
          else
            client2.failSocket(requestStart);
        } catch (IOException e) {
          client2.failSocket(requestStart);

          log.log(Level.FINE, e.toString(), e);
        } finally {
          if ((result & EXIT_MASK) == QUIT)
            stream.free();
          else
            stream.close();
        }
      }

      // POST can only fail over if the first returned a BUSY and
      // the data fits into a single chunk
      if (! res.isCommitted())
        res.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE);
    } finally {
      TempBuffer.free(tempBuf);
      tempBuf = null;
    }
  }

  /**
   * Opens the next available server.
   */
  private ClusterStream openServer(String sessionId, ServerConnector oldSrun)
  {
    ClusterStream stream = null;

    stream = openSessionServer(sessionId, oldSrun);
    if (stream != null) {
      if (log.isLoggable(Level.FINE))
        log.fine("Using session server " + stream.getServer() 
                 + " for session " + sessionId);
      return stream;
    }

    return _strategy.openServer(oldSrun);
  }

  /**
   * Opens the sticky session server.
   *
   * @param sessionId the request session cookie
   * @param oldSrun on failover, the original server
   */
  private ClusterStream openSessionServer(String sessionId,
                                          ServerConnector oldSrun)
  {
    if (sessionId == null || sessionId.length() < 3 || ! _isStickySessions)
      return null;

    int length = _srunClients.length;

    for (int i = 0; i < 3; i++) {
      int index = decode(sessionId.charAt(i)) % length;

      ServerConnector client = _srunClients[index];
      if (client == null || client == oldSrun)
        continue;

      ClusterStream stream = client.openForSession();
      if (stream != null)
        return stream;

      if (i == 0 && log.isLoggable(Level.FINE)) {
        log.fine("primary load-balance " + client.getDebugId()  
                 + " not available for session " + sessionId);
      }
    }

    return null;
  }

  private int handleRequest(HttpServletRequest req,
                            HttpServletResponse res,
                            ClusterStream stream,
                            ReadStream rs, WriteStream ws, int index,
                            byte []buf, int length, boolean isComplete,
                            boolean allowBusy)
    throws ServletException, IOException
  {
    AbstractHttpRequest cauchoReq = null;

    if (req instanceof AbstractHttpRequest)
      cauchoReq = (AbstractHttpRequest) req;

    int channel = 2;

    CharBuffer cb = new CharBuffer();

    ws.write(HmuxRequest.HMUX_CHANNEL);
    ws.write(channel >> 8);
    ws.write(channel);

    String uri;

    if (cauchoReq != null)
      uri = cauchoReq.getPageURI();
    else
      uri = (String) req.getAttribute(REQUEST_URI);

    String queryString;

    boolean isDebugFiner = log.isLoggable(Level.FINER);

    if (uri != null) {
      if (cauchoReq != null)
        queryString = cauchoReq.getPageQueryString();
      else
        queryString = (String) req.getAttribute(QUERY_STRING);
    }
    else {
      uri = req.getRequestURI();
      queryString = req.getQueryString();
    }

    if (req.isRequestedSessionIdFromUrl()) {
      if (_altSessionPrefix != null) {
        cb.clear();
        cb.append(_altSessionPrefix);
        cb.append(req.getRequestedSessionId());
        cb.append(uri);

        writeRequestString(ws, HmuxRequest.HMUX_URI,
                           cb.getBuffer(), 0, cb.getLength());
      }
      else {
        cb.clear();
        cb.append(uri);
        cb.append(_sessionPrefix);
        cb.append(req.getRequestedSessionId());

        writeRequestString(ws, HmuxRequest.HMUX_URI,
                           cb.getBuffer(), 0, cb.getLength());
      }
    }
    else
      writeRequestString(ws, HmuxRequest.HMUX_URI, uri);

    if (isDebugFiner)
      logFinest(stream,  "URL " + uri);

    writeRequestString(ws, HmuxRequest.CSE_SERVER_TYPE, "Resin");

    cb.clear();
    cb.append(req.getServerName());

    if (cb.equalsIgnoreCase(Management.HOST_NAME)) {
      log.warning(L.l("management service attempt through load balancer from ip {0}",
                      req.getRemoteAddr()));

      res.sendError(HttpServletResponse.SC_NOT_FOUND);

    }
    cb.append(':');
    cb.append(req.getServerPort());
    writeRequestString(ws, HmuxRequest.HMUX_SERVER_NAME,
                       cb.getBuffer(), 0, cb.getLength());

    writeRequestInt(ws, HmuxRequest.CSE_SERVER_PORT,
                    req.getServerPort());
    writeRequestString(ws, HmuxRequest.CSE_REMOTE_HOST,
                       req.getRemoteHost());
    writeRequestString(ws, HmuxRequest.CSE_REMOTE_ADDR,
                       req.getRemoteAddr());
    /* XXX: backend is responsible
    String remoteUser = req.getRemoteUser();
    if (remoteUser != null)
      writeRequestString(ws, HmuxRequest.CSE_REMOTE_USER,
                         remoteUser);
    */
    writeRequestString(ws, HmuxRequest.HMUX_METHOD,
                       req.getMethod());
    writeRequestString(ws, HmuxRequest.CSE_QUERY_STRING, queryString);
    writeRequestString(ws, HmuxRequest.CSE_PROTOCOL,
                       req.getProtocol());

    if (req.isSecure())
      writeRequestString(ws, HmuxRequest.CSE_IS_SECURE, "");

    int headerSize;
    if (cauchoReq != null && (headerSize = cauchoReq.getHeaderSize()) >= 0) {
      for (int i = 0; i < headerSize; i++) {
        CharSegment name = cauchoReq.getHeaderKey(i);
        CharSegment value = cauchoReq.getHeaderValue(i);

        writeRequestString(ws, HmuxRequest.HMUX_HEADER,
                           name.getBuffer(), name.getOffset(), name.getLength());
        writeRequestString(ws, HmuxRequest.HMUX_STRING,
                           value.getBuffer(), value.getOffset(), value.getLength());

        if (isDebugFiner)
          logFinest(stream, name + ": " + value);
      }
    }
    else {
      Enumeration e = req.getHeaderNames();
      while (e.hasMoreElements()) {
        String name = (String) e.nextElement();
        String value = req.getHeader(name);

        writeRequestString(ws, HmuxRequest.HMUX_HEADER, name);
        writeRequestString(ws, HmuxRequest.HMUX_STRING, value);

        if (isDebugFiner)
          logFinest(stream, name + ": " + value);
      }
    }

    InputStream is = req.getInputStream();
    boolean isFirst = true;
    OutputStream os;

    os = res.getOutputStream();

    CharBuffer key = cb;
    CharBuffer value = new CharBuffer();
    boolean hasHeader = ! req.getProtocol().equals("HTTP/0.9");
    int outChannel = -1;
    int len;

    if (length > 0)
      writeRequestData(ws, HmuxRequest.HMUX_DATA, buf, length);

  loop:
    while (! isComplete && (len = is.read(buf, 0, buf.length)) > 0) {
      writeRequestData(ws, HmuxRequest.HMUX_DATA, buf, len);

      if (isDebugFiner)
        logFinest(stream, "Y (yield)");

      ws.write(HmuxRequest.HMUX_YIELD);

      boolean hasStatus = false;

      while (true) {
        int code = rs.read();

        if (code < 0) {
          if (isDebugFiner)
            logFinest(stream, "end of file");

          if (hasStatus)
            return OK|EXIT;
          else {
            logConfig(stream, "unexpected end of file");

            return FAIL|EXIT;
          }
        }
        else if (code == HmuxRequest.HMUX_QUIT) {
          if (isDebugFiner)
            logFinest(stream, "Q (keepalive)");

          if (hasStatus)
            return OK|QUIT;
          else {
            logConfig(stream, "unexpected Q (quit) file");

            return FAIL|QUIT;
          }
        }
        else if (code == HmuxRequest.HMUX_EXIT) {
          if (isDebugFiner)
            logFinest(stream, "X (close)");

          if (hasStatus)
            return OK|EXIT;
          else {
            logConfig(stream, "unexpected X (close) file");

            return FAIL|EXIT;
          }
        }
        else if (code == HmuxRequest.HMUX_YIELD) {
          if (isDebugFiner)
            logFinest(stream, "Y (yield)");

          continue;
        }

        int sublen = readLength(rs);

        if (code == HmuxRequest.CSE_ACK) {
          if (isDebugFiner)
            logFinest(stream,  " ack:" + sublen);

          break;
        }
        else if (code == HmuxRequest.HMUX_CHANNEL) {
          outChannel = sublen;
          log.finer("Hmux channel:" + outChannel);
        }
        else if (code == HmuxRequest.HMUX_STATUS && hasHeader) {
          value.clear();
          rs.readAll(value, sublen);

          int status = 0;
          for (int i = 0; i < 3; i++)
            status = 10 * status + value.charAt(i) - '0';

          if (isDebugFiner)
            logFinest(stream, value.toString());

          if (status != 200)
            res.setStatus(status, value.substring(4));

          hasStatus = true;
        }
        else if (code == HmuxRequest.HMUX_HEADER && hasHeader) {
          key.clear();
          rs.readAll(key, sublen);

          code = rs.read();
          sublen = readLength(rs);
          value.clear();
          rs.readAll(value, sublen);

          if (isDebugFiner)
            logFinest(stream, key + ":" + value);

          res.addHeader(key.toString(), value.toString());
        }
        else if (code == HmuxRequest.HMUX_DATA) {
          if (isDebugFiner)
            logFinest(stream, "D " + sublen);

          rs.writeToStream(os, sublen);
        }
        else if (code == HmuxRequest.HMUX_META_HEADER) {
          key.clear();
          rs.readAll(key, sublen);

          code = rs.read();
          sublen = readLength(rs);
          value.clear();
          rs.readAll(value, sublen);

          if (isDebugFiner)
            logFinest(stream, key + ":" + value);

          if (key.matches("cpu-load")) {
            double loadAvg = 0.001 * Long.parseLong(value.toString());

            stream.getServer().setCpuLoadAvg(loadAvg);
          }
        }
        else
          rs.skip(sublen);
      }
    }

    ws.write(HmuxRequest.HMUX_QUIT);

    int code;

    if (outChannel < 0) {
      try {
        code = rs.read();
      } catch (IOException e1) {
        log.log(Level.FINE, stream.getDebugId() + e1.toString(), e1);

        return FAIL|EXIT;
      }

      if (allowBusy
          && (code < 0
              || code == HmuxRequest.HMUX_QUIT
              || code == HmuxRequest.HMUX_EXIT)) {
        stream.clearRecycle();

        logConfig(stream, "unexpected end of file " + (code) + " " + (char) code);

        return FAIL|EXIT;
      }
      else if (code != HmuxRequest.HMUX_CHANNEL) {
        log.warning(stream.getDebugId() + " hmux protocol error at " + code + " " + (char) code);
        res.sendError(res.SC_SERVICE_UNAVAILABLE);

        return FAIL|EXIT;
      }

      outChannel = readLength(rs);
    }

    // #2369 - A slow modem can cause the app-tier and web-tier times
    // to get out of sync, with the app-tier thinking it's completed
    // (and starts the keepalive timeout) 30s before the web-tier reads
    // its data.
    // As a temporary measure, we start the idle time at the first data
    // read (later we might mark the time it takes to read an app-tier
    // packet.  If it's short, e.g. 250ms, don't update the time.)
    stream.setFreeTime(Alarm.getCurrentTime());

    boolean isBusy = false;

    while (true) {
      code = rs.read();

      if (code <= 0) {
        if (isDebugFiner)
          logFinest(stream, "end of file");

        return isBusy ? BUSY|EXIT : OK|EXIT;
      }
      else if (code == HmuxRequest.HMUX_QUIT) {
        if (isDebugFiner)
          logFinest(stream, "Q (keepalive)");

        return isBusy ? BUSY|QUIT : OK|QUIT;
      }
      else if (code == HmuxRequest.HMUX_EXIT) {
        if (isDebugFiner)
          logFinest(stream, "X (close)");

        return isBusy ? BUSY|EXIT : OK|EXIT;
      }
      else if (code == HmuxRequest.HMUX_YIELD) {
        if (isDebugFiner)
          logFinest(stream, "Y (yield)");

        continue;
      }

      int v1 = rs.read();
      int v2 = rs.read();

      if (v2 < 0)
        return OK|EXIT;

      int sublen = (v1 << 8) + v2;

      if (code == HmuxRequest.HMUX_DATA) {
        if (! isBusy)
          rs.writeToStream(os, sublen);
        else
          rs.skip(sublen);

        if (isDebugFiner)
          logFinest(stream, "D: " + sublen);
      }
      else if (code == HmuxRequest.HMUX_STATUS && hasHeader) {
        value.clear();
        rs.readAll(value, sublen);

        int status = 0;
        for (int i = 0; i < 3; i++)
          status = 10 * status + value.charAt(i) - '0';

        if (status == 503 && allowBusy)
          isBusy = true;
        else if (status != 200)
          res.setStatus(status, value.substring(4));

        if (isDebugFiner)
          logFinest(stream, value.toString());
      }
      else if (code == HmuxRequest.HMUX_HEADER && hasHeader) {
        key.clear();
        rs.readAll(key, sublen);

        code = rs.read();
        sublen = readLength(rs);
        value.clear();
        rs.readAll(value, sublen);

        if (! isBusy)
          res.addHeader(key.toString(), value.toString());

        if (isDebugFiner)
          logFinest(stream, key + ":" + value);
      }
      else if (code == HmuxRequest.HMUX_META_HEADER) {
        key.clear();
        rs.readAll(key, sublen);

        code = rs.read();
        sublen = readLength(rs);
        value.clear();
        rs.readAll(value, sublen);

        if (isDebugFiner)
          logFinest(stream, "M " + key + ":" + value);

        if (key.matches("cpu-load")) {
          double loadAvg = 0.001 * Long.parseLong(value.toString());

          stream.getServer().setCpuLoadAvg(loadAvg);
        }
      }
      else
        rs.skip(sublen);
    }
  }

  private void logInfo(ClusterStream stream, String msg)
  {
    log.info("load balance " + stream.getDebugId() + " " + msg);
  }

  private void logConfig(ClusterStream stream, String msg)
  {
    log.config("load balance " + stream.getDebugId() + " " + msg);
  }

  private void logFinest(ClusterStream stream, String msg)
  {
    log.finest("load balance " + stream.getDebugId() + " " + msg);
  }

  private int readLength(ReadStream is)
    throws IOException
  {
    int d1 = is.read();
    int d2 = is.read();

    if (d2 < 0)
      return -1;
    else
      return (d1 << 8) + d2;
  }

  private void writeRequestString(WriteStream s, int code, String string)
    throws IOException
  {
    if (string == null) {
      s.write(code);
      s.write(0);
      s.write(0);
    }
    else {
      int length = string.length();
      s.write(code);
      s.write(length >> 8);
      s.write(length);

      s.print(string, 0, length);
    }
  }

  private void writeRequestString(WriteStream s, int code,
                                  char []buffer, int offset, int length)
    throws IOException
  {
    s.write(code);
    s.write(length >> 8);
    s.write(length);

    s.print(buffer, offset, length);
  }

  private void writeRequestInt(WriteStream s, int code, int value)
    throws IOException
  {
    CharBuffer cb = new CharBuffer();

    cb.append(value);
    int length = cb.length();

    s.write(code);
    s.write(length >> 8);
    s.write(length);

    s.print(cb.getBuffer(), 0, cb.getLength());
  }

  private void writeRequestData(WriteStream s, int code)
    throws IOException
  {
    s.write(code);
    s.write(0);
    s.write(0);
  }

  private void writeRequestData(WriteStream os,
                                int code, byte []data, int length)
    throws IOException
  {
    os.write(code);
    os.write(length >> 8);
    os.write(length);

    os.write(data, 0, length);
  }

  public void destroy()
  {
    super.destroy();

    for (int i = 0; i < _srunClients.length; i++) {
      ServerConnector client = _srunClients[i];

      if (client != null)
        client.close();
    }
  }

  /**
   * Converts an integer to a printable character
   */
  private static char convert(long code)
  {
    code = code & 0x3f;

    if (code < 26)
      return (char) ('a' + code);
    else if (code < 52)
      return (char) ('A' + code - 26);
    else if (code < 62)
      return (char) ('0' + code - 52);
    else if (code == 62)
      return '_';
    else
      return '-';
  }

  private static int decode(int code)
  {
    return DECODE[code & 0x7f];
  }

  static {
    DECODE = new int[128];
    for (int i = 0; i < 64; i++)
      DECODE[(int) convert(i)] = i;
  }
}
