/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.servlets;

import com.caucho.server.cluster.ClusterStream;
import com.caucho.server.cluster.ServerConnector;
import com.caucho.util.Alarm;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Strategy for selecting the next server
 */
public class LeastConnectionBalanceStrategy extends LoadBalanceStrategy
{
  private static final Logger log
    = Logger.getLogger(HmuxLoadBalanceServlet.class.getName());

  private int _cpuWeight = 0;
  private int _connectionWeight = 100;

  public void setCpuWeight(int weight)
  {
    _cpuWeight = weight;
  }

  public void setConnectionWeight(int weight)
  {
    _connectionWeight = weight;
  }
  
  /**
   * Opens the best load balance server.
   */
  protected ClusterStream openBestServer(int startIndex,
					 ServerConnector oldSrun)
  {
    int length = _serverGroup.length;
    
    ClusterStream stream = null;
    ServerConnector srun = null;

    double bestCost = Double.MAX_VALUE;
    srun = null;

    synchronized (this) {
      // first use the srun with the least number of active connections
      for (int i = 0; i < length; i++) {
	int index = (startIndex + i) % length;
	ServerConnector client = _serverGroup[index];

	if (client == null)
	  continue;
      
	// skip servers in warmup phase
	if (! client.canOpenSoftOrRecycle()) {
	  continue;
	}

	if (client == oldSrun)
	  continue;

	long weight = client.getLoadBalanceWeight();
	if (weight <= 0)
	  continue;

	long connections = (client.getActiveCount()
			    + client.getLoadBalanceAllocateCount());
	double cpu = client.getCpuLoadAvg();

	double cost = 0.0;

	cost += _connectionWeight * connections;
	cost += _cpuWeight * cpu;

	double latencyFactor = client.getLatencyFactor();

	if (connections > 0) {
	  cost += _connectionWeight * latencyFactor / 100;
	}
      
	cost = cost / (double) weight;
	if (cost < bestCost) {
	  srun = client;
	  bestCost = cost;
	}

	if (log.isLoggable(Level.FINER)) {
	  log.finer("load balance cost: " + cost
		    + " conn:" + connections
		    + " cpu:" + cpu
		    + " latency:" + latencyFactor
		    + " " + client);
	}
      }

      if (srun != null) {
	srun.allocateLoadBalance();
      }
    }

    if (srun != null) {
      try {
	stream = srun.openSoft();

	return stream;
      } finally {
	srun.freeLoadBalance();
      }
    }
    else
      return null;
  }
}
