/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.servlets;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.caucho.server.cluster.ClusterStream;
import com.caucho.server.cluster.ServerConnector;
import com.caucho.util.L10N;

/**
 * Strategy for selecting the next server
 */
abstract public class LoadBalanceStrategy 
{
  static protected final Logger log
    = Logger.getLogger(HmuxLoadBalanceServlet.class.getName());
  static final L10N L = new L10N(HmuxLoadBalanceServlet.class);
  
  protected ServerConnector []_serverGroup;
  private int _roundRobin = -1;

  /**
   * Sets the server group.
   */
  void setServerGroup(ServerConnector []serverGroup)
  {
    _serverGroup = serverGroup;
  }
  
  /**
   * Opens the next available server.
   *
   * @param oldClient on failover, the client which failed with 503
   *
   * @return the next available server
   */
  public final ClusterStream openServer(ServerConnector oldClient)
  {
    int startIndex = nextRoundRobin();
    
    ClusterStream stream = null;

    stream = openBestServer(startIndex, oldClient);
    if (stream != null) {
      if (log.isLoggable(Level.FINE))
        log.fine("Using best available server: " 
                 + stream.getServer().getDebugId());
      return stream;
    }

    stream = openBackupServer(startIndex,  oldClient);
    if (stream != null) {
      if (log.isLoggable(Level.FINE))
        log.fine("Using backup server: " 
                 + stream.getServer().getDebugId());
    }
    
    return stream;
  }
  
  /**
   * Opens the best load balance server.
   *
   * @param oldClient on failover, the client which failed with 503
   */
  protected ClusterStream openBestServer(int startIndex,
					 ServerConnector oldClient)
  {
    return null;
  }

  /**
   * Open a backup server.
   */
  private ClusterStream openBackupServer(int startIndex, ServerConnector oldSrun)
  {
    int length = _serverGroup.length;

    ServerConnector client;
    ClusterStream stream;
    
    // second time, try the first srun that works.
    for (int i = 0; i < length; i++) {
      int index = (startIndex + i) % length;
      client = _serverGroup[index];

      if (client == null || client == oldSrun)
	continue;

      stream = client.openSoft();
      if (stream != null)
	return stream;
    }

    // third time force the open
    for (int i = 0; i < length; i++) {
      int index = (startIndex + i) % length;
      client = _serverGroup[index];

      if (client == null || client == oldSrun)
	continue;

      stream = client.open();
      if (stream != null)
	return stream;
    }

    return null;
  }

  /**
   * Returns the index of the next round-robin server.
   */
  protected int nextRoundRobin()
  {
    int length = _serverGroup.length;
    
    _roundRobin = (_roundRobin + 1) % length;

    for (int i = 0; i < length; i++) {
      int index = _roundRobin;

      ServerConnector server = _serverGroup[index];

      if (server == null)
	continue;
      
      if (server.canOpenSoft())
        return _roundRobin;
      
      _roundRobin = (_roundRobin + 1) % length;
    }

    return _roundRobin;
  }
}
