/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * @author Scott Ferguson
 */

package com.caucho.vfs;

import com.caucho.license.LicenseCheck;
import com.caucho.log.Log;
import com.caucho.server.util.CauchoSystem;
import com.caucho.util.Alarm;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Map;
import java.util.logging.Logger;

/**
 * FilePath implements the native filesystem.
 */
public class JniFilePathImpl extends FilePath {
  private static final Logger log = Log.open(JniFilePathImpl.class);

  private static boolean _isJNIEnabled;
  private static boolean _isEnabled;
  private static boolean _isInit;
  private static Throwable _jniInitException;

  private byte []_bytes;

  private long _lastStatTime = 0L;
  private FileStatus _lastStat = null;
  private boolean _doLstat = false;

  /**
   * @param path canonical path
   */
  protected JniFilePathImpl(FilesystemPath root, String userPath, String path)
  {
    super(root, userPath, path);

    _isWindows = CauchoSystem.isWindows();
  }

  /**
   * @param path canonical path
   */
  private JniFilePathImpl(String userPath, String path, boolean isRoot)
  {
    super(null, userPath, path);

    _isWindows = CauchoSystem.isWindows();
    _root = this;
  }

  public JniFilePathImpl()
  {
    this(null);
  }

  JniFilePathImpl(String path)
  {
    this(Vfs.getGlobalPwd() != null ? Vfs.getGlobalPwd().getRoot() : null,
	 path, normalizePath("/", initialPath(path),
                             0, CauchoSystem.getFileSeparatorChar()));
    if (_root == null) {
      _root = new JniFilePathImpl("/", "/", true);
    }
  }

  /**
   * Returns true if the JNI file path exists.
   */
  public static boolean isEnabled()
  {
    if (! _isInit) {
      _isInit = true;

      try {
	if (_isJNIEnabled) {
	  Class cl = Class.forName("com.caucho.license.LicenseCheckImpl");
	  LicenseCheck license = (LicenseCheck) cl.newInstance();

	  license.validate(1);
	}
	
	_isEnabled = _isJNIEnabled;
      } catch (Throwable e) {
	log.fine(e.toString());
      }
    }

    return _isEnabled;
  }

  /**
   * Lookup the actual path relative to the filesystem root.
   *
   * @param userPath the user's path to lookup()
   * @param attributes the user's attributes to lookup()
   * @param path the normalized path
   *
   * @return the selected path
   */
  public Path fsWalk(String userPath,
		     Map<String,Object> attributes,
		     String path)
  {
    FilePath newPath = new JniFilePathImpl(_root, userPath, path);

    if (newPath._isWindows && newPath.isAux())
      return new FilePath(_root, userPath, path);
    else
      return newPath;
  }

  /**
   * Returns true if the path itself is cacheable
   */
  @Override
  protected boolean isPathCacheable()
  {
    return true;
  }

  /**
   * Returns true if the file exists.
   */
  public boolean exists()
  {
    return getStatus() != null;
  }

  /**
   * Returns the last access time of the file.
   *
   * @return 0 for non-files.
   */
  public long getLastAccessTime()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getAtime() * 1000L;
  }

  /**
   * Returns the last modified time.
   */
  public long getLastModified()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getMtime() * 1000L;
  }

  /**
   * Returns equivalent of struct stat.st_ctime if appropriate.
   */
  public long getLastStatusChangeTime()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getCtime() * 1000L;
  }

  /**
   * Returns the length modified time.
   */
  public long getLength()
  {
    FileStatus status = getStatus();

    if (status == null)
      return -1;
    else
      return status.getSize();
  }

  /**
   * Sets the file mode.
   */
  public boolean chmod(int mode)
  {
    byte []bytes = getBytes();

    boolean result = nativeChmod(bytes, bytes.length, mode) >= 0;

    _lastStatTime = 0;
    
    return result;
  }

  /**
   * Sets the file mode.
   */
  @Override
  public boolean changeOwner(String owner)
  {
    if (owner == null)
      return false;
    
    byte []bytes = getBytes();

    boolean result = nativeChangeOwner(bytes, bytes.length, owner);

    _lastStatTime = 0;
    
    return result;
  }

  /**
   * Returns true if the file can be read.
   */
  public boolean canRead()
  {
    byte []bytes = getBytes();

    return nativeCanRead(bytes);
  }

  /**
   * Returns true if the file is a directory.
   */
  public boolean isDirectory()
  {
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isDirectory();
  }  
  
  /**
   * Returns true if the file is a regular file.
   */
  public boolean isFile()
  {
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isRegularFile();
  }

  /**
   * Returns true if the file is a symbolic link.
   */
  public boolean isLink()
  {
    _doLstat = true;
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isLink();
  }

  /**
   * Tests if the path refers to a socket.
   */
  public boolean isSocket()
  {
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isSocket();
  }

  /**
   * Tests if the path refers to a FIFO.
   */
  public boolean isFIFO()
  {
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isFIFO();
  }

  /**
   * Tests if the path refers to a block device.
   */
  public boolean isBlockDevice()
  {
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isBlockDevice();
  }

  /**
   * Tests if the path refers to a block device.
   */
  public boolean isCharacterDevice()
  {
    FileStatus status = getStatus();

    if (status == null)
      return false;
    else
      return status.isCharacterDevice();
  }

  /**
   * Returns the crc64 code.
   */
  @Override
  public long getCrc64()
  {
    byte []bytes = getBytes();

    long crc64 = nativeCrc64(bytes, bytes.length);

    if (crc64 != 0)
      return crc64;
    else
      return super.getCrc64();
  }

  /**
   * Remove the underlying file.
   */
  @Override
  public boolean remove()
  {
    _lastStat = null;
    _lastStatTime = 0;
    
    return super.remove();
  }

  /**
   * Truncate the file.
   */
  @Override
  public boolean truncate(long length)
    throws IOException
  {
    _lastStat = null;
    _lastStatTime = 0;

    byte []bytes = getBytes();

    nativeTruncate(bytes, bytes.length);
    
    return true;
  }

  /**
   * Returns the stream implementation for a read stream.
   */
  public StreamImpl openReadImpl() throws IOException
  {
    if (_separatorChar == '\\' && isAux())
      throw new FileNotFoundException(getNativePath());

    byte []bytes = getBytes();

    int fd = nativeOpenRead(bytes, bytes.length);

    if (fd < 0)
      throw new FileNotFoundException(getNativePath());

    JniFileStream stream = new JniFileStream(fd, true, false);
    stream.setPath(this);

    return stream;
  }

  /**
   * Returns the stream implementation for a write stream.
   */
  public StreamImpl openWriteImpl() throws IOException
  {
    if (_separatorChar == '\\' && isAux())
      throw new FileNotFoundException(getNativePath());

    byte []bytes = getBytes();

    int fd = nativeOpenWrite(bytes, bytes.length, false);
    _lastStat = null;
    _lastStatTime = 0;

    if (fd < 0)
      throw new FileNotFoundException(getNativePath());

    JniFileStream stream = new JniFileStream(fd, false, true);
    stream.setPath(this);

    return stream;
  }

  /**
   * Returns the stream implementation for an append stream.
   */
  public StreamImpl openAppendImpl() throws IOException
  {
    if (_separatorChar == '\\' && isAux())
      throw new FileNotFoundException(getNativePath());

    byte []bytes = getBytes();

    int fd = nativeOpenWrite(bytes, bytes.length, true);
    _lastStat = null;
    _lastStatTime = 0;

    if (fd < 0)
      throw new FileNotFoundException(getNativePath());

    JniFileStream stream = new JniFileStream(fd, false, true);
    stream.setPath(this);

    return stream;
  }

  /**
   * Creates a link named by this path to another path.
   *
   * @param target the target of the link
   * @param hardLink true if the link should be a hard link
   */
  public boolean createLink(Path target, boolean hardLink)
    throws IOException
  {
    if (hardLink && ! target.exists())
      throw new FileNotFoundException(target.getNativePath());

    if (exists())
      throw new IOException(getFullPath() + " already exists");

    byte []bytes = getBytes();
    byte []targetBytes = getBytes(target);

    return nativeLink(bytes, targetBytes, hardLink);
  }

  /**
   * Returns equivalent of struct stat.st_dev if appropriate.
   */
  public long getDevice()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getDev();
  }

  /**
   * Returns equivalent of struct stat.st_ino if appropriate.
   */
  public long getInode()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getIno();
  }

  /**
   * Returns equivalent of struct stat.st_mode if appropriate.
   */
  public int getMode()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getMode();
  }

  /**
   * Returns equivalent of struct stat.st_nlink if appropriate.
   */
  public int getNumberOfLinks()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getNlink();
  }

  /**
   * Returns equivalent of struct stat.st_uid if appropriate.
   */
  public int getUser()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getUid();
  }

  /**
   * Returns equivalent of struct stat.st_gid if appropriate.
   */
  public int getGroup()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getGid();
  }

  /**
   * Returns equivalent of struct stat.st_rdev if appropriate.
   */
  public long getDeviceId()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getRdev();
  }

  /**
   * Returns equivalent of struct stat.st_blksize if appropriate.
   */
  public long getBlockSize()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getBlksize();
  }

  /**
   * Returns equivalent of struct stat.st_blocks if appropriate.
   */
  public long getBlockCount()
  {
    FileStatus status = getStatus();

    if (status == null)
      return 0;
    else
      return status.getBlocks();
  }

  private FileStatus getStatus()
  {
   long now;

   now = Alarm.getCurrentTime();

    // Wait at least 500ms between stats, limited by OS timestamp
    // XXX: This optimization might need to be removed at some point 
    if (now - _lastStatTime <= 500L && _lastStat != null && ! Alarm.isTest())
      return _lastStat;
    
    byte []bytes = getBytes();

    _lastStat = nativeStat(bytes, _doLstat);
    
    _lastStatTime = now;

    return _lastStat;
  }

  @Override
  protected Path copy()
  {
    return new JniFilePathImpl(getRoot(), getUserPath(), getPath());
  }

  /**
   * Returns the byte representation of the file.
   */
  private byte []getBytes()
  {
    if (_bytes == null) {
      try {
        String encoding = System.getProperty("file.encoding");
        _bytes = getNativePath().getBytes(encoding);
      } catch (RuntimeException e) {
	throw e;
      } catch (Exception e) {
        throw new RuntimeException(e);
      }
    }

    return _bytes;
  }

  /**
   * Returns the byte representation of the file.
   */
  private byte []getBytes(Path path)
  {
    try {
      String encoding = System.getProperty("file.encoding");
      return path.getNativePath().getBytes(encoding);
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private static native boolean nativeIsEnabled();
  private static native int nativeGetLastModified(byte []name, int length);
  private static native int nativeGetLength(byte []name, int length);
  private static native boolean nativeCanRead(byte []name);
  private static native boolean nativeIsDirectory(byte []name);
  private static native long nativeCrc64(byte []name, int length);
  private static native int nativeOpenRead(byte []name, int length);
  private static native int nativeOpenWrite(byte []name, int length,
					    boolean isAppend);
  private static native int nativeTruncate(byte []name, int length);
  private static native int nativeChmod(byte []name, int length, int mode);
  private static native boolean nativeChangeOwner(byte []name, int length,
						  String userName);
  private static native FileStatus nativeStat(byte []name, boolean doLstat);
  private static native boolean nativeLink(byte []name, byte []target, 
                                           boolean hardLink) 
    throws IOException;
  private static native String nativeReadLink(byte []name);
  private static native String nativeRealPath(byte []name);

  static {
    try {
      byte []test = "test".getBytes(System.getProperty("file.encoding"));

      System.loadLibrary("resin");
      _isJNIEnabled = nativeIsEnabled();
    } catch (Throwable e) {
      _jniInitException = e;
      
      /*
      System.err.println("Can't load JNI VFS library.  Resin will still run,");
      System.err.println("but with somewhat slower performance.");
      System.err.println("    " + e);
      if (File.separatorChar == '/') {
	System.err.println("  On Unix, you need to compile the JNI library with");
	System.err.println("    ./configure; make; make install");
      }
      if (File.separatorChar == '\\') {
	System.err.println("  On Windows, check for the resin.dll in resin/bin and");
	System.err.println("  check the dll load path.");
	System.err.println();
      }
      */
    }
  }
}
